/*
** Daedalus (Version 2.5) File: create2.cpp
** By Walter D. Pullen, Astara@msn.com, http://www.astrolog.org/labyrnth.htm
**
** IMPORTANT NOTICE: Daedalus and all Maze generation and general
** graphics routines used in this program are Copyright (C) 1998-2013 by
** Walter D. Pullen. Permission is granted to freely use, modify, and
** distribute these routines provided these credits and notices remain
** unmodified with any altered or distributed versions of the program.
** The user does have all rights to Mazes and other graphic output
** they make in Daedalus, like a novel created in a word processor.
**
** More formally: This program is free software; you can redistribute it
** and/or modify it under the terms of the GNU General Public License as
** published by the Free Software Foundation; either version 2 of the
** License, or (at your option) any later version. This program is
** distributed in the hope that it will be useful and inspiring, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** General Public License for more details, a copy of which is in the
** LICENSE.HTM included with Daedalus, and at http://www.gnu.org
**
** This file contains Maze creation algorithms, specifically algorithms that
** produce orthogonal Mazes, however which aren't standard 2D Mazes.
**
** Created: 11/22/1996.
** Last code change: 10/10/2013.
*/

#include <stdio.h>
#include <stdlib.h>
#include "util.h"
#include "graphics.h"
#include "color.h"
#include "threed.h"
#include "maze.h"


/*
******************************************************************************
** Nested Fractal Maze Routines
******************************************************************************
*/

// Quickly create a 2x2 passage Maze in a bitmap. The bitmap is assumed to be
// of appropriate size to contain it. Used to quickly create fractal Mazes
// that nest 2x2 Mazes.

void TwoByTwoGenerate(bitmap b)
{
  int d;

  BitmapOff(b);
  LineX(b, 0, 4, 0, fOn); LineX(b, 0, 4, 4, fOn);
  LineY(b, 0, 1, 3, fOn); LineY(b, 4, 1, 3, fOn);
  M1(b, 2, 2);

  // Randomly connect the center vertex to one of the four boundary walls.
  d = RndDir();
  M1(b, 2 + xoff[d], 2 + yoff[d]);
  MakeEntranceExit(b, 0);
  UpdateDisplay();
}


// Create a section of a fractal Maze. Recursively calls itself to create the
// subsections or Mazes within each cell of the Maze at this nesting level.

void FractalGenerate(bitmap b, bitmap bT, int z, int xp, int yp)
{
  int xsMax = 1, ysMax = 1, x, y, i;

  z--;
  for (i = 0; i < z; i++) {
    xsMax *= ms.xFractal; ysMax *= ms.yFractal;
  }

  // Create the Maze for this nesting level.
  if (ms.xFractal != 2 || ms.yFractal != 2)
    CreateMazePerfect(&bT);
  else
    TwoByTwoGenerate(bT);

  // Each Maze contains a top and left edge (but no bottom or right edge) with
  // a random entrance in those edges. Those entrances may or may not be
  // visible when this Maze is nested within surrounding Mazes. For the
  // outermost Maze, don't have either entrance (they'll be added later).
  if (z+1 < ms.zFractal)
    M0(bT, 0, (Rnd(1, ms.yFractal) << 1) - 1);
  else
    M1(bT, ms.xEntrance, ms.yEntrance);

  // Render the Maze at appropriate zoom level and location.
  BlockMoveMaze(bT, b, 0, 0, MX(bT)-2, MY(bT)-2, xp << 1, yp << 1,
    xsMax << 1, 0, 0, ysMax << 1);

  // Recursively create a Maze within each cell of this Maze.
  if (z > 0)
    for (y = 0; y < ms.yFractal; y++)
      for (x = 0; x < ms.xFractal; x++)
        FractalGenerate(b, bT, z, xp + x * xsMax, yp + y * ysMax);
}


// Create a new nested fractal Maze, composed of a Maze with smaller Mazes
// recursively nested within each cell. The fractal Maze's size is defined by
// the Fractal fields in the Create Settings dialog.

flag CreateMazeFractal(bitmap *pb)
{
  bitmap b, bT;
  char sz[cchSzMax];
  int xsMax = 1, ysMax = 1, i;

  // Ensure the fractal Maze will fit within a bitmap.
  for (i = 0; i < ms.zFractal; i++) {
    if (((long)xsMax * ms.xFractal << 1) + 1 > xBitmap ||
      ((long)ysMax * ms.yFractal << 1) + 1 > yBitmap) {
      sprintf(S(sz),
        "%d by %d fractal Maze can't have a nesting level greater than %d.",
        ms.xFractal, ms.yFractal, i); PrintSz_W(sz);
      return fFalse;
    }
    xsMax *= ms.xFractal; ysMax *= ms.yFractal;
  }

  // Create the bitmap for the fractal Maze, and a temporary internal bitmap
  // to store the Mazes at each level.
  if (!FBitmapSizeSet(pb, (xsMax << 1) + 1, (ysMax << 1) + 1))
    return fFalse;
  b = *pb;
  if (!FAllocateBitmap(&bT,
    (ms.xFractal << 1) + 1, (ms.yFractal << 1) + 1, b))
    return fFalse;
  BitmapOff(b);
  UpdateDisplay();
  SetXyh(bT);

  // Create the fractal Maze here.
  i = ms.nEntrancePos; ms.nEntrancePos = 3;
  FractalGenerate(b, bT, ms.zFractal, 0, 0);
  ms.nEntrancePos = i;

  // Draw the right and bottom boundary walls.
  SetXyh(b);
  LineY(b, xh, 1, yh-1, fOn);
  LineX(b, 1, xh, yh, fOn);
  MakeEntranceExit(b, 0);
  DeallocateBitmap(bT);
  return fTrue;
}


// Draw one subsection of a virtual fractal Maze using given random number
// seed. Called from FractalPartial() to create sections of the fractal Maze.

void FractalPartialGenerate(bitmap b, bitmap bT, int z, long x0, long y0,
  int xp, int yp, int nRndSeed, long *pxEntrance, long *pxExit)
{
  long xTotal, yTotal, xsMax = 1, ysMax = 1, xT, yT, xEntrance, xExit;
  int xpMax = 1, ypMax = 1, i, j, nRnd, xs, ys;
  flag fX, fY;

  for (i = 0; i < z; i++) {
    xpMax *= ms.xFractal; ypMax *= ms.yFractal;
  }
  for (i = z; i < ms.zFractal; i++) {
    xsMax *= ms.xFractal; ysMax *= ms.yFractal;
  }
  xTotal = xsMax * xpMax; yTotal = ysMax * ypMax;
  srand(nRndSeed);

  // Figure out where the entrance and exit to the fractal Maze should be.
  switch (ms.nEntrancePos)
    {
  case 0:
    xEntrance = 0;
    xExit     = xTotal - 1;
    break;
  case 1:
    xEntrance = (xTotal - 1) >> 1;
    xExit     = xEntrance + !FOdd(xTotal);
    break;
  case 2:
    xEntrance = RndL(0, xTotal - 1);
    xExit     = xTotal - 1 - xEntrance;
    break;
  case 3:
    xEntrance = RndL(0, xTotal - 1);
    xExit     = xTotal - 1 - RndL(0, xTotal - 1);
    break;
    }
  *pxEntrance = xEntrance; *pxExit = xExit;

  // Leave the section blank if it's outside the bounds of the whole Maze.
  if (x0 >= 0 && x0 < xsMax && y0 >= 0 && y0 < ysMax) {
    xT = xsMax; yT = ysMax;
    fX = fY = fFalse;

    // Get information about the Mazes surrounding the starting nesting level.
    for (i = z; i < ms.zFractal; i++) {
      if (ms.xFractal != 2 || ms.yFractal != 2)
        CreateMazePerfect(&bT);
      else
        TwoByTwoGenerate(bT);
      if (i > z)
        M0(bT, 0, (Rnd(1, ms.yFractal) << 1) - 1);
      else
        M1(bT, ms.xEntrance, ms.yEntrance);
      xT /= ms.xFractal; yT /= ms.yFractal;
      xs = (int)(x0 / xT % ms.xFractal); ys = (int)(y0 / yT % ms.yFractal);

      // The top and/or left edges of the section should be solid if the
      // section is adjacent to any surrounding Maze boundary, and that
      // surrounding Maze has a wall segment at the appropriate position.
      if (y0 % yT == 0)
        fX |= MG(bT, (xs << 1) + 1, ys << 1);
      if (x0 % xT == 0)
        fY |= MG(bT, xs << 1, (ys << 1) + 1);

      // Set a random number seed for this subsection. This is a crude way of
      // getting a different number for each cell based on the initial seed.
      for (j = ys * ms.xFractal + xs; j >= 0; j--)
        nRnd = Rnd(0, 16383);
      srand(nRnd);
    }

    // Create the fractal Maze section here.
    i = ms.nEntrancePos; ms.nEntrancePos = 3;
    FractalGenerate(b, bT, z, xp, yp);
    ms.nEntrancePos = i;
  } else {

    // Ensure the boundary walls at the right and bottom get drawn.
    fX = y0 == ysMax && FBetween(x0, 0, xsMax-1);
    fY = x0 == xsMax && FBetween(y0, 0, ysMax-1);
  }

  // Draw lines at the top and/or left of the section, if there should be no
  // passage between cells at this nesting level.
  if (fX)
    LineX(b, xp << 1, (xp + xpMax) << 1, yp << 1, fOn);
  if (fY)
    LineY(b, xp << 1, yp << 1, (yp + ypMax) << 1, fOn);

  // Draw the entrance or exit to the whole Maze if visible.
  if (y0 == 0 && x0 == xEntrance / xpMax)
    M0(b, ((xp + (int)LMod(xEntrance, xpMax)) << 1) + 1, yp << 1);
  if (y0 == ysMax && x0 == xExit / xpMax)
    M0(b, ((xp + (int)LMod(xExit, xpMax)) << 1) + 1, yp << 1);
}


// Create a grid of subsections of a virtual fractal Maze, where only part of
// the fractal Maze is rendered. Implements the FractalPart operation.

flag FractalPartial(bitmap *pb, long x0, long y0, int z, int nRndSeed,
  long *pxEntrance, long *pxExit)
{
  bitmap b, bT;
  char sz[cchSzMax];
  int xpMax = 1, ypMax = 1, x, y, i;
  long xT, yT;

  // Ensure the subsections of the fractal Maze to make will fit in a bitmap.
  for (i = 0; i < z; i++) {
    if ((long)xpMax * ms.xFractal * 8 > xBitmap ||
      (long)ypMax * ms.yFractal * 8 > yBitmap) {
      sprintf(S(sz),
        "%d by %d fractal section can't have nesting level greater than %d.",
        ms.xFractal, ms.yFractal, i); PrintSz_W(sz);
      return fFalse;
    }
    xpMax *= ms.xFractal; ypMax *= ms.yFractal;
  }

  // Ensure the total size of the fractal Maze doesn't overflow 32 bits.
  xT = xpMax; yT = ypMax;
  for (i = z; i < ms.zFractal; i++) {
    if (xT * ms.xFractal / ms.xFractal != xT ||
      yT * ms.yFractal / ms.yFractal != yT) {
      sprintf(S(sz),
        "%d by %d fractal Maze can't have total nesting greater than %d.",
        ms.xFractal, ms.yFractal, i); PrintSz_W(sz);
      return fFalse;
    }
    xT *= ms.xFractal; yT *= ms.yFractal;
  }

  // Create the bitmap for the fractal Maze, and a temporary internal bitmap
  // to store the Mazes at each level.
  if (!FBitmapSizeSet(pb, xpMax * 8, ypMax * 8))
    return fFalse;
  b = *pb;
  if (!FAllocateBitmap(&bT,
    (ms.xFractal << 1) + 1, (ms.yFractal << 1) + 1, b))
    return fFalse;
  BitmapOff(b);
  UpdateDisplay();
  SetXyh(bT);

  // Create a 4x4 grid of fractal Maze sections.
  for (y = 0; y < 4; y++)
    for (x = 0; x < 4; x++)
      FractalPartialGenerate(b, bT, z, x0 - 1 + x, y0 - 1 + y,
        x * xpMax, y * ypMax, nRndSeed, pxEntrance, pxExit);
  DeallocateBitmap(bT);
  return fTrue;
}


/*
******************************************************************************
** 3D Maze Routines
******************************************************************************
*/

CONST KV rgkvClar[DIRS] = {kvRed, kvBlue, kvGreen, kvYellow};


// Return the status of a pixel in a 3D Maze. Daedalus 3D Maze bitmaps don't
// have ceilings above the top level or floors below the bottom, so assume
// them here.

bit Qg(constcubemap t, int x, int y, int z)
{
  if (x < 0 || x >= TX(t) || y < 0 || y >= TY(t) ||
    z < -1 || z > Odd(TZ(t)))
    return fOff;
  if (z == -1 || z == Odd(TZ(t)))
    return fOn;
  x = Tx(t, x, z); y = Ty(t, y, z);
  return MGCore(t, x, y);
}


// Return the status of a pixel in a 3D Maze, where the 3D Maze is assumed to
// be in the perspective inside view. This is like Qg() except each floor
// extends forever throughout the plane. This prevents the Maze from appearing
// suspended in space, with just the top level sticking out of the ground.

bit Qg3(constcubemap t, int x, int y, int z)
{
  if (x < 0 || x >= TX(t) || y < 0 || y >= TY(t) ||
    z < 0 || z >= Odd(TZ(t)))
    return z & 1;
  x = Tx(t, x, z); y = Ty(t, y, z);
  return MGCore(t, x, y);
}


// Create a new perfect 3D Maze in the bitmap, stored as a sequence of 2D
// levels. The dimensions of the Maze are taken from the Bitmap Size dialog.

flag CreateMaze3D(bitmap *b)
{
  cubemap t;
  int tx, ty, tz, x, y, z, xnew, ynew, znew,
    fHunt = fFalse, pass = 0, d, i;
  long count;

  if (!FCubeSizeSet(b, TX(*b), TY(*b), TZ(*b), TW(*b)))
    return fFalse;
  t = *b;
  BitmapOff(t);
  if (!FEnsureMazeSize(&t, 3, femsNoResize))
    return fFalse;
  tx = Even(TX(t)); ty = Even(TY(t)); tz = Even(TZ(t));
  CubeBlock(t, 0, 0, 0, tx-2, ty-2, tz-2, fOn);
  MakeEntranceExit(t, 4);
  x = y = 1; z = 0;
  T0(t, x, y, z);
  count = (long)((tx - 1) >> 1) * ((ty - 1) >> 1) * (tz >> 1) - 1;
  UpdateDisplay();
  if (count <= 0)
    goto LDone;

  // Use the Hunt and Kill algorithm to create the 3D Maze.
  loop {
    if (ms.nRndBias > 0 && Rnd(0, ms.nRndBias) > 0)
      i = DIRS1;
    else
      i = DIRS3-1;
    d = Rnd(0, i);
    if (!TG(t, x, y, z)) {
      for (i = 0; i < DIRS3; i++) {
        xnew = x + (xoff3[d] << 1);
        ynew = y + (yoff3[d] << 1);
        znew = z + (zoff3[d] << 1);
        if (FLegalMaze3(t, xnew, ynew, znew) && TG(t, xnew, ynew, znew))
          break;
        d++;
        if (d >= DIRS3)
          d = 0;
      }
      if (ms.fRiver)
        fHunt = (i >= DIRS3);
      else {
        fHunt = i > 0;
        if (i < DIRS3)
          pass = 0;
      }
    }
    if (!fHunt) {
      if (fCellMax)
        break;
      T0(t, (x + xnew) >> 1, (y + ynew) >> 1, (z + znew) >> 1);
      T0(t, xnew, ynew, znew);
      x = xnew; y = ynew; z = znew;
      pass = 0;
      count--;
      if (count <= 0)
        goto LDone;
    } else {
      x += 2;
      if (x >= tx-1) {
        x = 1;
        y += 2;
        if (y >= ty-1) {
          y = 1;
          z += 2;
          if (z >= tz) {
            z = 0;
            pass++;
            if (pass > 1) {
              Assert(fFalse);
              goto LDone;
            }
            UpdateDisplay();
          }
        }
      }
    }
  }
LDone:
  return fTrue;
}


// Create a new perfect 3D Maze in the bitmap using the Binary Tree algorithm
// (technically a trinary tree in this case) by carving passages.

flag CreateMazeBinary3D(bitmap *b)
{
  cubemap t;
  int tx, ty, tz, x, y, z, d;

  if (!FCubeSizeSet(b, TX(*b), TY(*b), TZ(*b), TW(*b)))
    return fFalse;
  t = *b;
  BitmapOff(t);
  if (!FEnsureMazeSize(&t, 3, femsNoResize))
    return fFalse;
  tx = Even(TX(t)); ty = Even(TY(t)); tz = Even(TZ(t));
  CubeBlock(t, 0, 0, 0, tx-2, ty-2, tz-2, fOn);
  MakeEntranceExit(t, 4);
  UpdateDisplay();

  for (z = 0; z < tz; z += 2)
    for (y = 1; y < ty-1; y += 2)
      for (x = 1; x < tx-1; x += 2) {

        // For each cell, carve a passage north, west, or up.
        if (fCellMax)
          break;
        T0(t, x, y, z);
        if (x <= 1 && y <= 1 && z <= 0)
          continue;
        do {
          d = Rnd(0, 2);
        } while ((d == 0 && x <= 1) || (d == 1 && y <= 1) ||
          (d == 2 && z <= 0));
        if (d < 2)
          T0(t, x - (d ^ 1), y - d, z);
        else
          T0(t, x, y, z - 1);
      }
  return fTrue;
}


// Replace the 3D Maze in a bitmap with a clarified version of itself. Cells
// are wider and passages between levels are indicated with diagonal lines
// representing up and down staircases.

flag FClarify3D(bitmap *pb)
{
  bitmap t, b;
  int tw, tx, ty, tz, x, y, z, m, n;

  t = *pb;
  tw = Even(TW(t)); tx = Even(TX(t)); ty = Even(TY(t)); tz = Even(TZ(t));
  x = (tx * tw / 4 - 1) * CLAR + 1;
  y = (ty * ((tz + tw-1) / tw) / 2 - 1) * CLAR + 1;
  if (!FAllocateBitmap(&b, x, y, t))
    return fFalse;
  BitmapOff(b);
  for (z = 0; z < tz-1; z += 2)
    for (y = 0; y < ty-1; y += 2) {
      for (x = 0; x < tx-1; x += 2) {
        m = (z % tw * tx / 4 + x / 2) * CLAR;
        n = (z / tw * ty / 2 + y / 2) * CLAR;
        if (TG(t, x+1, y+1, z))
          Block(b, m, n, m+CLAR, n+CLAR, fOn);
        else {
          if (TG(t, x+1, y, z))
            LineX(b, m, m+CLAR, n, fOn);
          if (TG(t, x, y+1, z))
            LineY(b, m, n, n+CLAR, fOn);

          // Draw stairs between levels.
          if (x < tx-2 && y < ty-2) {
            if (!Qg(t, x+1, y+1, z-1))
              LineZ(b, m+3, n+CLAR-3, n+3, 1, fOn);
            if (!Qg(t, x+1, y+1, z+1))
              LineZ(b, m+CLAR-3, n+CLAR-3, n+3, -1, fOn);
          }
        }
      }
    }
  DeallocateBitmap(*pb);
  *pb = b;
  return fTrue;
}


// Like FClarify3D() but draws the clarified version of the Maze on a color
// bitmap. Colors are used to highlight staircases, where up and down
// staircases that connect with each other are given the same color, to make
// manually navigating between levels easier.

flag FColmapClarify3D(constcubemap t, colmap *pc, KV kv0, KV kv1)
{
  bitmap c;
  int tw, tx, ty, tz, x, y, z, m, n, i;

  tw = Even(TW(t)); tx = Even(TX(t)); ty = Even(TY(t)); tz = Even(TZ(t));
  x = (tx * tw / 4 - 1) * CLAR + 1;
  y = (ty * ((tz + tw-1) / tw) / 2 - 1) * CLAR + 1;
  if (!FColmapSizeSet(pc, x, y))
    return fFalse;
  c = *pc;
  ColmapSet(c, kv0);
  for (z = 0; z < tz-1; z += 2)
    for (y = 0; y < ty-1; y += 2) {
      for (x = 0; x < tx-1; x += 2) {
        m = (z % tw * tx / 4 + x / 2) * CLAR;
        n = (z / tw * ty / 2 + y / 2) * CLAR;
        if (TG(t, x+1, y+1, z))
          BlockC(c, m, n, m+CLAR, n+CLAR, kv1);
        else {
          if (TG(t, x+1, y, z))
            LineXC(c, m, m+CLAR, n, kv1);
          if (TG(t, x, y+1, z))
            LineYC(c, m, n, n+CLAR, kv1);

          // Draw colored stairs between levels.
          if (x < tx-2 && y < ty-2) {
            i = ((x ^ y) & 2) >> 1 ^ ((z & 2) >> 1) * 3 ^ 3;
            if (!Qg(t, x+1, y+1, z-1))
              LineZC(c, m+3, n+CLAR-3, n+3, 1, rgkvClar[i]);
            i ^= 3;
            if (!Qg(t, x+1, y+1, z+1)) {
              LineZC(c, m+CLAR-3, n+CLAR-3, n+3, -1, rgkvClar[i]);
              if (!Qg(t, x+1, y+1, z-1))
                CP(c, m+CLAR/2, n+CLAR/2,
                  KvBlend(rgkvClar[i], rgkvClar[i^3]));
            }
          }
        }
      }
    }
  return fTrue;
}


// Return the status of a pixel in a 4D Maze. Daedalus 4D Maze bitmaps don't
// have ceilings above the top levels or floors below the bottom, and don't
// have walls at the extreme ends of the 4th dimension, so assume them here.
// They also don't have solid blocks for levels that are between floors in
// both the 3rd and 4th dimensions (done to make the 4D Maze easier to
// understand in its raw form) so assume that here too.

bit Rg(constcubemap t, int w, int x, int y, int z)
{
  if (x < 0 || x >= TX(t) || y < 0 || y >= TY(t) ||
    z < -1 || z > Odd(TZ(t)) || w < -1 || w > Odd(TW(t)))
    return fOff;
  if (z == -1 || z == Odd(TZ(t)) || w == -1 || w == Odd(TW(t)) || (w & z & 1))
    return fOn;
  return UG(t, w, x, y, z);
}


// Create a new perfect 4D Maze in the bitmap, stored as a grid of 2D levels.
// The dimensions of the 4D Maze are taken from the Bitmap Size dialog.

flag CreateMaze4D(bitmap *b)
{
  cubemap t;
  int tw, tx, ty, tz, w, x, y, z, wnew, xnew, ynew, znew,
    fHunt = fFalse, pass = 0, d, i;
  long count;

  if (!FTesseractSizeSet(b, TW(*b), TX(*b), TY(*b), TZ(*b)))
    return fFalse;
  t = *b;
  BitmapOff(t);
  if (!FEnsureMazeSize(&t, 3, femsNoResize))
    return fFalse;
  tw = Even(TW(t)); tx = Even(TX(t)); ty = Even(TY(t)); tz = Even(TZ(t));
  for (z = 0; z < tz-1; z++) {
    y = Uy(t, 0, z);
    for (w = 0; w < tw-1; w++) {
      x = Ux(t, w, 0);
      if ((w & z & 1) == 0)
        Block(t, x, y, x + tx - 2, y + ty - 2, fOn);
    }
  }
  MakeEntranceExit(t, 5);
  x = y = 1; w = z = 0;
  U0(t, w, x, y, z);
  count = (long)(tw >> 1) * ((tx - 1) >> 1) * ((ty - 1) >> 1) * (tz >> 1) - 1;
  UpdateDisplay();
  if (count <= 0)
    goto LDone;

  // Use the Hunt and Kill algorithm to create the 4D Maze.
  loop {
    if (ms.nRndBias > 0 && Rnd(0, ms.nRndBias) > 0)
      i = DIRS1;
    else
      i = DIRS4-1;
    d = Rnd(0, i);
    if (!UG(t, w, x, y, z)) {
      for (i = 0; i < DIRS4; i++) {
        wnew = w + (woff3[d] << 1);
        xnew = x + (xoff3[d] << 1);
        ynew = y + (yoff3[d] << 1);
        znew = z + (zoff3[d] << 1);
        if (FLegalMaze4(t, wnew, xnew, ynew, znew) &&
          UG(t, wnew, xnew, ynew, znew))
          break;
        d = (d + 1) & (DIRS4 - 1);
      }
      if (ms.fRiver)
        fHunt = (i >= DIRS4);
      else {
        fHunt = i > 0;
        if (i < DIRS4)
          pass = 0;
      }
    }
    if (!fHunt) {
      if (fCellMax)
        break;
      U0(t,
        (w + wnew) >> 1, (x + xnew) >> 1, (y + ynew) >> 1, (z + znew) >> 1);
      U0(t, wnew, xnew, ynew, znew);
      w = wnew; x = xnew; y = ynew; z = znew;
      pass = 0;
      count--;
      if (count <= 0)
        goto LDone;
    } else {
      x += 2;
      if (x >= tx-1) {
        x = 1;
        y += 2;
        if (y >= ty-1) {
          y = 1;
          z += 2;
          if (z >= tz) {
            z = 0;
            w += 2;
            if (w >= tw) {
              w = 0;
              pass++;
              if (pass > 1) {
                Assert(fFalse);
                goto LDone;
              }
              UpdateDisplay();
            }
          }
        }
      }
    }
  }
LDone:
  return fTrue;
}


// Replace the 4D Maze in a bitmap with a clarified version of itself. Cells
// are wider and passages between levels and through the 4th dimension are
// indicated with lines in the cell, representing hatches or portals to an
// adjacent floor in the grid of levels.

flag FClarify4D(bitmap *pb)
{
  bitmap t, b;
  int w, x, y, z, m, n;

  t = *pb;
  x = (TW(t) / 2 * TX(t) / 2 - 1) * CLAR + 1;
  y = (TZ(t) / 2 * TY(t) / 2 - 1) * CLAR + 1;
  if (!FAllocateBitmap(&b, x, y, t))
    return fFalse;
  BitmapOff(b);
  for (w = 0; w < TW(t)-1; w += 2)
    for (z = 0; z < TZ(t)-1; z += 2)
      for (y = 0; y < TY(t)-1; y += 2) {
        n = (z * TY(t) / 4 + y / 2) * CLAR;
        for (x = 0; x < TX(t)-1; x += 2) {
          m = (w * TX(t) / 4 + x / 2) * CLAR;
          if (UG(t, w, x+1, y+1, z))
            Block(b, m, n, m+CLAR, n+CLAR, fOn);
          else {
            if (UG(t, w, x+1, y, z))
              LineX(b, m, m+CLAR, n, fOn);
            if (UG(t, w, x, y+1, z))
              LineY(b, m, n, n+CLAR, fOn);

            // Draw portals between levels in the 3rd and 4th dimensions.
            if (x < TX(t)-2 && y < TY(t)-2) {
              if (!Rg(t, w, x+1, y+1, z-1)) {
                LineX(b, m+3, m+CLAR-3, n+2, fOn);
                M1(b, m+CLAR/2, n+3);
              }
              if (!Rg(t, w, x+1, y+1, z+1)) {
                LineX(b, m+3, m+CLAR-3, n+CLAR-2, fOn);
                M1(b, m+CLAR/2, n+CLAR-3);
              }
              if (!Rg(t, w-1, x+1, y+1, z)) {
                LineY(b, m+2, n+3, n+CLAR-3, fOn);
                M1(b, m+3, n+CLAR/2);
              }
              if (!Rg(t, w+1, x+1, y+1, z)) {
                LineY(b, m+CLAR-2, n+3, n+CLAR-3, fOn);
                M1(b, m+CLAR-3, n+CLAR/2);
              }
            }
          }
        }
      }
  DeallocateBitmap(*pb);
  *pb = b;
  return fTrue;
}


// Like FClarify4D() but draws the clarified version of the Maze on a color
// bitmap. Colors are used to highlight portals between levels, where portals
// that connect with each other are given the same color, to make manually
// navigating among the grid of levels easier.

flag FColmapClarify4D(constcubemap t, colmap *pc, KV kv0, KV kv1)
{
  bitmap c;
  int w, x, y, z, m, n, i;

  x = (TW(t) / 2 * TX(t) / 2 - 1) * CLAR + 1;
  y = (TZ(t) / 2 * TY(t) / 2 - 1) * CLAR + 1;
  if (!FColmapSizeSet(pc, x, y))
    return fFalse;
  c = *pc;
  ColmapSet(c, kv0);
  for (w = 0; w < TW(t)-1; w += 2)
    for (z = 0; z < TZ(t)-1; z += 2)
      for (y = 0; y < TY(t)-1; y += 2) {
        n = (z * TY(t) / 4 + y / 2) * CLAR;
        for (x = 0; x < TX(t)-1; x += 2) {
          m = (w * TX(t) / 4 + x / 2) * CLAR;
          if (UG(t, w, x+1, y+1, z))
            BlockC(c, m, n, m+CLAR, n+CLAR, kv1);
          else {
            if (UG(t, w, x+1, y, z))
              LineXC(c, m, m+CLAR, n, kv1);
            if (UG(t, w, x, y+1, z))
              LineYC(c, m, n, n+CLAR, kv1);

            // Draw colored portals between levels in the 3rd/4th dimensions.
            if (x < TX(t)-2 && y < TY(t)-2) {
              i = ((y & 2) | ((x & 2) >> 1)) ^ ((w ^ z) & 2);
              if (!Rg(t, w, x+1, y+1, z-1)) {
                LineXC(c, m+3, m+CLAR-3, n+2, rgkvClar[i]);
                CP(c, m+CLAR/2, n+3, rgkvClar[i]);
              }
              DirInc(i);
              if (!Rg(t, w-1, x+1, y+1, z)) {
                LineYC(c, m+2, n+3, n+CLAR-3, rgkvClar[i]);
                CP(c, m+3, n+CLAR/2, rgkvClar[i]);
              }
              DirInc(i);
              if (!Rg(t, w, x+1, y+1, z+1)) {
                LineXC(c, m+3, m+CLAR-3, n+CLAR-2, rgkvClar[i]);
                CP(c, m+CLAR/2, n+CLAR-3, rgkvClar[i]);
              }
              DirInc(i);
              if (!Rg(t, w+1, x+1, y+1, z)) {
                LineYC(c, m+CLAR-2, n+3, n+CLAR-3, rgkvClar[i]);
                CP(c, m+CLAR-3, n+CLAR/2, rgkvClar[i]);
              }
            }
          }
        }
      }
  return fTrue;
}


// Convert a 4D Maze to a 3D Maze. The 4D Maze will be treated as a list of 3D
// Mazes placed along the X-axis. A copy of the Maze will be put in the given
// color bitmap, with portals through the 4th dimension between 3D bitmaps
// colored appropriately. The given monochrome bitmap will have on pixels at
// places where there are portals. Implements the 4D3D operation.

flag FConvert4DTo3D(cubemap *pt, bitmap *pb, colmap *pc,
  KV kvWall, KV kvMark, KV kvPast, KV kvFuture, KV kvBoth)
{
  cubemap t;
  bitmap b, bT;
  colmap c;
  int w, z;

  if (!FTesseractSizeSet(pt, TW(*pt), TX(*pt), TY(*pt), TZ(*pt)))
    return fFalse;
  t = *pt;
  if (!FBitmapSizeSet(pb, MX(t), MY(t)))
    return fFalse;
  if (!FColmapSizeSet(pc, MX(t), MY(t)))
    return fFalse;
  if (!FAllocateBitmap(&bT, MX(t), MY(t), t))
    return fFalse;
  b = *pb; c = *pc;
  BitmapOff(b);
  BitmapOff(bT);
  for (w = 0; w < TW(t); w += 2) {
    for (z = 0; z < TZ(t); z += 2) {

      // Copy 4D portals leading to the past to one bitmap.
      if (w > 0) {
        BlockMove(t, b, Ux(t, w-1, 0), Uy(t, 0, z),
          Ux(t, w-1, 0) + TX(t) - 2, Uy(t, 0, z) + TY(t) - 2,
          Ux(t, w, 0), Uy(t, 0, z));
        BlockReverse(b, Ux(t, w, 0), Uy(t, 0, z),
          Ux(t, w, 0) + TX(t) - 2, Uy(t, 0, z) + TY(t) - 2);
      }

      // Copy 4D portals leading to the future to another bitmap.
      if (w < TW(t)-2) {
        BlockMove(t, bT, Ux(t, w+1, 0), Uy(t, 0, z),
          Ux(t, w+1, 0) + TX(t) - 2, Uy(t, 0, z) + TY(t) - 2,
          Ux(t, w, 0), Uy(t, 0, z));
        BlockReverse(bT, Ux(t, w, 0), Uy(t, 0, z),
          Ux(t, w, 0) + TX(t) - 2, Uy(t, 0, z) + TY(t) - 2);
      }
    }
  }
  TX(t) *= Odd(TW(t)); TZ(t) = Odd(TZ(t)); TW(t) = 1;
  CopyClient(t, b);
  CopyClient(t, c);
  FColmapBlendFromBitmap(pc, b, bT, NULL);
  BitmapOr(b, bT);

  // Color portals through the 4th dimension appropriately.
  if (kvPast != kvRed)
    ColmapReplaceSimple(c, kvRed, kvPast);
  if (kvFuture != kvCyan)
    ColmapReplaceSimple(c, kvCyan, kvFuture);
  if (kvBoth != kvWhite)
    ColmapReplaceSimple(c, kvWhite, kvBoth);

  // Color walls and passages in the Maze appropriately.
  ColmapOrAndFromBitmap(c, t, kvBlack, kvWhite, 1);
  if (kvWall != kvWhite)
    ColmapReplaceSimple(c, kvWhite, kvWall);
  if (kvMark != kvBlack)
    ColmapReplaceSimple(c, kvBlack, kvMark);

  DeallocateBitmap(bT);
  return fTrue;
}


// Create a new Hypermaze in the bitmap, or a Maze where the solving object is
// a line instead of a point, and the solution a surface instead of a line.
// Like a 3D Maze this is a 3D construct stored as a sequence of 2D levels,
// where the dimensions of the Maze are taken from the Bitmap Size dialog.

flag CreateMazeHyper(bitmap *b)
{
  cubemap t;
  PT3 *rgpt;
  int tx, ty, tz, x, y, z, xnew, ynew, znew, d, i;
  long count, cpt = 1, ipt = 0, iptLo, iptHi;

  if (!FCubeSizeSet(b, TX(*b), TY(*b), TZ(*b), TW(*b)))
    return fFalse;
  t = *b;
  if (F64K3(t))
    return fFalse;
  BitmapOff(t);
  tx = Even(TX(t)); ty = Even(TY(t)); tz = Even(TZ(t));

  // Start with solid faces at the top and bottom levels.
  CubeBlock(t, 0, 0, 0,    tx-2, ty-2, 0,    fOn);
  CubeBlock(t, 0, 0, tz-2, tx-2, ty-2, tz-2, fOn);
  count = (long)(tx >> 1) * (ty >> 1) * (tz >> 1);
  if (count < 1)
    return fTrue;
  rgpt = RgAllocate(count, PT3);
  if (rgpt == NULL)
    return fFalse;
  for (y = 0; y < ty; y += 2)
    for (x = 0; x < tx; x += 2) {
      PutPt3(ipt, x, y, 0); PutPt3(ipt + 1, x, y, tz-2);
      ipt += 2;
      count -= 2;
    }
  cpt = ipt;
  ipt = RndL(0, cpt-1);
  GetPt3(ipt, x, y, z);
  UpdateDisplay();

  // Use a 3D version of the Growing Tree algorithm to create the Hypermaze.
  // Grow wall beams into the Maze from both the top and bottom faces.
  loop {
    d = Rnd(0, DIRS3-1);
    for (i = 0; i < DIRS3; i++) {
      xnew = x + (xoff3[d] << 1);
      ynew = y + (yoff3[d] << 1);
      znew = z + (zoff3[d] << 1);
      if (FLegalCube(t, xnew, ynew, znew) && !TG(t, xnew, ynew, znew))
        break;
      d++;
      if (d >= DIRS3)
        d = 0;
    }

    // When extending the Maze into a new cell, add the new cell to the list.
    // When nothing can be created from a cell, remove it from the list.
    if (i < DIRS3) {
      if (fCellMax)
        break;
      x = xnew; y = ynew; z = znew;
      T1(t, x - xoff3[d], y - yoff3[d], z - zoff3[d]);
      T1(t, x, y, z);
      count--;
      if (count <= 0)
        break;
      PutPt3(cpt, x, y, z);
      cpt++;
    } else {
      cpt--;
      if (cpt < 1) {
        Assert(fFalse);
        break;
      }
      rgpt[ipt] = rgpt[cpt];
    }

    // Figure out which cell to look at next.
    if (ms.fTreeRandom) {
      if (Rnd(0, ms.nTreeRiver) == 0)
        ipt = RndL(0, cpt-1);
      else
        ipt = cpt-1;
    } else {
      if (ms.nTreeRiver >= 0) {
        iptLo = cpt-1 - ms.nTreeRiver;
        if (iptLo < 0)
          iptLo = 0;
        iptHi = cpt-1;
      } else {
        iptHi = -(ms.nTreeRiver + 1);
        if (iptHi >= cpt)
          iptHi = cpt-1;
        iptLo = 0;
      }
      ipt = RndL(iptLo, iptHi);
    }
    GetPt3(ipt, x, y, z);
  }
  DeallocateP(rgpt);
  return fTrue;
}


/*
******************************************************************************
** N-Dimensional Maze Routines
******************************************************************************
*/

// Map a point within an n-Dimensional bitmap, whose coordinates are defined
// in an array, to X and Y pixel coordinates within a 2D bitmap. This is the
// internal representation of the ND bitmap, as opposed to how to display it.

void MapND(constcubemap t, CONST int *pn, int *px, int *py)
{
  int x = 0, y = 0, d, i;

  // X pixel coordinate is based on even numbered dimensions.
  d = 1;
  for (i = 0; i < TW(t); i += 2) {
    x += pn[i] * d;
    d *= i < 2 ? TX(t) : TZ(t);
  }

  // Y pixel coordinate is based on odd numbered dimensions.
  d = 1;
  for (i = 1; i < TW(t); i += 2) {
    y += pn[i] * d;
    d *= i < 2 ? TY(t) : TZ(t);
  }

  *px = x; *py = y;
}


// Create a new perfect n-Dimensional Maze in the bitmap, stored as a 2D grid
// of (n-2)-Dimensional Maze levels, which in turn contain 2D levels. The
// dimensions of the Maze are taken from the Bitmap Size dialog, where X and Y
// are the size of the 1st and 2nd dimensions, Z is 3rd dimension and up, and
// W is the number of dimensions. Implements the ND operation.

flag CreateMazeND(bitmap *b)
{
  bitmap t = *b;
  int n[dimMax], nNew[dimMax], nT[dimMax], nMax[dimMax],
    cDim = TW(t), tx, ty, tz, cDir, x, y, fHunt = fFalse, pass = 0, d, i, j;
  long count;

  x = TX(t) * (int)LPower(TZ(t), (cDim - 1) >> 1);
  y = TY(t) * (int)LPower(TZ(t), (cDim - 2) >> 1);
  if (cDim < 2 || cDim > dimMax || x < 1 || y < 1) {
    PrintSzNN_W("%d dimensional Maze of size %d is too large.", cDim, TZ(t));
    return fFalse;
  }
  if (!FBitmapSizeSet(b, x, y))
    return fFalse;
  t = *b;
  tx = Even(TX(t)); ty = Even(TY(t)); tz = Even(TZ(t));
  cDir = cDim << 1;
  nMax[0] = tx; nMax[1] = ty;
  n[0] = n[1] = 1;
  for (i = 2; i < cDim; i++) {
    nMax[i] = tz;
    n[i] = 0;
  }

  // Fill out the 2D levels that passages can actually carve through.
  BitmapOff(t);
  loop {
    j = 0;
    for (i = 2; i < cDim; i++)
      if (FOdd(n[i]))
        j++;
    if (j < 2) {
      MapND(t, n, &x, &y);
      Block(t, x - 1, y - 1, x + tx - 3, y + ty - 3, fOn);
    }
    for (i = 2; i < cDim; i++) {
      n[i]++;
      if (n[i] <= tz-2)
        break;
      n[i] = 0;
    }
    if (i >= cDim)
      break;
  }

  // Carve the entrance and initial cell.
  MapND(t, n, &x, &y);
  M0(t, x, y);
  n[1]--;
  MapND(t, n, &x, &y);
  M0(t, x, y);
  n[1]++;
  count = (long)((tx - 1) >> 1) * ((ty - 1) >> 1) *
    LPower(tz >> 1, cDim-2) - 1;
  UpdateDisplay();
  if (count <= 0)
    goto LDone;

  // Use the Hunt and Kill algorithm to create the n-Dimensional Maze.
  loop {
    if (ms.nRndBias > 0 && Rnd(0, ms.nRndBias) > 0)
      i = DIRS1;
    else
      i = cDir - 1;
    d = Rnd(0, i);
    MapND(t, n, &x, &y);
    if (!MG(t, x, y)) {
      for (i = 0; i < cDir; i++) {
        for (j = 0; j < cDim; j++)
          nNew[j] = n[j];
        j = d >> 1;
        nNew[j] += (FOdd(d) << 2) - 2;
        if (!(nNew[j] < 0 || nNew[j] > nMax[j]-2)) {
          MapND(t, nNew, &x, &y);
          if (MG(t, x, y))
            break;
        }
        d = (d + 1) % cDir;
      }
      if (ms.fRiver)
        fHunt = (i >= cDir);
      else {
        fHunt = i > 0;
        if (i < cDir)
          pass = 0;
      }
    }
    if (!fHunt) {
      for (i = 0; i < cDim; i++)
        nT[i] = (n[i] + nNew[i]) >> 1;
      MapND(t, nT, &x, &y);
      M0(t, x, y);
      MapND(t, nNew, &x, &y);
      M0(t, x, y);
      for (i = 0; i < cDim; i++)
        n[i] = nNew[i];
      pass = 0;
      count--;
      if (count <= 0)
        goto LDone;
    } else {
      for (i = 0; i < cDim; i++) {
        n[i] += 2;
        if (n[i] <= nMax[i]-2)
          break;
        n[i] = i < 2;
      }
      if (i >= cDim) {
        pass++;
        if (pass > 1) {
          Assert(fFalse);
          goto LDone;
        }
        UpdateDisplay();
      }
    }
  }
LDone:
  for (i = 0; i < cDim; i++)
    nMax[i] -= 2 + (i < 1);
  MapND(t, nMax, &x, &y);
  M0(t, x, y);
  return fTrue;
}


// Offset the coordinates of a pixel, based on the axis sizes of an
// n-Dimensional cube, so the cube when rendered will be centered around the
// origin, as opposed to the origin always being one of the cube's corners.

void CenterND(constcubemap t, CONST long *nLine, int *px, int *py)
{
  int x = 0, y = 0, i;

  for (i = 0; i < TW(t); i++) {
    x += (int)nLine[i << 1];
    y += (int)nLine[(i << 1) + 1];
  }
  *px -= x >> 1; *py -= y >> 1;
}


// Compute the pixel coordinate of a point within an n-Dimensional cube, based
// on the horizontal and vertical offsets of each dimensional axis. This is
// for the screen display of the cube, as opposed to internal representation.

void MapNDRender(constcubemap t, int *px, int *py, CONST int *n,
  CONST int *nMax, CONST long *nLine)
{
  int x = 0, y = 0, cDim = TW(t), i;

  for (i = 0; i < cDim; i++) {
    x += NMultDiv(nLine[i << 1],       n[i], nMax[i]-1);
    y += NMultDiv(nLine[(i << 1) + 1], n[i], nMax[i]-1);
  }
  *px = x; *py = y;
}


// Draw a n-Dimensional Maze on the bitmap. Implements the DrawND operation.

void RenderND(bitmap b, colmap c, constcubemap t, int *px, int *py, int *px2,
  int *py2, CONST long *nCoor, CONST long *nLine, KV kv, int grf)
{
  int n[dimMax], n2[dimMax], nMax[dimMax], x0 = *px, y0 = *py,
    cDim = TW(t), tx, ty, tz, x, y, x1, y1, x2, y2, xmax, ymax, i, j, k;
  KV rgkv[dimMax];
  flag fColor = (b == NULL), o;

  if (!fColor) {
    o = (kv != kvBlack);
    xmax = MX(b); ymax = MY(b);
  } else {
    for (i = 0; i < cDim; i++)
      rgkv[i] = Hue(i * nHueMax / cDim);
    xmax = CX(c); ymax = CY(c);
  }
  tx = (TX(t) - 1) >> 1; ty = (TY(t) - 1) >> 1; tz = (TZ(t) + 1) >> 1;
  nMax[0] = tx; nMax[1] = ty;
  for (i = 2; i < cDim; i++)
    nMax[i] = tz;
  for (i = 0; i < cDim; i++)
    n[i] = 0;
  if (grf & 1)
    CenterND(t, nLine, &x0, &y0);

  loop {
    for (i = 0; i < cDim; i++)
      n2[i] = (i < 2) + (n[i] << 1);
    MapND(t, n2, &x1, &y1);
    if (!MG(t, x1, y1)) {
      MapNDRender(t, &x1, &y1, n, nMax, nLine);
      x1 += x0; y1 += y0;
      if (!fColor && (grf & 2) > 0 &&
        x1 > -1 && x1 <= xmax && y1 > -1 && y1 <= ymax)
        Block(b, x1-1, y1-1, x1+1, y1+1, o);

      // For each cell, draw a passage if present to cells adjacent to it.
      for (i = 0; i < cDim; i++)
        if (n[i] < nMax[i]-1) {
          n[i]++;
          for (j = 0; j < cDim; j++)
            n2[j] = (j < 2) + (n[j] << 1) - (j == i);
          MapND(t, n2, &x, &y);
          if (!MG(t, x, y)) {
            MapNDRender(t, &x2, &y2, n, nMax, nLine);
            x2 += x0; y2 += y0;
            if (!fColor)
              Line(b, x1, y1, x2, y2, o);
            else {
              if ((grf & 4) == 0)
                k = i;
              else {
                k = (n[i] <= nCoor[i]) - 1;
                for (j = 0; j < cDim; j++)
                  k += NAbs(n[j] - nCoor[j]);
                if (k >= cDim)
                  k = cDim - 1;
              }
              LineC(c, x1, y1, x2, y2, rgkv[k]);
            }
          }
          n[i]--;
        }
    }

    // Advance to the next cell in the Maze.
    for (i = 0; i < cDim; i++) {
      n[i]++;
      if (n[i] < nMax[i])
        break;
      n[i] = 0;
    }
    if (i >= cDim)
      break;
  }

  // Draw the vertex points in a color Maze afterward, so they appear on top.
  if (fColor && (grf & 2) > 0)
    loop {
      for (i = 0; i < cDim; i++)
        n2[i] = (i < 2) + (n[i] << 1);
      MapND(t, n2, &x1, &y1);
      if (!MG(t, x1, y1)) {
        MapNDRender(t, &x1, &y1, n, nMax, nLine);
        x1 += x0; y1 += y0;
        if (x1 > -1 && x1 <= xmax && y1 > -1 && y1 <= ymax)
          BlockC(c, x1-1, y1-1, x1+1, y1+1, kv);
      }

      // Advance to the next cell in the Maze.
      for (i = 0; i < cDim; i++) {
        n[i]++;
        if (n[i] < nMax[i])
          break;
        n[i] = 0;
      }
      if (i >= cDim)
        break;
    }

  // Set x and y to screen coordinates of the current coordinate in the Maze.
  for (i = 0; i < cDim; i++)
    n[i] = (int)nCoor[i];
  MapNDRender(t, px, py, n, nMax, nLine);
  *px += x0; *py += y0;

  // Set x2 and y2 to the screen coordinates of the finish cell of the Maze.
  for (i = 0; i < cDim; i++)
    n[i] = nMax[i] - 1;
  MapNDRender(t, px2, py2, n, nMax, nLine);
  *px2 += x0; *py2 += y0;
}


// Move a ND point to an adjacent point, based on screen coordinates where the
// user clicked. Go to the adjacent point in the ND cube closest to that
// clicked coordinate (or do nothing if the closest ND point is the current
// point). Return whether the point moved to the far corner, which is
// considered to be the solution cell. Implements the MoveND operation.

flag MoveND(constcubemap t, long *nCoor, CONST long *nLine, int *px, int *py,
  flag fCenter)
{
  int nMax[dimMax], n[dimMax], np[dimMax], cDim = TW(t), tx, ty, tz,
    x0, y0, xMouse = *px, yMouse = *py, iNear = -1, i, ip, j, f;
  long lDist, lDistMin, x, y;

  tx = (TX(t) - 1) >> 1; ty = (TY(t) - 1) >> 1; tz = (TZ(t) + 1) >> 1;
  nMax[0] = tx; nMax[1] = ty;
  for (i = 2; i < cDim; i++)
    nMax[i] = tz;
  for (i = 0; i < cDim; i++)
    n[i] = (int)nCoor[i];
  MapNDRender(t, &x0, &y0, n, nMax, nLine);
  if (fCenter)
    CenterND(t, nLine, &x0, &y0);
  xMouse -= x0; yMouse -= y0;

  lDistMin = Sq(xMouse) + Sq(yMouse);
  for (i = 0; i < cDim << 1; i++) {
    j = i >> 1;
    f = FOdd(i) ? n[j] < nMax[j]-1 : n[j] > 0;
    if (f) {
      for (ip = 0; ip < cDim; ip++)
        np[ip] = (ip < 2) + (n[ip] << 1);
      np[j] += (FOdd(i) << 1) - 1;
      MapND(t, np, &x0, &y0);
      if (MG(t, x0, y0))
        f = fFalse;
    }
    if (f) {
      x = nLine[j << 1]       / (nMax[j] - 1);
      y = nLine[(j << 1) + 1] / (nMax[j] - 1);
      if (!FOdd(i)) {
        neg(x); neg(y);
      }
      lDist = Sq(xMouse - x) + Sq(yMouse - y);
      if (lDist < lDistMin) {
        lDistMin = lDist;
        iNear = i;
      }
    }
  }

  if (iNear >= 0) {
    n[iNear >> 1] += (FOdd(iNear) << 1) - 1;
  }
  f = fTrue;
  for (i = 0; i < cDim; i++) {
    nCoor[i] = n[i];
    if (n[i] < nMax[i] - 1)
      f = fFalse;
  }
  MapNDRender(t, px, py, n, nMax, nLine);
  if (fCenter)
    CenterND(t, nLine, px, py);
  return f;
}


/*
******************************************************************************
** Planair Maze Routines
******************************************************************************
*/

#define PZTo(sz, z, d) (int)(sz[(z)*9 + (d)*2]-'a')
#define PDir(sz, z, d) (int)(sz[(z)*9 + (d)*2 + 1]-'0')
#define PSpc(sz, z) sz[(z)*9 + 8]

#define JX(x, z) (((z) % jw) * TX(b) + (x))
#define JY(y, z) (((z) / jw) * TY(b) + (y))
#define JG(b, x, y, z) MG(b, JX(x, z), JY(y, z))
#define J0(b, x, y, z) M0(b, JX(x, z), JY(y, z))

// Check whether a planair Maze definition string is formatted correctly.

flag FValidPlanair(constbitmap b, CONST char *sz)
{
  char szT[cchSzMax], sz1[cchSzDef], sz2[cchSzDef], sz3[cchSzDef];
  int cch, jz, i, j, z, d, z1, z2;

  // Ensure the length of the string implies an integer number of sections.
  cch = CchSz(sz);
  if ((cch+1) % 9 > 0) {
    sprintf(S(sz1), "The Planair string length is invalid.\n");
    sprintf(S(sz2), "The length %d implies a section count of %.1f.\n",
      cch, (real)(cch + 1)/9.0);
    sprintf(S(sz3),
      "The length needs to be one less than a multiple of 9.\n");
    sprintf(S(szT), "%s%s%s", sz1, sz2, sz3);
    PrintSz_W(szT);
    return fFalse;
  }

  // Check each section for validity.
  jz = (cch+1) / 9;
  for (i = 0; i < jz; i++) {
    for (j = 0; j < DIRS; j++) {
      z = PZTo(sz, i, j);
      if (!FBetween(z, 0, jz-1)) {
        sprintf(S(sz1),
          "The Planair string has an invalid group destination.\n");
        sprintf(S(sz2), "Edge %d of group '%c' points to group '%c'.\n",
          j, (char)(i+'a'), (char)(z+'a'));
        sprintf(S(sz3),
          "This destination group should range from 'a' to '%c'.\n",
          (char)(jz-1+'a'));
        sprintf(S(szT), "%s%s%s", sz1, sz2, sz3);
        PrintSz_W(szT);
        return fFalse;
      }
      d = PDir(sz, i, j);
      if (!FBetween(d, 0, 7)) {
        sprintf(S(sz1),
          "The Planair string has an invalid edge destination.\n");
        sprintf(S(sz2), "Edge %d of group '%c' points to edge %d.\n",
          j, (char)(i+'a'), d);
        sprintf(S(sz3), "This destination edge should range from 0 to 7.\n",
          (char)(jz+'a'));
        sprintf(S(szT), "%s%s%s", sz1, sz2, sz3);
        PrintSz_W(szT);
        return fFalse;
      }
      z1 = ((FOdd(j) ? TY(b) : TX(b)) - 1) >> 1;
      z2 = ((FOdd(d) ? TY(b) : TX(b)) - 1) >> 1;
      if (z1 != z2) {
        sprintf(S(sz1), "The Planair string has an unaligned edge link.\n");
        sprintf(S(sz2),
          "Edge %d of group '%c' points to edge %d of group '%c'.\n",
          j, (char)(i+'a'), d, (char)(z+'a'));
        sprintf(S(sz3), "These edges are of different lengths (%d and %d).\n",
          z1, z2);
        sprintf(S(szT), "%s%s%s", sz1, sz2, sz3);
        PrintSz_W(szT);
        return fFalse;
      }
    }
    if (i < jz-1 && PSpc(sz, i) != ' ') {
      sprintf(S(sz1), "The Planair string has an invalid separator.\n");
      sprintf(S(sz2), "The character after group '%c' should be a space.\n",
        (char)(i+'a'));
      sprintf(S(szT), "%s%s", sz1, sz2);
      PrintSz_W(szT);
      return fFalse;
    }
  }

  // Check for asymmetric edge connections. If one edge connects to another,
  // the other edge should connect up with the first one to avoid one way
  // passages. It's not required, but at least display a warning if so.
  for (i = 0; i < jz; i++)
    for (j = 0; j < DIRS; j++) {
      z = PZTo(sz, i, j); d = PDir(sz, i, j);
      z1 = PZTo(sz, z, d & DIRS1); z2 = PDir(sz, z, d & DIRS1);
      if (z1 != i) {
        sprintf(S(sz1), "The Planair string has a one way group link.\n");
        sprintf(S(sz2),
          "Edge %d of group '%c' points to edge %d of group '%c',\n",
          j, (char)(i+'a'), d, (char)(z+'a'));
        sprintf(S(sz3),
          "However edge %d of group '%c' points to group '%c'.\n",
          d, (char)(z+'a'), (char)(z1+'a'));
        sprintf(S(szT), "%s%s%s", sz1, sz2, sz3);
        PrintSz(szT);
        return fTrue;
      }
      if ((z2 & DIRS1) != j || (d & DIRS) != (z2 & DIRS)) {
        sprintf(S(sz1), "The Planair string has a one way edge link.\n");
        sprintf(S(sz2),
          "Edge %d of group '%c' points to edge %d of group '%c',\n",
          j, (char)(i+'a'), d, (char)(z+'a'));
        sprintf(S(sz3),
          "However edge %d of group '%c' points to edge %d.\n",
          d, (char)(z+'a'), z2);
        sprintf(S(szT), "%s%s%s", sz1, sz2, sz3);
        PrintSz(szT);
        return fTrue;
      }
    }
  return fTrue;
}


// Create a new planair Maze in a bitmap, or a Maze on the surface of a set of
// rectangles, whose edges can connect in arbitrary ways. Like a 3D Maze this
// is technically a 3D construct stored as a sequence of levels, where the
// horizontal and vertical dimensions of each rectangle are taken from the
// 3D settings in the Bitmap Size dialog.

flag CreateMazePlanair(bitmap *pb)
{
  bitmap b;
  int jx = Even(TX(*pb)), jy = Even(TY(*pb)), jz, jw = TW(*pb),
    fHunt = fFalse, pass = 0, x, y, z, xnew, ynew, znew,
    xInc = 2, yInc = 2, zInc = 1, d, e, i, iMax, j;
  long count;

  b = *pb;
  if (!FValidPlanair(b, ms.szPlanair))
    return fFalse;
  jz = (CchSz(ms.szPlanair) + 1) / 9;
  if (!FBitmapSizeSet(pb, TX(b) * Min(jz, jw), TY(b) * ((jz + jw-1) / jw)))
    return fFalse;
  b = *pb;
  BitmapOff(b);
  if (!FEnsureMazeSize(&b, 3, femsNoResize))
    return fFalse;
  CubeBlock(b, 0, 0, 0, jx-2, jy-2, jz-1, fOn);
  count = (long)((jx - 1) >> 1) * ((jy - 1) >> 1) * jz - 1;
  iMax = ms.fRiver ? DIRS : 1;
  x = y = 1; z = 0;
  J0(b, x, y, z);
  UpdateDisplay();
  if (count <= 0)
    goto LDone;

  // Use the Hunt and Kill algorithm to create the planair Maze.
  loop {
    if (!JG(b, x, y, z)) {
LNext:
      d = RndDir();
      for (i = fHunt ? DIRS : iMax; i > 0; i--) {
        xnew = x + xoff2[d];
        ynew = y + yoff2[d];
        if (xnew <= 0 || xnew >= jx-1 || ynew <= 0 || ynew >= jy-1) {

          // If checking a cell off the edge of a rectangle, figure out which
          // new rectangle and which edge cell to flow onto.
          znew = PZTo(ms.szPlanair, z, d);
          e = PDir(ms.szPlanair, z, d);
          j = xnew <= 0 || xnew >= jx-1 ? ynew : xnew;
          switch (e) {
          case 0: xnew = j;    ynew = 1;    break;
          case 1: xnew = 1;    ynew = j;    break;
          case 2: xnew = j;    ynew = jy-3; break;
          case 3: xnew = jx-3; ynew = j;    break;
          case 4: xnew = jx-2-j; ynew = 1;      break;
          case 5: xnew = 1;      ynew = jy-2-j; break;
          case 6: xnew = jx-2-j; ynew = jy-3;   break;
          case 7: xnew = jx-3;   ynew = jy-2-j; break;
          }
          e &= DIRS1;
        } else {

          // For movement within a rectangle, the rectangle of course doesn't
          // change, and the undo direction is simply the reverse.
          znew = z;
          e = d ^ 2;
        }
        if (JG(b, xnew, ynew, znew)) {
          if (fCellMax)
            break;
          J0(b, x+xoff[d], y+yoff[d], z);
          J0(b, xnew+xoff[e], ynew+yoff[e], znew);
          J0(b, xnew, ynew, znew);
          x = xnew; y = ynew; z = znew;
          pass = 0;
          fHunt = fFalse;
          count--;
          if (count <= 0)
            goto LDone;
          goto LNext;
        }
        DirInc(d);
      }
      fHunt = fTrue;
    }
    if (x + xInc > 0 && x + xInc < jx-1)
      x += xInc;
    else {
      neg(xInc);
      if (y + yInc > 0 && y + yInc < jy-1)
        y += yInc;
      else {
        neg(yInc);
        if (z + zInc >= 0 && z + zInc < jz)
          z += zInc;
        else {
          neg(zInc);
          pass++;
          if (pass > 1) {
            Assert(fFalse);
            goto LDone;
          }
          UpdateDisplay();
        }
      }
    }
  }
LDone:
  return fTrue;
}


/*
******************************************************************************
** Segment Maze Routines
******************************************************************************
*/

#define SList 16
#define SGroup 9
#define SEGRIVER 1  // Op #1: River percentage.
#define SEGBIAS 2   // Op #2: Random bias.
#define SEGRUN 3    // Op #3: Random run.

int segment[SList*SGroup] = {
  SEGRIVER,   0,  0, 255,  95,   0,   0,   0,   0,
  SEGRIVER,   0, 96, 255, 191, 100, 100, 100, 100,
  SEGBIAS,    0,  0, 127, 191, -20, -20, -20, -20,
  SEGBIAS,  128,  0, 255, 191,  20,  20,  20,  20,
  SEGRUN,    64, 48, 191, 144,   4,   4,   4,   4,
  0};

#define SOp(i) segment[(i)*SGroup]
#define SX1(i) segment[(i)*SGroup + 1]
#define SY1(i) segment[(i)*SGroup + 2]
#define SX2(i) segment[(i)*SGroup + 3]
#define SY2(i) segment[(i)*SGroup + 4]
#define SVal1(i) segment[(i)*SGroup + 5]
#define SVal2(i) segment[(i)*SGroup + 6]
#define SVal3(i) segment[(i)*SGroup + 7]
#define SVal4(i) segment[(i)*SGroup + 8]

// Convert the segment Maze definition array to a list of numbers in a string.
// Called from the Create Settings dialog to show the current definition.

void SegmentGetSz(char *sz, int cchMax)
{
  int i, j;
  char *pch = sz;

  for (i = 0; i < SList && SOp(i); i++) {
    for (j = 0; j < SGroup; j++) {
      sprintf(SS(pch, cchMax - (pch - sz)), "%d%s",
        segment[i*SGroup + j], j < 8 ? " " : "  ");
      while (*pch)
        pch++;
    }
  }
  sprintf(SS(pch, cchMax - (pch - sz)), "0");
}


// Given a list of numbers in a string, fill the segment Maze definition array
// with them. The reverse of SegmentGetSz().

void SegmentParseSz(CONST char *sz)
{
  int i, j;
  CONST char *pch = sz;
  for (i = 0; i < SList*SGroup && *pch; i++) {
    SkipSpaces(pch);
    sscanf(pch, "%d", &j); segment[i] = j;
    SkipToSpace(pch);
  }
}


// Given a coordinate and an operation, return the first segment rectangle for
// that operation that overlaps the coordinate, if any.

int SegLookup(int x, int y, int op)
{
  int i;

  for (i = 0; i < SList && SOp(i) != 0; i++) {
    if (NAbs(SOp(i)) == op &&
      x >= SX1(i) && x <= SX2(i) && y >= SY1(i) && y <= SY2(i))
      return i;
  }
  return -1;
}


// Return the correct operation value for a coordinate that lies within a
// segment rectangle. A rectangle may have different values blend across it.

int SegValBlend(int x, int y, int i)
{
  int s1, s2, s3, s4, x1, x2;
  real r;

  // Simple case: The value is uniform across the rectangle.
  s1 = SVal1(i); s2 = SVal2(i); s3 = SVal3(i); s4 = SVal4(i);
  if (s1 == s2 && s2 == s3 && s3 == s4)
    return s1;

  // Complex case: The value has a gradient between the four corners.
  r = ((real)(y - SY1(i)) / (real)(SY2(i) - SY1(i)));
  x1 = s1 + (int)(r * (real)(s3 - s1));
  x2 = s2 + (int)(r * (real)(s4 - s2));
  r = ((real)(x - SX1(i)) / (real)(SX2(i) - SX1(i)));
  return x1 + (int)(r * (real)(x2 - x1));
}


// Return a random one of the four directions. A more advanced version of
// RndDir() that gets the random bias and run from any segment rectangles that
// overlap the coordinates.

int SegRndDir(int x, int y)
{
  int i, n;

  // If in the middle of a random run, return the previous direction again.
  if (ms.cRunRndSeg > 0) {
    ms.cRunRndSeg--;
    return ms.dirRndSeg;
  }

  // Randomly pick a new direction, and a new random run length if any.
  i = SegLookup(x, y, SEGRUN);
  if (i >= 0)
    ms.cRunRndSeg = SegValBlend(x, y, i);
  else
    ms.cRunRndSeg = ms.nRndRun;
  i = SegLookup(x, y, SEGBIAS);
  if (i >= 0)
    n = SegValBlend(x, y, i);
  else
    n = ms.nRndBias;
  ms.dirRndSeg = Rnd(0, DIRS1 + NAbs(n)*2);

  // Return a standard direction.
  if (ms.dirRndSeg < DIRS)
    return ms.dirRndSeg;

  // Higher random numbers represent bias. Map them to a standard direction.
  ms.dirRndSeg = ((ms.dirRndSeg & 1) << 1) + (n > 0);
  return ms.dirRndSeg;
}


// Return whether to check all directions (as opposed to just the first)
// before entering hunting mode while creating the Maze. This is a random
// percentage taken from a segment rectangle that overlaps the coordinates.
// Like a version of the Create With River setting that changes dynamically.

flag SegRiver(int x, int y)
{
  int i, n;

  i = SegLookup(x, y, SEGRIVER);
  if (i > -1) {
    n = SegValBlend(x, y, i);
    return Rnd(0, 99) < n;
  }
  return ms.fRiver;
}


// Create a new segment Maze in the bitmap, or a standard Maze that can have
// different textures defined for different rectangular sections in it.

flag CreateMazeSegment(bitmap *pb)
{
  bitmap b;
  int x, y, xnew, ynew,
    xInc = 2, yInc = 2, fHunt = fFalse, pass = 0, d, i;
  long count;

  if (!FEnsureMazeSize(pb, 3, femsOddSize | femsNoResize | femsMinSize))
    return fFalse;
  b = *pb;
  ms.cRunRndSeg = 0;
  MazeClear(b, fOn);
  count = (long)((xh - xl) >> 1) * ((yh - yl) >> 1) - 1;
  x = xh - 1; y = yh - 1;
  M0(b, x, y);
  MakeEntranceExit(b, 0);
  UpdateDisplay();

  // Use the Hunt and Kill algorithm to create the segmented Maze.
  loop {
    if (!MG(b, x, y)) {
      d = SegRndDir(x, y);
      for (i = 0; i < DIRS; i++) {
        xnew = x + xoff2[d];
        ynew = y + yoff2[d];
        if (FLegalMaze2(xnew, ynew) && MG(b, xnew, ynew))
          break;
        DirInc(d);
      }
      if (SegRiver(xnew, ynew))
        fHunt = (i >= DIRS);
      else {
        fHunt = i > 0;
        if (i < DIRS)
          pass = 0;
      }
    }
    if (!fHunt) {
      if (fCellMax)
        break;
      M0(b, (x + xnew) >> 1, (y + ynew) >> 1);
      M0(b, xnew, ynew);
      x = xnew; y = ynew;
      pass = 0;
      count--;
      if (count <= 0)
        goto LDone;
    } else {
      if (x + xInc >= xl && x + xInc <= xh)
        x += xInc;
      else {
        neg(xInc);
        if (y + yInc >= yl && y + yInc <= yh)
          y += yInc;
        else {
          neg(yInc);
          pass++;
          if (pass > 1)
            goto LDone;
          UpdateDisplay();
        }
      }
    }
  }
LDone:
  return fTrue;
}


/*
******************************************************************************
** Infinite Length Maze Routines
******************************************************************************
*/

#define INFSTART 0
#define INFCONTINUE 1
#define INFEND 2
#define INFRESTART 3

// Create or open a file in preparation of appending infinite Maze sections to
// it. Used when the Save Infinite To File setting is on.

FILE *FileOpenInfinite(CONST char *szType)
{
  char sz[cchSzDef];
  FILE *file;

  // If writing to multiple files, include numbers in the filename.
  if (ms.nInfFileCutoff > 0)
    sprintf(S(sz), "inf%05d.txt", ms.nInfGen / ms.nInfFileCutoff);
  else
    sprintf(S(sz), "infinite.txt");
  file = FileOpen(sz, szType);
  return file;
}


// Create an infinite Maze, or a Maze that has a finite width and no limit to
// how long it can be. The Maze is created in sections, which scroll down the
// bitmap, where two sections are visible in the bitmap at any time. This can
// start a new infinite Maze in the bitmap, add onto the existing infinite
// Maze by scrolling up and adding a new Maze section, or finish the existing
// infinite Maze in the bitmap.

flag CreateMazeInfinite(bitmap *pb, int phase)
{
  int x, y, xnew, ynew, yhalf, xInc = 2, yInc = 2, fHunt, pass = 0, d,
    xs, xinc1, xinc2, yp;
  long count, count2;
  bitmap b, b2;

  x = MX(*pb);
  if (x < 3)
    x = 3;
  y = MY(*pb);
  if (y < 7)
    y = 7;
  else
    y -= 2*((y + 1 & 2) > 0);
  if (!FBitmapSizeSet(pb, x, y))
    return fFalse;
  b = *pb;
  SetXyh(b);

  // Automatically start a new infinite Maze if not in the middle of one.
  // Automatically end the current infinite Maze if the number of sections
  // created so far reaches the Generation Cutoff setting.
  if (!ms.fInfMiddle) {
    if (phase != INFSTART && ms.fInfAutoStart)
      return fFalse;
    phase = INFSTART;
  } else {
    if (ms.nInfCutoff > 0 && ms.nInfGen >= ms.nInfCutoff - 1)
      phase = INFEND;
  }

  if (ms.fInfEller) {
    xs = (xh - xl) >> 1;
    x = xs-1;              xinc1 = (int)LPrime(x);
    x = xs-2*ms.fTreeWall; xinc2 = (int)LPrime(x);

    // When using Eller's algorithm, initialize the connection information for
    // the current row, and put each cell in the row in its own set. Do this
    // when starting, restarting (for lack of anything better to do although
    // it will cause loops) and if the user resizes the bitmap.

    if (ms.cnInfEller != xs || phase == INFSTART || phase == INFRESTART) {
      if (ms.cnInfEller != xs) {
        if (ms.rgnInfEller != NULL)
          DeallocateP(ms.rgnInfEller);
        ms.rgnInfEller = RgAllocate(xs*2, short);
        if (ms.rgnInfEller == NULL)
          return fFalse;
        ms.cnInfEller = xs;
      }
      for (x = 0; x < xs; x++)
        ms.rgnInfEller[x] = ms.rgnInfEller[xs + x] = x;
    }
  }

  // Restarting an infinite Maze is the same as continuing it, except the
  // number of cells created so far in the bottom section is recomputed.
  if (phase == INFRESTART) {
    if (!ms.fInfEller) {
      ms.fInfMiddle = fTrue;
      ms.nInfCount = 0L;
      for (y = (yh >> 1) + 2; y < yh; y += 2)
        for (x = 1; x < xh; x += 2)
          ms.nInfCount += !MG(b, x, y);
    }
    phase = INFCONTINUE;
  }

  x = y = 1;
  if (phase == INFSTART) {

    // Get ready to start a new infinite Maze.
    if (ms.fileInf != NULL) {
      fclose(ms.fileInf);
      ms.fileInf = NULL;
    }
    ms.nInfGen = ms.nInfCount = 0L;
    if (ms.fInfFile) {
      ms.fileInf = FileOpenInfinite("w");
      if (ms.fileInf == NULL)
        PrintSz_E("The file could not be created.");
    }
    MazeClear(b, fOn);
    if (FOdd(xh))
      LineY(b, xh, yl, yh, fOff);
    if (FOdd(yh))
      LineX(b, xl, xh, yh, fOff);
    MakeEntranceExit(b, 6);
    if (!ms.fInfEller) {
      M0(b, x, y);
      fHunt = fFalse;
    } else
      yp = 1;
  } else {

    // Get ready to continue or finish an existing infinite Maze.
    if (ms.fInfFile) {
      if (ms.nInfFileCutoff > 0 && ms.nInfGen % ms.nInfFileCutoff == 0) {
        if (ms.fileInf != NULL)
          fclose(ms.fileInf);
        ms.fileInf = FileOpenInfinite("w");
        if (ms.fileInf == NULL)
          PrintSz_E("The next file could not be created.");
      } else if (ms.fileInf == NULL) {
        ms.fileInf = FileOpenInfinite("a");
        if (ms.fileInf == NULL)
          PrintSz_E("The file could not be reopened.");
      }
    }

    // Copy the bottom section to the top, and blank out a new bottom section.
    BlockMove(b, b, xl, (yh >> 1) + 1, xh, yh & ~1, xl, yl);
    Block(b, xl, yh >> 1, xh, yh & ~1, fOn);
    if (FOdd(xh))
      LineY(b, xh, yh >> 1, yh & ~1, fOff);
    if (!ms.fInfEller)
      fHunt = fTrue;
    else
      yp = yh >> 1;
  }

  UpdateDisplay();
  if (phase != INFEND) {
    yhalf = yh >> 1;
    count2 = 0;
  } else {
    MakeEntranceExit(b, 7);
    yhalf = yh;
    count2 = 1;
  }

  // Figure out how many cells need to be carved to finish the section(s).
  if (phase == INFSTART)
    count = (long)(xh >> 1) * ((yhalf + 1) >> 1) - 1;
  else if (phase == INFCONTINUE)
    count = (long)(xh >> 1) * ((yhalf + 1) >> 1) - ms.nInfCount;
  else
    count = (long)(xh >> 1) * (yh >> 1) - ms.nInfCount;

  // If the infinite Maze is being created with Eller's algorithm, just create
  // rows one by one to completely fill the new section.
  if (ms.fInfEller) {
    while (yp <= (yh | 1)-2) {
      EllerMakeRow(b, ms.rgnInfEller, ms.rgnInfEller + xs, xs, xinc1, xinc2,
        yp, 1, phase == INFEND && yp >= (yh | 1)-2);
      yp += 2;
    }
    goto LAfterMaze;
  }

  // Use a version of the Hunt and Kill algorithm to create the next section
  // of the infinite Maze. This stops when the top half is done, leaving the
  // bottom half partially created to continue with in the next iteration.
  loop {
    if (!MG(b, x, y)) {
      if (ms.fRiver || fHunt) {
        d = DirFindUncreated(b, &x, &y);
        fHunt = d < 0;
      } else {
        d = RndDir();
        xnew = x + xoff2[d];
        ynew = y + yoff2[d];
        if (FLegalMaze2(xnew, ynew) && MG(b, xnew, ynew)) {
          x = xnew; y = ynew;
        } else
          fHunt = fTrue;
      }
    }
    if (!fHunt) {
      M0(b, x-xoff[d], y-yoff[d]);
      M0(b, x, y);
      pass = 0;
      if (y <= yhalf) {
        count--;

        // Done when all cells in the top section have been carved, and at
        // least one cell in the bottom section has been carved.
        if (count <= 0 && count2 > 0)
          break;
      } else {
        count2++;

        // If carving reaches bottom row of the bottom section, teleport to
        // top row of the top section, to give bias to finishing top section.
        if (y >= yh-1) {
          y = 1;
          fHunt = fTrue;
        }
      }
    } else {
      if (x + xInc > xl && x + xInc < xh)
        x += xInc;
      else {
        neg(xInc);
        if (y >= yhalf) {
          pass++;
          if (pass > 2)
            break;

          // If hunting reaches an edge in the bottom section, teleport to the
          // top row of top section, to give bias to finishing top section.
          y = 1;
          UpdateDisplay();
        } else if (y + yInc > yl && y + yInc < yh) {
          y += yInc;
        } else {
          neg(yInc);
          pass++;
          if (pass > 2)
            break;
          UpdateDisplay();
        }
      }
    }
  }

LAfterMaze:
  ms.nInfGen++;
  ms.fInfMiddle = phase != INFEND;
  ms.nInfCount = count2;
  if (ms.fInfFile && ms.fileInf != NULL) {

    // Write top section (or both sections if done with the Maze) to file.
    if (phase != INFEND)
      y = MY(b) >> 1;
    else
      y = MY(b)-1 + ms.fInfFileDos;
    if (FAllocateBitmap(&b2, MX(b), y, b)) {
      BlockMove(b, b2, 0, 0, MX(b)-1, y-1, 0, 0);
      if (ms.fInfFileDos)
        WriteTextDOS(b2, ms.fileInf, 0, phase == INFEND ? 1 : 2);
      else
        WriteText(b2, ms.fileInf, fTrue, fFalse, fFalse);
      DeallocateBitmap(b2);
    }

    // Close the file or the current file when done writing to it.
    if (phase == INFEND ||
      (ms.nInfFileCutoff > 0 && ms.nInfGen % ms.nInfFileCutoff == 0)) {
      fclose(ms.fileInf);
      ms.fileInf = NULL;
    } else
      fflush(ms.fileInf);
  }
  return fTrue;
}

/* create2.cpp */
