/*
** Daedalus (Version 2.5) File: graphics.cpp
** By Walter D. Pullen, Astara@msn.com, http://www.astrolog.org/labyrnth.htm
**
** IMPORTANT NOTICE: Daedalus and all Maze generation and general
** graphics routines used in this program are Copyright (C) 1998-2013 by
** Walter D. Pullen. Permission is granted to freely use, modify, and
** distribute these routines provided these credits and notices remain
** unmodified with any altered or distributed versions of the program.
** The user does have all rights to Mazes and other graphic output
** they make in Daedalus, like a novel created in a word processor.
**
** More formally: This program is free software; you can redistribute it
** and/or modify it under the terms of the GNU General Public License as
** published by the Free Software Foundation; either version 2 of the
** License, or (at your option) any later version. This program is
** distributed in the hope that it will be useful and inspiring, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** General Public License for more details, a copy of which is in the
** LICENSE.HTM included with Daedalus, and at http://www.gnu.org
**
** This file contains general graphics routines, that operate on a monochrome
** bitmap, unrelated to Mazes.
**
** Converted from Modula2: 3/8/1991.
** Last code change: 10/10/2013.
*/

#include <stdio.h>
#include <math.h>
#include "util.h"
#include "graphics.h"


CONST char *rgszDir[DIRS] = {"North", "West", "South", "East"};

// Offset table for orthogonal and diagonal directions.
CONST int xoff[DIRS2] = { 0,-1, 0, 1, -1, -1, 1, 1};
CONST int yoff[DIRS2] = {-1, 0, 1, 0, -1, 1, 1, -1};

CONST int xoff2[DIRS2] = { 0,-2, 0, 2, -2, -2, 2, 2};
CONST int yoff2[DIRS2] = {-2, 0, 2, 0, -2, 2, 2, -2};

// Simple truth table for logical operators, given the number of true bits.
CONST int rgfOr[3]  = {0, 1, 1};
CONST int rgfAnd[3] = {0, 0, 1};
CONST int rgfXor[3] = {0, 1, 0};

// Dither table giving the order to set points in a 4x4 pixel patch.
CONST int rgnDitherPoint[16] =
  {6, 14,  7, 15,
   9,  1, 12,  4,
   8, 16,  5, 13,
  11,  3, 10,  2};

// Bit flags for all combinations of a consecutive group of <= 3 neighbors.
CONST int rggrfNeighbor[25] = {0x00,
  0x01, 0x10, 0x02, 0x20, 0x04, 0x40, 0x08, 0x80,
  0x11, 0x12, 0x22, 0x24, 0x44, 0x48, 0x88, 0x81,
  0x13, 0x32, 0x26, 0x64, 0x4C, 0xC8, 0x89, 0x91};

// Turtle font for height 5 variable width characters.
CONST char *szTurtleCh1[128-32] = {
  "BR4", "D2BD2D0BU4BR2", "DBR2UBR2", "BRD4BR2U4BFL4BD2R4BU3BR2",
  "NR2D2R2D2LNLU4BR3", "D0BR2DG2DBR2D0BU4BR2", "BFD2LUR2BE2", "NDBR2",
  "BRGD2FBU4BR2", "FD2GBU4BR3", "BFDNLNDNRNHNGNFEBUBR2", "BFD2BHR2BE2",
  "BD4EBU3BR2", "BD2R2BE2", "BD4D0BU4BR2", "BD4UE2UBR2", // Symbols

  "BRGD2FEU2HBR3", "RD4NRLBE4", "R2D2L2D2NR2BE4", "R2D2NL2D2L2BE4",
  "D2R2ND2U2BR2", "NR2D2RFGLBE4", "NR2D4R2U2NL2BE2", "R2DG2DBE4",
  "D4R2U2NL2U2L2BR4", "ND2R2D2NL2D2L2BE4", // Digits

  "BDD0BD2D0BU3BR2", "BFD0BD2GBU4BR3", "BR2G2F2BU4BR2", "BDR2BD2L2BU3BR4",
  "F2G2BE4", "R2D2L2BD2D0BE4", "BF2U2L2D4NR2BE4", // Symbols

  "NR2D2ND2R2ND2U2BR2", "ND4RFGNLFGLBE4", "ND4R2DBD2DL2BE4", "ND4RFD2GLBE4",
  "NR2D2NR2D2NR2BE4", "NR2D2NR2D2BE4", "NR2D4R2U2BE2", "D2ND2R2ND2U2BR2",
  "RNRD4NRLBE4", "BD3DR2U4BR2", "D2ND2RFDBU3NGUBR2", "D4NR2BE4",
  "ND4RFND3ERND4BR2", "ND4R2ND4BR2", "D4R2U4L2BR4", "D2ND2R2U2L2BR4",
  "D4REU3NL2BD4HBE3", "D2ND2RFDBU3NGHLBR4", "NR2D2R2D2L2BE4", "RND4RBR2",
  "D4R2U4BR2", "D3FEU3BR2", "D4RENU3FRU4BR2", "DF2DBL2UE2UBR2",
  "D2RND2RU2BR2", "R2DG2DNR2BE4", // Upper Case Letters

  "NR2D4NR2BE4", "DF2DBU4BR2", "R2D4L2BE4", "BDEFBUBR2",
  "BD4NR2BE4", "NFBR3", // Symbols

  "", "", "", "", "", "", "", "", "", "", "", "", "",
  "", "", "", "", "", "", "", "", "", "", "", "", "", // Lower Case Letters

  "BR2LDGFDRBU4BR2", "ND4BR2", "RDFGDLBE4", "BDEDEBR2",
  "BRGFEHBR3"}; // Symbols

// Turtle font for 5x9 characters.
CONST char *szTurtleCh2[128-32] = {
  "", "BR2D4BD2D0", "BRD2BR2U2", "BD2R4BD2L4BFU4BR2D4", "BR2D6BENL3EHL2HER3",
  "RDLNUBR4G4BR4DLUR", "BD2NF4UEFDG2DFRE2", "BR2DG", "BR3G2D2F2", "BRF2D2G2",
  "BD2FNGRNU2ND2RNEF", "BD3R2NU2ND2R2", "BD5BR2DG", "BD3R4", "BD6BRRULD",
  "BD5E4", // Symbols

  "BDD4NE4FR2EU4HL2G", "BFED6NLR", "BDER2FDG4R4", "BDER2FDGNLFDGL2H",
  "D3R3NU3ND3R", "NR4D3R3FDGL2H", "BR3NFL2GD4FR2EUHL3", "R4DG4D",
  "BDDFNR2GDFR2EUHEUHL2G", "BD5FR2EU4HL2GDFR3", // Numbers

  "BR2BD2D0BD2D0", "BR2BD2D0BD2G", "BR3G3F3", "BD2R4BD2L4", "BRF3G3",
  "BDER2FDGLDBD2D0", "BF2DFEU2HL2GD4FR2", // Symbols

  "BD6U4E2F2D2NL4D2", "D6R3EUHNL3EUHL3", "BR3NFL2GD4FR2E", "D6R2E2U2H2L2",
  "NR4D3NR3D3R4", "NR4D3NR3D3", "BR3NFL2GD4FR2EU2L2", "D3ND3R4NU3D3",
  "BRRNRD6NLR", "BD4DFR2EU5", "D3ND3RNE3F3", "D6R4", "ND6F2NDE2D6",
  "ND6F4ND2U4", "BDD4FR2EU4HL2G", "R3FDGL3NU3D3", "BDD4FRENHNFEU3HL2G",
  "ND6R3FDGL2NLF3", "BR3NFL2GDFR2FDGL2H", "R2NR2D6", "D5FR2EU5",
  "D2FDFNDEUEU2", "D6E2NUF2U6", "DF4DBL4UE4U", "D2FRND3REU2",
  "R4DG4DR4", // Upper Case Letters

  "BR3L2D6R2", "BDF4", "BRR2D6L2", "BD2E2F2", "BD6R4", "BR2DF", // Symbols

  "BF4G2LHU2ER2FD3", "D5NDFR2EU2HL2G", "BF4BUHL2GD2FR2E", "BR4D5NDGL2HU2ER2F",
  "BD4R4UHL2GD2FR3", "BD3RNR3ND3U2ERF", "BD8R3EU4HL2GD2FR2E", "D3ND3ER2FD3",
  "BR2D0BD2D4", "BR3D0BD2D5GLH", "D4ND2REREBD4HLH", "BR2D6",
  "BD2DND3EFNDEFD3", "BD2DND3ER2FD3", "BD3D2FR2EU2HL2G", "BD2DND5ER2FD2GL2H",
  "BR4BD8U5HL2GD2FR2E", "BD2DND3ER2F", "BD6R3EHL2HER3", "BR2D2NL2NR2D4",
  "BD2D3FRE2NU2D2", "BD2DFDFEUEU", "BD2D3FENUFEU3", "BD2F2NG2NE2F2",
  "BD2D3FR2ENU3D2GL3", "BD2R4G4R4", // Lower Case Letters

  "BR3GDGFDF", "BR2D2BD2D2", "BRFDFGDG", "BFEFE", "BR2GFEH"}; // Symbols

// Turtle font for 8x13 unicursal characters.
CONST char *szTurtleCh3[128-32] = {
  "R7", "R3ULU2RU2LU7R3D7LD2RD2LDR3", "RU8LU4R2D4R3U4R2D4LD8R",
  "R2U3RUL3UR2U3L2URU3RD3R3U3RD3RDL2D3R2DL3DRD3R2",
  "URDR2U5RD5R2U6L6U6R7D2LUL2D3LU3L2D4R6D7",
  "U3RURURURURURU2L6U2R2DR4URD5LDLDLDLDLDLD2R4U2R2DLDR",
  "R3U4LULULUR3U3RD3R3DL3D6R3", "R3U8LU2RU2R2D4LD8R3",
  "R5ULULU2LU4RU2RURURD2LDLD2LD2RD2RDRD2R",
  "RU2RURU2RU2LU2LULU2RDRDRD2RD4LD2LDLDR5",
  "R3U5LDL2URUR2UL2ULUR2DRU3RD3RUR2DLDL2DR2DRDL2ULD5R3",
  "R3U6L3UR3U3RD3R3DL3D6R3", "R2U3RURD2LD2R4", "R3U6L3UR7DL3D6R3",
  "R2U3R2D2LDR4", "U2RU2RU2RU2RU2RU2R2DLD2LD2LD2LD2LD2LDR6", // Symbols

  "R2ULULU9RUR5DRD9LDL3DR4", "RURDRU11L2UR3D12RURDR",
  "U5RUR5U5L5DLU2R7D7L5DLD4R6", "URDR5U6L4UR4U4L5DLU2R7D12",
  "R5U6L5U6RD5R4U5RD5RDLD6R", "URDR5U6L6U6R7DL6D4R6D7",
  "U12R7D2LUL5D5R6D5L6DR6", "U3RURURURURURU3L6UR7D5LDLDLDLDLDLD2R6",
  "U12R7D5L6DR6D5L6DR6", "U2RD2R5U7L6U5R7D12", // Digits

  "R3U2LU2RU4LU2R3D2LD4RD2LD2R3", "RU2RURU5LU2R3D2LD4RD2LDL2DR5",
  "R6ULULULULULULURURURURURUR2DLDLDLDLDLDRDRDRDRDRD2",
  "R3U4L3UR3U3L3UR7DL3D3R3DL3D4R3",
  "U2RURURURURULULULULULUR2DRDRDRDRDRDLDLDLDLDLDLDR6",
  "U3RU3RUR3URU2LUL3DLDLU2RUR5DRD4LDL3DLD2RD2L2DR6",
  "RULU10RUR5DRD5LDL3U3RD2RURU3LUL3DLD8RD2RURDRURDR", // Symbols

  "U11RUR5DRD5L6DR6D5", "U12R6DRD3LDL5DR5DRD3LDL5DR6",
  "U12R7D3LU2L5D11RURDRURDRU2RD2", "U12R5DRDRD7LDLDL4DR6",
  "U12R7DL6D4R5DL5D6RURDRURDRURD", "U12R7DL6D5R5DL5D5R6",
  "U12R7D3LU2L5D11RURDRURDRU5LUR2D6", "U12RD6R5U6RD12",
  "RURDRU11L2UR5DL2D11RURDR", "U2RD2RURDRURDRU12RD12",
  "U12RD6R2URURURU3RD4LDLDLDLDRDRDRDRD", "U12RD12RURDRURDRURD",
  "U12R3D11RU11R3D12", "U12R5DRDRD10", "U12R7D11L6DR6",
  "U12R6DRD4LDL5D6R6", "U12R7D10LULULULD2RDRDL4DR6",
  "U12R6DRD4LDL5DRDRDRDRDRDR", "URDR5U5L6U7R7D2LUL5D5R6D6",
  "R3U11L3UR7DL3D11R3", "U12RD12RURDRURDRU12RD12",
  "R3U2LULULU8RD7RDRDRURURU7RD8LDLDLD2R3", "U12RD12R2U11RD11R2U12RD12",
  "U3RURURU2LULULU3RD2RDRDRURURU2RD3LDLDLD2RDRDRD3",
  "R3U5L3U7RD6R5U6RD7L3D5R3",
  "U3RURURURURURU3L6UR7D5LDLDLDLDLDLD2R6", // Upper Case Letters

  "RU12R5DL4D11RURDRURDR", "R6ULU2LU2LU2LU2LU2LUR2D2RD2RD2RD2RD2RD2",
  "RURDRURDRU11L4UR5D12R", "RU10RURURDRDRD10R", "URDRURDRURDRURD",
  "R3U8LU4R2D2RD2LD8R3", // Symbols

  "", "", "", "", "", "", "", "", "", "", "", "", "",
  "", "", "", "", "", "", "", "", "", "", "", "", "", // Lower Case Letters

  "R5ULULU2LU2L2UR2U2RU2RUR2DLDLD2LD3RD2RDRD2R", "R3U12RD12R3",
  "RU2RURU2RU3LU2LULUR2DRD2RD2R2DL2D2LD2LDLDR5", "U10RURURDRDRURURD12",
  "R3U8LULU2RUR3DRD2LDLD8R3"}; // Symbols

GS gs = {fFalse, fTrue,
  fFalse, NULL, 0, 0, 1, 0, fOn, fTrue, fFalse, 0x1F3, 0x008};


/*
******************************************************************************
** Bitmap Primitives
******************************************************************************
*/

// Allocate a new bitmap of a given size.

flag FAllocateBitmap(bitmap *pb, int x, int y, constbitmap bOld)
{
  bitmap b;
  long cb;

  if (x < 0 || y < 0 || x > xBitmap || y > yBitmap) {
    PrintSzNN_E("Can't create bitmap larger than %d by %d!\n",
      xBitmap, yBitmap);
    return fFalse;
  }
  cb = cbBH + CbBitmap(x, y);
  if (cb < 0) {
    PrintSzNN_E("Can't allocate bitmap of size %d by %d!\n", x, y);
    return fFalse;
  }
  *pb = (bitmap)PAllocate(cb);
  b = *pb;
  if (b == NULL)
    return fFalse;
  MX(b) = x; MY(b) = y;
  b->clRow = CbBitmapRow(x) >> 2;
  b->fColor = fFalse;

  // Inherit the client fields (i.e. 3D dimensions) from passed in bitmap.
  if (bOld != NULL) {
    CopyClient(bOld, b);
  } else {
    b->wClient1 = wClient1Def; b->wClient2 = wClient2Def;
    b->wClient3 = wClient3Def; b->wClient4 = wClient4Def;
  }
  return fTrue;
}


// Assert a point is within the bounds of a bitmap. If not display an error.
// Used by bitmap accessing functions to avoid crashing. If this ever happens,
// it means a bug in whatever feature called the bitmap function.

flag FValidate(constbitmap b, int x, int y, CONST char *pch)
{
  char sz[cchSzDef*2];

  if (!FLegal(b, x, y)) {
    sprintf(S(sz), "Daedalus attempted to %s the dot at (%d, %d).\n"
      "Valid coordinates are from (0, 0) to (%d, %d).\n",
      pch, x, y, MX(b)-1, MY(b)-1);
    PrintSz_E(sz);

    // Turn off error checking to avoid future messages of this type.
    gs.fErrorCheck = fFalse;
    gs.fSlowDot = FSlowDot();
    return fFalse;
  }
  return fTrue;
}


// Turn off a pixel in a bitmap.

void M0(bitmap b, int x, int y)
{
  if (gs.fSlowDot) {
    if (gs.fErrorCheck && !FValidate(b, x, y, "clear"))
      return;
    if (gs.fTraceDot && b == *gs.pbFocus)
      ScreenDot(x, y, fOff, ~0);
  }
  MA(b, x, y) &= ~MB(x);
}


// Turn on a pixel in a bitmap.

void M1(bitmap b, int x, int y)
{
  if (gs.fSlowDot) {
    if (gs.fErrorCheck && !FValidate(b, x, y, "set"))
      return;
    if (gs.fTraceDot && b == *gs.pbFocus)
      ScreenDot(x, y, fOn, ~0);
  }
  MA(b, x, y) |= MB(x);
}


// Invert a pixel in a bitmap.

void MR(bitmap b, int x, int y)
{
  if (gs.fSlowDot) {
    if (gs.fErrorCheck && !FValidate(b, x, y, "invert"))
      return;
    if (gs.fTraceDot && b == *gs.pbFocus)
      ScreenDot(x, y, !MGCore(b, x, y), ~0);
  }
  MA(b, x, y) ^= MB(x);
}


// Return the on/off status of a pixel in a bitmap.

bit _MG(constbitmap b, int x, int y)
{
  if (gs.fErrorCheck) {
    if (!FValidate(b, x, y, "get"))
      return fOff;
  }
  return MGCore(b, x, y);
}


// Set the on/off status of a pixel in a bitmap.

void _MP(bitmap b, int x, int y, bit o)
{
  if (gs.fSlowDot) {
    if (gs.fErrorCheck && !FValidate(b, x, y, "put"))
      return;
    if (gs.fTraceDot && b == *gs.pbFocus)
      ScreenDot(x, y, o, ~0);
  }
  if (o)
    MA(b, x, y) |= MB(x);
  else
    MA(b, x, y) &= ~MB(x);
}


// Return the status of a pixel in a bitmap. Pixels not on the bitmap are
// assumed to be off.

bit Mg(constbitmap b, int x, int y)
{
  return MgFast(b, x, y);
}


// Set the status of a pixel in a bitmap. Does nothing if an attempt is made
// to set a pixel not on the bitmap.

void Mp(bitmap b, int x, int y, bit o)
{
  if (!FLegal(b, x, y))
    return;
  if (gs.fTraceDot && b == *gs.pbFocus)
    ScreenDot(x, y, o, ~0);
  if (o)
    MA(b, x, y) |= MB(x);
  else
    MA(b, x, y) &= ~MB(x);
}


// Force coordinates to be within the bounds of a bitmap, moving them to the
// edge of the bitmap if not within it already.

void Legalize(constbitmap b, int *x, int *y)
{
  if (*x < 0)
    *x = 0;
  else if (*x >= MX(b))
    *x = MX(b)-1;
  if (*y < 0)
    *y = 0;
  else if (*y >= MY(b))
    *y = MY(b)-1;
}


/*
******************************************************************************
** Bitmap Block Editing
******************************************************************************
*/

// Set all pixels within a rectangle in a bitmap.

void Block(bitmap b, int x1, int y1, int x2, int y2, bit o)
{
  int x, y;
  long il1, il2, il, l, l1, l2;

  Legalize(b, &x1, &y1); Legalize(b, &x2, &y2);
  SortN(&x1, &x2);
  SortN(&y1, &y2);

  // Fastest case: If the horizontal coordinates cover the whole width of the
  // bitmap, it covers consecutive memory addresses. Set 32 pixels at a time.
  if (x1 == 0 && x2 == MX(b)-1) {
    il1 = MI(b, x1, y1);
    il2 = MI(b, x2, y2);
    l = o ? ~0 : 0;
    while (il1 <= il2)
      *MJ(b, il1++) = l;
    goto LDone;
  }

  // Fast case: Can still set 32 pixels at a time, just need to do each row
  // separately, and calculate the bit pattern for start and end of each row.
  il1 = MI(b, x1, y1); il2 = MI(b, x2, y1);
  l1 = ~((1L << (x1-1 & 31 ^ 7)) - 1 ^ (255L << (x1-1 & 24)));
  l2 = ((1L << (x2 & 31 ^ 7)) - 1 ^ (255L << (x2 & 24)));
  for (y = y1; y <= y2; y++) {
    for (il = il1; il <= il2; il++) {
      l = ~0;
      if (il == il1 && (x1 & 31) > 0)
        l &= l1;
      if (il == il2)
        l &= l2;
      if (o)
        *MJ(b, il) |= l;
      else
        *MJ(b, il) &= ~l;
    }
    il1 += b->clRow; il2 += b->clRow;
  }

  // Need to update the screen if Show Pixel Edits is on.
LDone:
  if (gs.fTraceDot && b == *gs.pbFocus)
    for (y = y1; y <= y2; y++)
      for (x = x1; x <= x2; x++)
        ScreenDot(x, y, o, ~0);
}


// Invert all pixels within a rectangle in a bitmap.

void BlockReverse(bitmap b, int x1, int y1, int x2, int y2)
{
  int x, y;

  Legalize(b, &x1, &y1); Legalize(b, &x2, &y2);
  SortN(&x1, &x2);
  SortN(&y1, &y2);
  for (y = y1; y <= y2; y++)
    for (x = x1; x <= x2; x++)
      MR(b, x, y);
}


// Set pixels around the border of a rectangle in a bitmap.

void Box(bitmap b, int x1, int y1, int x2, int y2, int xsiz, int ysiz, bit o)
{
  SortN(&x1, &x2);
  SortN(&y1, &y2);
  Block(b, x1, y1, x2, y1 + ysiz - 1, o);
  if (y1 + ysiz < y2) {
    Block(b, x1, y1 + ysiz, x1 + xsiz - 1, y2 - ysiz, o);
    Block(b, x2 - xsiz + 1, y1 + ysiz, x2, y2 - ysiz, o);
  }
  Block(b, x1, y2 - ysiz + 1, x2, y2, o);
}


// Copy pixels in a rectangle from one bitmap to another bitmap. If the
// bitmaps are the same, the rectangles should not overlap.

void BlockMove(constbitmap b1, bitmap b2, int x1, int y1, int x2, int y2,
  int x0, int y0)
{
  int x, y, xT, yT;
  long il1, il2, il0;

  Legalize(b1, &x1, &y1); Legalize(b1, &x2, &y2);
  SortN(&x1, &x2);
  SortN(&y1, &y2);

  // Fast case: If source and destination bitmaps have the same width, and the
  // horizontal spans cover the whole bitmap width, both source and dest
  // rectangles cover consecutive memory addresses. Copy 32 pixels at a time.
  if (MX(b1) == MX(b2) &&
    x1 == 0 && x2 == MX(b1)-1 && x0 == 0 && y0+(y2-y1) < MY(b2)) {
    il1 = MI(b1, x1, y1);
    il2 = MI(b1, x2, y2);
    il0 = MI(b2, x0, y0);
    while (il1 <= il2)
      *MJ(b2, il0++) = *MJ(b1, il1++);
    return;
  }

  // Normal case: Copy one pixel at a time from source to dest rectangle.
  for (y = y1; y <= y2; y++)
    for (x = x1; x <= x2; x++) {
      xT = x0+(x-x1); yT = y0+(y-y1);
      if (FLegal(b2, xT, yT))
        MP(b2, xT, yT, MG(b1, x, y));
    }
}


// Copy pixels in a rectangle from one bitmap to another bitmap, combining
// them using the specified logical operator table.

void BlockMoveOrAnd(constbitmap b1, bitmap b2, int x1, int y1, int x2, int y2,
  int x0, int y0, CONST int *rgf)
{
  int x, y, xT, yT;

  Legalize(b1, &x1, &y1); Legalize(b1, &x2, &y2);
  SortN(&x1, &x2);
  SortN(&y1, &y2);
  for (y = y1; y <= y2; y++)
    for (x = x1; x <= x2; x++) {
      xT = x0+(x-x1); yT = y0+(y-y1);
      if (FLegal(b2, xT, yT))
        MP(b2, xT, yT, rgf[MG(b1, x, y) + MG(b2, xT, yT)]);
    }
}


/*
******************************************************************************
** Bitmap Line Drawing
******************************************************************************
*/

// Draw a horizontal line on a bitmap.

void LineX(bitmap b, int x1, int x2, int y, bit o)
{
  int x;
  long il1, il2, il, l;

  // Don't worry about pixels that are off the bitmap.
  if (y < 0 || y >= MY(b))
    return;
  SortN(&x1, &x2);
  if (x1 < 0)
    x1 = 0;
  if (x2 >= MX(b))
    x2 = MX(b)-1;

  // Quickly draw the line, setting up to 32 pixels at a time.
  il1 = MI(b, x1, y); il2 = MI(b, x2, y);
  for (il = il1; il <= il2; il++) {
    l = ~0;
    if (il == il1 && (x1 & 31) > 0)
      l &= ~((1L << (x1-1 & 31 ^ 7)) - 1 ^ (255 << (x1-1 & 24)));
    if (il == il2)
      l &= ((1L << (x2 & 31 ^ 7)) - 1 ^ (255 << (x2 & 24)));
    if (o)
      *MJ(b, il) |= l;
    else
      *MJ(b, il) &= ~l;
  }

  // Need to update the screen if Show Pixel Edits is on.
  if (gs.fTraceDot && b == *gs.pbFocus)
    for (x = x1; x <= x2; x++)
      ScreenDot(x, y, o, ~0);
}


// Draw a vertical line on a bitmap.

void LineY(bitmap b, int x, int y1, int y2, bit o)
{
  int y;
  long il, mb;

  // Don't worry about pixels that are off the bitmap.
  if (x < 0 || x >= MX(b))
    return;
  SortN(&y1, &y2);
  if (y1 < 0)
    y1 = 0;
  if (y2 >= MY(b))
    y2 = MY(b)-1;

  // Quickly draw the line, where things like the bit mask and memory address
  // only need to be calculated once.
  il = MI(b, x, y1);
  if (o) {
    mb = MB(x);
    for (y = y1; y <= y2; y++) {
      *MJ(b, il) |= mb;
      il += b->clRow;
    }
  } else {
    mb = ~MB(x);
    for (y = y1; y <= y2; y++) {
      *MJ(b, il) &= mb;
      il += b->clRow;
    }
  }

  // Need to update the screen if Show Pixel Edits is on.
  if (gs.fTraceDot && b == *gs.pbFocus)
    for (y = y1; y <= y2; y++)
      ScreenDot(x, y, o, ~0);
}


// Draw a diagonal line of the specified tilt on a bitmap.

void LineZ(bitmap b, int x, int y1, int y2, int dir, bit o)
{
  int i, j;

  if (NAbs(dir) <= 1) {
    for (i = 0; i <= y1 - y2; i++)
      MP(b, x + i*dir, y1 - i, o);
  } else {
    j = dir < 0 ? -1 : 1;
    for (i = 0; i <= y1 - y2; i++)
      LineX(b, x + i*dir, x + (i+1)*dir - j, y1 - i, o);
  }
}


// Draw a line between any two points on a bitmap. This is an implementation
// of Bresenham's algorithm.

void Line(bitmap b, int x1, int y1, int x2, int y2, bit o)
{
  int x = x1, y = y1, dx = x2 - x1, dy = y2 - y1, xInc, yInc, xInc2, yInc2,
    d, dInc, z, zMax;
  flag fAllLegal = FLegal(b, x1, y1) && FLegal(b, x2, y2);

  // Determine slope.
  if (NAbs(dx) >= NAbs(dy)) {
    xInc = NSgn(dx); yInc = 0;
    xInc2 = 0; yInc2 = NSgn(dy);
    zMax = NAbs(dx); dInc = NAbs(dy);
    d = zMax - (!FOdd(dx) && x1 > x2);
  } else {
    xInc = 0; yInc = NSgn(dy);
    xInc2 = NSgn(dx); yInc2 = 0;
    zMax = NAbs(dy); dInc = NAbs(dx);
    d = zMax - (!FOdd(dy) && y1 > y2);
  }
  d >>= 1;

  // Loop over long axis, adjusting short axis for slope as needed.
  for (z = 0; z <= zMax; z++) {
    if (fAllLegal || FLegal(b, x, y)) {
      if (o)
        MA(b, x, y) |= MB(x);
      else
        MA(b, x, y) &= ~MB(x);
    }
    x += xInc; y += yInc; d += dInc;
    if (d >= zMax) {
      x += xInc2; y += yInc2; d -= zMax;
    }
  }

  // Need to update the screen if Show Pixel Edits is on.
  if (gs.fTraceDot && b == *gs.pbFocus)
    UpdateDisplay();
}


// Draw a line between any two points on a bitmap, stopping if a pixel already
// changed is reached.

flag FLineUntil(bitmap b, int x1, int y1, int x2, int y2,
  int *xEnd, int *yEnd, bit o, flag fDraw, flag fWall)
{
  int x = x1, y = y1, xLast = x1, yLast = y1, dx = x2 - x1, dy = y2 - y1,
    xInc, yInc, xInc2, yInc2, d, dInc, z, zMax;
  flag fAllLegal = FLegal(b, x1, y1) && FLegal(b, x2, y2);

  // Determine slope.
  if (NAbs(dx) >= NAbs(dy)) {
    xInc = NSgn(dx); yInc = 0;
    xInc2 = 0; yInc2 = NSgn(dy);
    zMax = NAbs(dx); dInc = NAbs(dy);
    d = zMax - (!FOdd(dx) && x1 > x2);
  } else {
    xInc = 0; yInc = NSgn(dy);
    xInc2 = NSgn(dx); yInc2 = 0;
    zMax = NAbs(dy); dInc = NAbs(dx);
    d = zMax - (!FOdd(dy) && y1 > y2);
  }
  d >>= 1;

  // Loop over long axis, adjusting short axis for slope as needed.
  for (z = 0;; z++) {
    if (fAllLegal || FLegal(b, x, y)) {
      if (MGCore(b, x, y) != o) {
        if (fDraw)
          MP(b, x, y, o);
      } else
        break;
    }
    xLast = x; yLast = y;
    if (z >= zMax)
      break;
    x += xInc; y += yInc; d += dInc;
    if (d >= zMax) {
      x += xInc2; y += yInc2; d -= zMax;
    }
  }

  // Return either the last pixel before the barrier, or the barrier itself.
  if (fWall) {
    xLast = x; yLast = y;
  }
  if (xEnd != NULL)
    *xEnd = xLast;
  if (yEnd != NULL)
    *yEnd = yLast;

  // Need to update the screen if Show Pixel Edits is on.
  if (gs.fTraceDot && b == *gs.pbFocus)
    UpdateDisplay();
  return xLast != x2 || yLast != y2;
}


// Calculate a line between two points, recording the horizontal coordinate at
// each row if it's smaller or larger than the current limits. This is used
// when drawing triangles and quadrilaterals.

void LineSetup(constbitmap b, MM *xminmax, int x1, int y1, int x2, int y2)
{
  int x = x1, y = y1, dx = x2 - x1, dy = y2 - y1, xInc, yInc, xInc2, yInc2,
    d, dInc, z, zMax;

  // Determine slope.
  if (NAbs(dx) >= NAbs(dy)) {
    xInc = NSgn(dx); yInc = 0;
    xInc2 = 0; yInc2 = NSgn(dy);
    zMax = NAbs(dx); dInc = NAbs(dy);
    d = zMax - (!FOdd(dx) && x1 > x2);
  } else {
    xInc = 0; yInc = NSgn(dy);
    xInc2 = NSgn(dx); yInc2 = 0;
    zMax = NAbs(dy); dInc = NAbs(dx);
    d = zMax - (!FOdd(dy) && y1 > y2);
  }
  d >>= 1;

  // Loop over long axis, adjusting short axis for slope as needed.
  for (z = 0; z <= zMax; z++) {
    if (y >= 0 && y < MY(b)) {
      if (x < xminmax[y].min)
        xminmax[y].min = x;
      if (x > xminmax[y].max)
        xminmax[y].max = x;
    }
    x += xInc; y += yInc; d += dInc;
    if (d >= zMax) {
      x += xInc2; y += yInc2; d -= zMax;
    }
  }
}


// Draw a filled in triangle on a bitmap, defined by three points.

flag FTriangle(bitmap b, int x1, int y1, int x2, int y2, int x3, int y3,
  int k)
{
  MM rgmm[zBitmapStack], *pmm = NULL, *xminmax;
  int ymin, ymax, y, yT, xmax, x;

  if (MY(b) <= zBitmapStack)
    xminmax = rgmm;
  else {
    pmm = RgAllocate(MY(b), MM);
    if (pmm == NULL)
      return fFalse;
    xminmax = pmm;
  }
  ymin = NMax(NMin(Min(y1, y2), y3), 0);
  ymax = NMin(NMax(Max(y1, y2), y3), MY(b) - 1);
  for (y = ymin; y <= ymax; y++) {
    xminmax[y].min = xBitmap;
    xminmax[y].max = -1;
  }
  LineSetup(b, xminmax, x1, y1, x2, y2);
  LineSetup(b, xminmax, x2, y2, x3, y3);
  LineSetup(b, xminmax, x3, y3, x1, y1);

  // Draw one horizontal line per row.
  if (k >= 0) {
    for (y = ymin; y <= ymax; y++)
      LineX(b, xminmax[y].min, xminmax[y].max, y, k);
  } else {
    neg(k);
    for (y = ymin; y <= ymax; y++) {
      yT = y & 3;
      xmax = Min(xminmax[y].max, MX(b) - 1);
      for (x = Max(xminmax[y].min, 0); x <= xmax; x++)
        MP(b, x, y, FDitherCore(k, x & 3, yT));
    }
  }
  if (pmm != NULL)
    DeallocateP(pmm);
  return fTrue;
}


// Draw a filled in four sided quadrilateral on a bitmap.

flag FQuadrilateral(bitmap b, int x1, int y1, int x2, int y2, int x3, int y3,
  int x4, int y4, int k)
{
  MM rgmm[zBitmapStack], *pmm = NULL, *xminmax;
  int ymin, ymax, y, yT, xmax, x;

  if (MY(b) <= zBitmapStack)
    xminmax = rgmm;
  else {
    pmm = RgAllocate(MY(b), MM);
    if (pmm == NULL)
      return fFalse;
    xminmax = pmm;
  }
  ymin = NMax(NMin(NMin(Min(y1, y2), y3), y4), 0);
  ymax = NMin(NMax(NMax(Max(y1, y2), y3), y4), MY(b) - 1);
  for (y = ymin; y <= ymax; y++) {
    xminmax[y].min = xBitmap;
    xminmax[y].max = -1;
  }
  LineSetup(b, xminmax, x1, y1, x2, y2);
  LineSetup(b, xminmax, x2, y2, x3, y3);
  LineSetup(b, xminmax, x3, y3, x4, y4);
  LineSetup(b, xminmax, x4, y4, x1, y1);

  // Since one horizontal line is drawn per row, the quadrilateral should be
  // convex, or at least not concave any along the vertical axis.
  if (k >= 0) {
    for (y = ymin; y <= ymax; y++)
      LineX(b, xminmax[y].min, xminmax[y].max, y, k);
  } else {
    neg(k);
    for (y = ymin; y <= ymax; y++) {
      yT = y & 3;
      xmax = Min(xminmax[y].max, MX(b) - 1);
      for (x = Max(xminmax[y].min, 0); x <= xmax; x++)
        MP(b, x, y, FDitherCore(k, x & 3, yT));
    }
  }
  if (pmm != NULL)
    DeallocateP(pmm);
  return fTrue;
}


/*
******************************************************************************
** Bitmap Curve Drawing
******************************************************************************
*/

// Draw a filled in circle or ellipse, within a rectangle on a bitmap.

void Disk(bitmap b, int x1, int y1, int x2, int y2, bit o)
{
  int rx, ry, r, i, j, iT, jT, q, qLo, qHi, m1, m2, n1, n2;

  // This efficient algorithm has no calculations more complicated than
  // addition (and multiplication to proportion in the case of an ellipse).
  rx = NAbs(x2 - x1) >> 1; ry = NAbs(y2 - y1) >> 1;
  if (rx == 0 || ry == 0) {
    Block(b, x1, y1, x2, y2, o);
    return;
  }
  r = Max(rx, ry);
  j = r; q = r+1; qLo = 1; qHi = r+r-1;
  m1 = (x1 + x2) >> 1; m2 = m1 + FOdd(x2 - x1);
  n1 = (y1 + y2) >> 1; n2 = n1 + FOdd(y2 - y1);
  for (i = 0; i <= r; i++) {
    iT = (rx <= ry) ? i : i * ry / rx;
    jT = (rx >= ry) ? j : j * rx / ry;
    LineX(b, m1 - jT, m2 + jT, n1 - iT, o);
    LineX(b, m1 - jT, m2 + jT, n2 + iT, o);
    q -= qLo;
    while (q < 0) {
      q += qHi;
      qHi -= 2;
      j--;
    }
    qLo += 2;
  }
}


// Draw an outline of a circle or ellipse, within a rectangle on a bitmap.

void Circle(bitmap b, int x1, int y1, int x2, int y2, bit o)
{
  int rx, ry, r, imax, i, j, i1, i2, j1, j2, q, qLo, qHi, m1, m2, n1, n2;

  // This efficient algorithm has no calculations more complicated than
  // addition (and multiplication to proportion in the case of an ellipse).
  rx = NAbs(x2 - x1) >> 1; ry = NAbs(y2 - y1) >> 1;
  if (rx == 0 || ry == 0) {
    Block(b, x1, y1, x2, y2, o);
    return;
  }
  r = Max(rx, ry);
  j = r; q = r+1; qLo = 1; qHi = r+r-1;
  m1 = (x1 + x2) >> 1; m2 = m1 + FOdd(x2 - x1);
  n1 = (y1 + y2) >> 1; n2 = n1 + FOdd(y2 - y1);
  imax = NMultDiv(r + 1, 708, 1000);
  for (i = 0; i <= imax; i++) {
    i1 = j2 = j; j1 = i2 = i;
    if (rx < ry) {
      i1 = i1 * rx / ry; i2 = i2 * rx / ry;
    } else if (ry < rx) {
      j1 = j1 * ry / rx; j2 = j2 * ry / rx;
    }
    Mp(b, m1 - i1, n1 - j1, o); Mp(b, m1 - i2, n1 - j2, o);
    Mp(b, m2 + i1, n1 - j1, o); Mp(b, m2 + i2, n1 - j2, o);
    Mp(b, m1 - i1, n2 + j1, o); Mp(b, m1 - i2, n2 + j2, o);
    Mp(b, m2 + i1, n2 + j1, o); Mp(b, m2 + i2, n2 + j2, o);
    q -= qLo;
    if (q < 0) {
      q += qHi;
      qHi -= 2;
      j--;
    }
    qLo += 2;
  }
}


// Draw one quadrant or 90 degrees of a circle.

void ArcQuadrant(bitmap b, int x, int y, int r, int d, bit o)
{
  int imax = NMultDiv(r, 708, 1000), i, j = r, q = r+1, qLo = 1, qHi = r+r-1,
    xi, yi;

  // This efficient algorithm has no calculations more complicated than
  // addition (and multiplication for the sign to indicate the quadrant).
  if (r == 0) {
    MP(b, x, y, o);
    return;
  }
  xi = d > 0 && d < 3 ? -1 : 1; yi = d < 2 ? 1 : -1;
  for (i = 0; i <= imax; i++) {
    q -= qLo;
    if (q < 0) {
      q += qHi;
      qHi -= 2;
      j--;
    }
    qLo += 2;
    MP(b, x + xi*j, y + yi*i, o);
    MP(b, x + xi*i, y + yi*j, o);
  }
}


// Draw a subsection of one quadrant or 90 degrees of a circle.

void ArcQuadrantSub(bitmap b, int x, int y, int r, int d,
  int dx, int dy, int *pdx, int *pdy, bit o)
{
  int imax = NMultDiv(r, 708, 1000), i, j = r, q = r+1, qLo = 1, qHi = r+r-1,
    xi, yi;

  // Use the same efficient algorithm as in ArcQuadrant().
  if (r == 0) {
    MP(b, x, y, o);
    return;
  }
  if (dx < 0)
    dx = 0;
  if (dy < 0)
    dy = 0;
  xi = d > 0 && d < 3 ? -1 : 1; yi = d < 2 ? 1 : -1;
  for (i = 0; i <= imax; i++) {
    q -= qLo;
    if (q < 0) {
      q += qHi;
      qHi -= 2;
      j--;
    }
    qLo += 2;
    if (i >= dy && j >= dx) {
      MP(b, x + xi*j, y + yi*i, o);
      if (i == dy && pdx != NULL)
        *pdx = j;
      if (j == dx && pdy != NULL)
        *pdy = i;
    }
    if (i >= dx && j >= dy) {
      MP(b, x + xi*i, y + yi*j, o);
      if (i == dx && pdy != NULL)
        *pdy = j;
      if (j == dy && pdx != NULL)
        *pdx = i;
    }
  }
}


// Draw an arc of an ellipse, specifically some subsection of sides of a dMax
// sized polygon.

void ArcPolygon(bitmap b, int x, int y, int rx, int ry,
  int d1, int d2, int dInc, int dMax, bit o)
{
  int i, m, n, m0, n0;
  real rMax;

  if (dMax < 0) {
    dMax = 64;
    if (rx + ry > 384)
      dMax = 256;
    d1 = 0;
    d2 = dMax;
  }
  rMax = (real)dMax;
  for (i = d1; i <= d2; i += dInc) {
    AngleR(&m, &n, x, y, (real)rx, (real)ry, (real)i * rDegMax / rMax);
    if (i > d1)
      Line(b, m0, n0, m, n, o);
    m0 = m; n0 = n;
  }
}


// Draw an arc of a circle.

void ArcReal(bitmap b, real x, real y, real r, real d1, real d2, bit o)
{
  real h, v, d, dInc;
  int m, n, m0, n0, i, iMax;

  // Figure out how many lines should be drawn, based on the radius.
  iMax = (int)((d2 - d1) / rDegQuad * r + rRound);
  iMax = Max(iMax, 1);
  dInc = (d2 - d1) / (real)iMax;

  // Draw the sides as if this were a subsection of a polygon.
  for (i = 0; i <= iMax; i++) {
    d = (d1 + (real)i * dInc);
    h = x + r*RCosD(d) + rRound;
    v = y + r*RSinD(d) + rRound;
    m = (int)h; n = (int)v;
    if (i > 0)
      Line(b, m0, n0, m, n, o);
    m0 = m; n0 = n;
  }
}


/*
******************************************************************************
** Bitmap Special Editing
******************************************************************************
*/

// Parse and return a number from a turtle subaction string, advancing the
// string index to after the number.

int TurtleNumber(CONST char *command, int *index)
{
  int num;
  char cmd;

  cmd = command[*index];
  if (!FDigitCh(cmd))
    return 1;
  num = 0;
  loop {
    if (!FDigitCh(cmd))
      return num;
    if (num >= 100000) {
      PrintSz_E("Turtle number overflow.\n");
      return 0;
    }
    num = num*10 + (cmd - '0');
    (*index)++;
    cmd = command[*index];
  }
}


// Draw a sequence of lines on the bitmap, based on the subactions in command.
// This is similar to the "draw" command in BASIC.

flag Turtle(bitmap b, CONST char *command)
{
  int index, x, y, deltax, deltay, deltaz, relx, rely, i, temp;
  flag fBlank = fFalse, fNoupdate = fFalse;
  char szT[cchSzDef], cmd, ch, chT;

  index = 0;
  loop {
    cmd = ChCap(command[index]);
    index++;
    switch (cmd) {
    case chNull:
      goto LDone;

    // Blank: The next action will just update the pen position and not draw.
    case 'B':
      fBlank = fTrue;
      break;

    // No update: The next action will not update the pen position.
    case 'N':
      fNoupdate = fTrue;
      break;

    // Color: Set whether to draw using on or off pixels.
    case 'C':
      gs.turtlec = FOdd(TurtleNumber(command, &index));
      break;

    // Scale: Set the scale to apply to all relative motion.
    case 'S':
      gs.turtles = TurtleNumber(command, &index);
      break;

    // Angle: Set the rotation to apply to all relative motion.
    case 'A':
      gs.turtlea = TurtleNumber(command, &index);
      break;

    // Move to the specified absolute or relative coordinates.
    case 'M':
      relx = rely = 0;
      if (command[index] == '-') {
        index++;
        relx = -1;
      } else if (command[index] == '+') {
        index++;
        relx = 1;
      }
      deltax = TurtleNumber(command, &index);
      if (relx < 0)
        neg(deltax);
      if (command[index] != ',') {
        PrintSz_E("Bad turtle coordinates.\n");
        return fFalse;
      }
      index++;
      if (command[index] == '-') {
        index++;
        rely = -1;
      } else if (command[index] == '+') {
        index++;
        rely = 1;
      }
      deltay = TurtleNumber(command, &index);
      if (rely < 0)
        neg(deltay);
      if (relx != 0 && rely != 0) {
        switch (gs.turtlea) {
        case 1:
          temp   = deltax;
          deltax = -deltay;
          deltay = temp;
          break;
        case 2:
          neg(deltax);
          neg(deltay);
          break;
        case 3:
          temp   = deltax;
          deltax = deltay;
          deltay = -temp;
          break;
        }
      }
      x = gs.turtlex;
      y = gs.turtley;
      if (relx == 0)
        gs.turtlex = deltax;
      else
        gs.turtlex += deltax * gs.turtles;
      if (rely == 0)
        gs.turtley = deltay;
      else
        gs.turtley += deltay * gs.turtles;
      if (!fBlank)
        Line(b, x, y, gs.turtlex, gs.turtley, gs.turtlec);
      if (fNoupdate) {
        gs.turtlex = x;
        gs.turtley = y;
      }
      break;

    // Draw a line (or arc) from the current pen position.
    case 'U': case 'D': case 'L': case 'R':
    case 'E': case 'F': case 'G': case 'H':
    case 'O':
      switch (cmd) {
      case 'U': deltax =  0; deltay = -1; break;
      case 'D': deltax =  0; deltay = 1;  break;
      case 'L': deltax = -1; deltay = 0;  break;
      case 'R': deltax =  1; deltay = 0;  break;
      case 'E': deltax =  1; deltay = -1; break;
      case 'F': deltax =  1; deltay = 1;  break;
      case 'G': deltax = -1; deltay = 1;  break;
      case 'H': deltax = -1; deltay = -1; break;
      case 'O':
        ch = ChCap(command[index]);
        index++;
        switch (ch) {
          case 'A': deltax =  1; deltay = -1; deltaz = fFalse; break;
          case 'B': deltax =  1; deltay = -1; deltaz = fTrue;  break;
          case 'C': deltax =  1; deltay = 1;  deltaz = fTrue;  break;
          case 'D': deltax =  1; deltay = 1;  deltaz = fFalse; break;
          case 'E': deltax = -1; deltay = 1;  deltaz = fFalse; break;
          case 'F': deltax = -1; deltay = 1;  deltaz = fTrue;  break;
          case 'G': deltax = -1; deltay = -1; deltaz = fTrue;  break;
          case 'H': deltax = -1; deltay = -1; deltaz = fFalse; break;
          default:
            gs.turtleo = (ch == '1'); continue;
        }
        break;
      }
      switch (gs.turtlea) {
      case 1:
        temp   = deltax;
        deltax = -deltay;
        deltay = temp;
        break;
      case 2:
        deltax = -deltax;
        deltay = -deltay;
        break;
      case 3:
        temp   = deltax;
        deltax = deltay;
        deltay = -temp;
        break;
      }
      temp = TurtleNumber(command, &index) * gs.turtles;
      x = gs.turtlex;
      y = gs.turtley;
      if (fBlank) {
        gs.turtlex += deltax*temp;
        gs.turtley += deltay*temp;
      } else {
        Mp(b, gs.turtlex, gs.turtley, gs.turtlec);
        if (cmd != 'O') {

          // Draw an orthogonal or diagonal line.
          for (i = 0; i < temp; i++) {
            gs.turtlex += deltax;
            gs.turtley += deltay;
            Mp(b, gs.turtlex, gs.turtley, gs.turtlec);
          }
        } else {

          // Draw an arc quadrant.
          if (!gs.turtleo) {
            if (deltaz)
              for (i = 0; i < temp; i++) {
                gs.turtlex += deltax;
                Mp(b, gs.turtlex, gs.turtley, gs.turtlec);
              }
            for (i = 0; i < temp; i++) {
              gs.turtley += deltay;
              Mp(b, gs.turtlex, gs.turtley, gs.turtlec);
            }
            if (!deltaz)
              for (i = 0; i < temp; i++) {
                gs.turtlex += deltax;
                Mp(b, gs.turtlex, gs.turtley, gs.turtlec);
              }
          } else {
            relx = gs.turtlex + (deltaz ? 0 : deltax*temp);
            rely = gs.turtley + (deltaz ? deltay*temp : 0);
            i = ((ch - 'A') >> 1) + gs.turtlea;
            if (!FOdd(ch - 'A'))
              i += 2;
            i &= 3;
            ArcQuadrant(b, relx, rely, temp, i, gs.turtlec);
            gs.turtlex += deltax*temp;
            gs.turtley += deltay*temp;
          }
        }
      }
      if (fNoupdate) {
        gs.turtlex = x;
        gs.turtley = y;
      }
      break;

    // Draw text or set the font to use.
    case 'T':
      chT = command[index++];
      if (chT == chNull)
        goto LDone;
      if (FBetween(chT, '0', '2')) {
        gs.turtlet = (chT - '0');
        break;
      }
      if (ChCap(chT) == 'T')
        chT = chNull;
      x = gs.turtlex; y = gs.turtley;
      loop {
        ch = command[index++];
        if (ch == chNull)
          goto LDone;
        if (ch == chT)
          break;
        if (ch == '\n') {
          i = gs.turtlet == 0 ? 6 : (gs.turtlet == 1 ? 10 : 16);
          i *= gs.turtles;
          gs.turtlex = x + xoff[gs.turtlea]*i;     // 0, -1, 0, 1
          gs.turtley = y + yoff[gs.turtlea ^ 2]*i; // 1, 0, -1, 0
          x = gs.turtlex; y = gs.turtley;
        }
        if (FBetween(ch, 32, 127)) {
          switch (gs.turtlet) {
          case 0:
            ch = ChCap(ch);
            Turtle(b, szTurtleCh1[ch - 32]);
            break;
          case 1:
            x = gs.turtlex; y = gs.turtley;
            Turtle(b, szTurtleCh2[ch - 32]);
            i = 6 * gs.turtles;
            gs.turtlex = x + xoff[gs.turtlea ^ 3]*i; // 1, 0, -1, 0
            gs.turtley = y + yoff[gs.turtlea ^ 3]*i; // 0, 1, 0, -1
            break;
          case 2:
            ch = ChCap(ch);
            Turtle(b, "R2");
            Turtle(b, szTurtleCh3[ch - 32]);
            Turtle(b, "R2");
            break;
          }
        }
      }
      break;

    // Fill or flood the region the pen is over.
    case '`':
      FFill(b, gs.turtlex, gs.turtley, gs.turtlec);
      break;
    case '~':
      FFlood(b, gs.turtlex, gs.turtley, gs.turtlec);
      break;

    // Reset Turtle settings to their defaults.
    case '_':
      gs.turtlex = gs.turtley = gs.turtlea = 0;
      gs.turtles = 1;
      gs.turtlec = fOn;
      gs.turtleo = fTrue;
      gs.turtlet = fFalse;
      break;

    default:
      sprintf(S(szT), "Bad turtle subaction '%c'.\n", cmd); PrintSz_E(szT);
      return fFalse;
    }
    if (cmd != 'B' && cmd != 'N')
      fBlank = fNoupdate = fFalse;
  }
LDone:
  return fTrue;
}


// Flood an irregular shape in a bitmap with the given bit color.

flag FFillCore(bitmap b, int x, int y, bit o, flag fFlood)
{
  PT *rgpt;
  long ipt = 0;
  int xnew, ynew, d, dMax;

  if (!FLegalFill(b, x, y, o))
    return fTrue;
  if (F64K(b))
    return fFalse;
  rgpt = RgAllocate(MX(b)*MY(b), PT);
  if (rgpt == NULL)
    return fFalse;
  dMax = DIRS + fFlood*DIRS;
LSet:
  MP(b, x, y, o);
LNext:
  for (d = 0; d < dMax; d++) {
    xnew = x + xoff[d]; ynew = y + yoff[d];
    if (FLegalFill(b, xnew, ynew, o)) {
      FillPush(x, y);
      x = xnew; y = ynew;
      goto LSet;
    }
  }
  if (ipt > 0) {
    FillPop(x, y);
    goto LNext;
  }
  DeallocateP(rgpt);
  return fTrue;
}


// Calculate and display the perimeter of a shape of on pixels.

long AnalyzePerimeter(constbitmap b, int x0, int y0, flag fCorner)
{
  long cLength = 0, cSide = 0;
  int x, y, x1, y1, x2, y2, d = 0, dOld = -1;
  char sz[cchSzMax];

  // Find an edge pixel of a shape.
  if (!FLegal(b, x0, y0) || !MG(b, x0, y0)) {
    if (!FBitmapFind(b, &x0, &y0, fOn))
      return 0;
  }
  while (Mg(b, x0 + 1, y0))
    x0++;
  x = x0; y = y0;

  // Loop around the shape boundary, counting the number of segments.
  do {
    dOld = d;
    x1 = x + xoff[d]; y1 = y + yoff[d];
    x2 = x1 + xoff[d-1 & DIRS1]; y2 = y1 + yoff[d-1 & DIRS1];
    if (!fCorner) {
      if (Mg(b, x1, y1)) {
        if (Mg(b, x2, y2)) {
          x = x2; y = y2;
          d = d-1 & DIRS1;
        } else {
          x = x1; y = y1;
        }
      } else
        d = d+1 & DIRS1;
    } else {
      if (Mg(b, x2, y2)) {
        x = x2; y = y2;
        d = d-1 & DIRS1;
      } else if (Mg(b, x1, y1)) {
        x = x1; y = y1;
      } else
        d = d+1 & DIRS1;
    }
    cLength++;
    if (d != dOld)
      cSide++;
    dOld = d;
  } while (x != x0 || y != y0 || d != 0);

  cSide = cSide & ~1;
  sprintf(S(sz),
    "Perimeter length: %d\nPerimeter sides: %d\nAverage length: %.2lf\n",
    cLength, cSide, (real)cLength / (real)cSide);
  PrintSz_N(sz);
  return cLength;
}


/*
******************************************************************************
** Bitmap Transformations
******************************************************************************
*/

// Return the number of on pixels in a bitmap.

long BitmapCount(constbitmap b)
{
  int x, y;
  long count = 0;
  dword l;
  byte *pb = (byte *)&l;

  for (y = 0; y < MY(b); y++)
    for (x = 0; x < MX(b); x += 32) {
      l = MA(b, x, y);
      if (x + 32 > MX(b)) {
        SwapN(pb[0], pb[3]); SwapN(pb[1], pb[2]);
        // Clear bits off the right edge of the bitmap.
        l &= ~((1L << (x + 32 - MX(b))) - 1);
      }

      // Fast way of counting the number of on bits in a long. Add neighboring
      // bits together, then groups of four, then eight, etc.
      l = ((l & 0xaaaaaaaaL) >> 1) + (l & 0x55555555L);
      l = ((l & 0xccccccccL) >> 2) + (l & 0x33333333L);
      l = ((l & 0xf0f0f0f0L) >> 4) + (l & 0x0f0f0f0fL);
      l = ((l & 0xff00ff00L) >> 8) + (l & 0x00ff00ffL);
      l = (l >> 16)                + (l & 0x0000ffffL);

      count += l;
    }
  return count;
}


// Find the first on or off pixel in a bitmap, and return its coordinates.

flag FBitmapFind(constbitmap b, int *x0, int *y0, bit o)
{
  int x, y;

  for (y = 0; y < MY(b); y++)
    for (x = 0; x < MX(b); x++)
      if (MGCore(b, x, y) == o) {
        *x0 = x; *y0 = y;
        return fTrue;
      }
  return fFalse;
}


// Set or clear all pixels in a bitmap.

void BitmapSet(bitmap b, bit o)
{
  dword l;
  long clBitmap, il;

  l = o ? ~0 : 0;
  clBitmap = CbBitmap(MX(b), MY(b)) >> 2;
  // Set 32 pixels at a time.
  for (il = 0; il < clBitmap; il++)
    *MJ(b, il) = l;
  if (gs.fTraceDot && b == *gs.pbFocus)
    UpdateDisplay();
}


// Invert all pixels in a bitmap.

void BitmapReverse(bitmap b)
{
  long clBitmap, il;

  clBitmap = CbBitmap(MX(b), MY(b)) >> 2;
  // Invert 32 pixels at a time.
  for (il = 0; il < clBitmap; il++)
    *MJ(b, il) ^= ~0;
}


// Set all pixels in a bitmap randomly, with on pixels at a given percentage.

void BitmapRandom(bitmap b, int nPercent)
{
  int x, y;

  for (y = 0; y < MY(b); y++)
    for (x = 0; x < MX(b); x++)
      MP(b, x, y, Rnd(0, 99) < nPercent);
}


// Draw a moire pattern in a bitmap, composed of on or off lines radiating
// from the center of the bitmap to the edges.

void BitmapMoire(bitmap b)
{
  int x0, y0, x, y;

  x0 = MX(b) >> 1; y0 = MY(b) >> 1;
  for (y = 0; y < MY(b); y++) {
    Line(b, x0, y0, 0, y, FOdd(y));
    Line(b, x0, y0, MX(b)-1, y, FOdd(y) == FOdd(MX(b)));
  }
  for (x = 0; x < MX(b); x++) {
    Line(b, x0, y0, x, 0, FOdd(x));
    Line(b, x0, y0, x, MY(b)-1, FOdd(x) == FOdd(MY(b)));
  }
}


// Copy the contents of one bitmap to another bitmap.

flag FBitmapCopy(constbitmap b1, bitmap *b2)
{
  long clBitmap, il;

  if (!FBitmapSizeSet(b2, MX(b1), MY(b1)))
    return fFalse;
  CopyClient(b1, *b2);
  clBitmap = CbBitmap(MX(b1), MY(b1)) >> 2;
  for (il = 0; il < clBitmap; il++)
    *MJ(*b2, il) = *MJ(b1, il);
  return fTrue;
}


// Copy pixels from one bitmap to another bitmap, combining them using the
// specified logical operator table.

void BitmapMoveOrAnd(bitmap b1, constbitmap b2, CONST int *rgf)
{
  int x, y, x1, y1;

  x1 = Min(MX(b1), MX(b2));
  y1 = Min(MY(b1), MY(b2));
  for (y = 0; y < y1; y++)
    for (x = 0; x < x1; x++)
      MP(b1, x, y, rgf[MG(b1, x, y) + MG(b2, x, y)]);
}


// Logical "or" one bitmap into another bitmap.

void BitmapOr(bitmap b1, constbitmap b2)
{
  long clBitmap, il;

  // Fast case: For bitmaps of the same size, combine 32 pixels at a time.
  if (MX(b1) == MX(b2) && MY(b1) == MY(b2)) {
    clBitmap = CbBitmap(MX(b1), MY(b1)) >> 2;
    for (il = 0; il < clBitmap; il++)
      *MJ(b1, il) |= *MJ(b2, il);
    return;
  }

  BitmapMoveOrAnd(b1, b2, rgfOr);
}


// Logical "and" one bitmap into another bitmap.

void BitmapAnd(bitmap b1, constbitmap b2)
{
  long clBitmap, il;

  // Fast case: For bitmaps of the same size, combine 32 pixels at a time.
  if (MX(b1) == MX(b2) && MY(b1) == MY(b2)) {
    clBitmap = CbBitmap(MX(b1), MY(b1)) >> 2;
    for (il = 0; il < clBitmap; il++)
      *MJ(b1, il) &= *MJ(b2, il);
    return;
  }

  BitmapMoveOrAnd(b1, b2, rgfAnd);
}


// Logical "exclusive or" one bitmap into another bitmap.

void BitmapXor(bitmap b1, constbitmap b2)
{
  long clBitmap, il;

  // Fast case: For bitmaps of the same size, combine 32 pixels at a time.
  if (MX(b1) == MX(b2) && MY(b1) == MY(b2)) {
    clBitmap = CbBitmap(MX(b1), MY(b1)) >> 2;
    for (il = 0; il < clBitmap; il++)
      *MJ(b1, il) ^= *MJ(b2, il);
    return;
  }

  BitmapMoveOrAnd(b1, b2, rgfXor);
}


// Blend one bitmap into another bitmap, where half the pixels come from one
// bitmap and half from the other, in a checkerboard pattern.

void BitmapBlend(bitmap b1, constbitmap b2)
{
  int x, y;

  for (y = 0; y < MY(b1); y++)
    for (x = 0; x < MX(b1); x++)
      MP(b1, x, y, Mg(FOdd(x ^ y) ? b2 : b1, x, y));
}


// Copy one bitmap into another bitmap, tessellating or forming a grid of the
// source bitmap pattern within the bounds of the destination bitmap.

void BitmapTessellate(bitmap b1, constbitmap b2)
{
  int x, y;

  for (y = 0; y < MY(b1); y += MY(b2))
    for (x = 0; x < MX(b1); x += MX(b2))
      BlockMove(b2, b1, 0, 0, MX(b2)-1, MY(b2)-1, x, y);
}


// Resize a bitmap to the specified size, then tessellate the old pattern
// within the bounds of the new dimensions.

flag FBitmapTessellateTo(bitmap *b, int x, int y)
{
  bitmap b2;

  if (!FAllocateBitmap(&b2, x, y, *b))
    return fFalse;
  BitmapTessellate(b2, *b);
  DeallocateBitmap(*b);
  *b = b2;
  return fTrue;
}


// Flip the pixels in a bitmap horizontally.

void BitmapFlipX(bitmap b)
{
  int x, y, o;
  long il, ilT;
  dword l;

  if ((MX(b) & 31) == 0) {
    // Fast case: For bitmaps with horizontal dimensions divisible by 32, flip
    // 32 bits at a time.
    for (y = 0; y < MY(b); y++) {
      il = ilT = MI(b, 0, y);
      for (x = 0; x < b->clRow; x++) {
        l = *MJ(b, il);

        // Fast way of flipping the bits in a long. Flip neighboring bits with
        // each other, then groups of four, then eight, etc.
        l = (l & 0xaaaaaaaaL) >> 1 | (l & 0x55555555L) << 1;
        l = (l & 0xccccccccL) >> 2 | (l & 0x33333333L) << 2;
        l = (l & 0xf0f0f0f0L) >> 4 | (l & 0x0f0f0f0fL) << 4;
        l = (l & 0xff00ff00L) >> 8 | (l & 0x00ff00ffL) << 8;
        l = l >> 16                | (l & 0x0000ffffL) << 16;

        *MJ(b, il) = l;
        il++;
      }
      il--;
      while (ilT < il) {
        l = *MJ(b, ilT);
        *MJ(b, ilT) = *MJ(b, il);
        *MJ(b, il) = l;
        ilT++, il--;
      }
    }
  } else {
    // Normal case: Flip one bit at a time.
    for (y = 0; y < MY(b); y++)
      for (x = 0; x < MX(b) >> 1; x++) {
        o = MG(b, x, y);
        MP(b, x, y, MG(b, MX(b)-1-x, y));
        MP(b, MX(b)-1-x, y, o);
      }
  }
}


// Flip the pixels in a bitmap vertically.

void BitmapFlipY(bitmap b)
{
  int x, y, o;

  for (x = 0; x < MX(b); x++)
    for (y = 0; y < MY(b) >> 1; y++) {
      o = MG(b, x, y);
      MP(b, x, y, MG(b, x, MY(b)-1-y));
      MP(b, x, MY(b)-1-y, o);
    }
}


// Transpose the pixels in a bitmap i.e. flip across the main diagonal.

flag FBitmapTranspose(bitmap *pb)
{
  bitmap b = *pb, b2;
  int x, y, o;

  if (MX(b) == MY(b)) {
    // For square bitmaps, just swap the pixels across the main diagonal.
    for (y = 1; y < MY(b); y++)
      for (x = 0; x < y; x++) {
        o = MG(b, x, y);
        MP(b, x, y, MG(b, y, x));
        MP(b, y, x, o);
      }
  } else {
    // For rectangular bitmaps, create a new bitmap of the transposed size.
    if (!FAllocateBitmap(&b2, MY(b), MX(b), b))
      return fFalse;
    for (y = 0; y < MY(b2); y++)
      for (x = 0; x < MX(b2); x++)
        MP(b2, x, y, MG(b, y, x));
    DeallocateBitmap(b);
    *pb = b2;
  }
  return fTrue;
}


// Make a bitmap be a given size, reallocating or allocating in the first
// place if necessary. The contents of the bitmap are undefined.

flag FBitmapSizeSet(bitmap *b, int xnew, int ynew)
{
  bitmap b2;

  if (*b != NULL) {
    // Check if bitmap will have same dimensions, or same number of bytes.
    if (MX(*b) == xnew && MY(*b) == ynew)
      return fTrue;
    if (xnew > xBitmap || ynew > yBitmap) {
      PrintSzNN_E("Can't create bitmap larger than %d by %d!\n",
        xBitmap, yBitmap);
      return fFalse;
    }
    if (CbBitmap(MX(*b), MY(*b)) == CbBitmap(xnew, ynew)) {
      MX(*b) = xnew; MY(*b) = ynew;
      (*b)->clRow = CbBitmapRow(xnew) >> 2;
      return fTrue;
    }
  }
  if (!FAllocateBitmap(&b2, xnew, ynew, *b))
    return fFalse;
  if (*b != NULL)
    DeallocateBitmap(*b);
  *b = b2;
  return fTrue;
}


// Resize a bitmap to be a given size, by either truncating or appending off
// pixels to its right and bottom edges.

flag FBitmapResizeTo(bitmap *b, int xnew, int ynew)
{
  bitmap b2;

  if (MX(*b) == xnew && MY(*b) == ynew)
    return fTrue;
  if (!FAllocateBitmap(&b2, xnew, ynew, *b))
    return fFalse;
  BitmapOff(b2);
  if (!FNull(*b))
    BlockMove(*b, b2, 0, 0, MX(*b)-1, MY(*b)-1, 0, 0);
  DeallocateBitmap(*b);
  *b = b2;
  return fTrue;
}


// Resize a bitmap by a given offset, by either truncating or appending off
// pixels to its left and top edges.

flag FBitmapShiftBy(bitmap *b, int xnew, int ynew)
{
  bitmap b2;
  int x, y;

  if (xnew == 0 && ynew == 0)
    return fTrue;
  if (!FAllocateBitmap(&b2, MX(*b) + xnew, MY(*b) + ynew, *b))
    return fFalse;
  for (y = 0; y < MY(b2); y++)
    for (x = 0; x < MX(b2); x++)
      MP(b2, x, y, Mg(*b, x - xnew, y - ynew));
  DeallocateBitmap(*b);
  *b = b2;
  return fTrue;
}


// Resize a bitmap to be a given size. The old contents of the bitmap are
// stretched or compacted to fit within the new dimensions.

flag FBitmapZoomTo(bitmap *b, int xnew, int ynew, flag fPreserve)
{
  bitmap b2;
  int x, y, x1, y1, x2, y2;
  real deltax, deltay;

  if (!FAllocateBitmap(&b2, xnew, ynew, *b))
    return fFalse;
  BitmapOff(b2);
  if (FNull(*b) || FNull(b2))
    goto LDone;
  deltax = (real)MX(*b) / (real)xnew;
  deltay = (real)MY(*b) / (real)ynew;
  if (fPreserve && deltax >= 1.0 && deltay >= 1.0) {

    // Preserve on pixels when shrinking case: If the bitmap isn't growing in
    // either dimension, then there are always one or more source pixels to
    // map to a destination pixel. Make each destination pixel on if any of
    // the source pixels that map to it are on.
    for (y = 0; y < ynew; y++)
      for (x = 0; x < xnew; x++) {
        y2 = (int)((real)(y+1)*deltay);
        for (y1 = (int)((real)y*deltay); y1 < y2; y1++) {
          x2 = (int)((real)(x+1)*deltax);
          for (x1 = (int)((real)x*deltax); x1 < x2; x1++)
            if (MG(*b, x1, y1)) {
              M1(b2, x, y);
              goto LNext;
            }
        }
LNext:
        ;
      }
  } else {

    // Normal case: Make each destination pixel a copy of the one source pixel
    // that best maps to it.
    for (y = 0; y < ynew; y++)
      for (x = 0; x < xnew; x++)
        if (MG(*b, (int)((real)x*deltax + rMinute),
          (int)((real)y*deltay + rMinute)))
          M1(b2, x, y);
  }
LDone:
  DeallocateBitmap(*b);
  *b = b2;
  return fTrue;
}


// Zoom a bitmap and its contents by the given factors. Each source pixel
// becomes an x by y block in the destination, where x and y can be different
// for even and odd numbered rows and columns.

flag FBitmapBias(bitmap *pb, int xbias1, int xbias2, int ybias1, int ybias2)
{
  bitmap b = *pb, b2;
  int x, y;

  if (!FAllocateBitmap(&b2,
    (MX(b) >> 1)*(xbias1 + xbias2) + FOdd(MX(b))*xbias1,
    (MY(b) >> 1)*(ybias1 + ybias2) + FOdd(MY(b))*ybias1, b))
    return fFalse;
  BitmapOff(b2);
  for (y = 0; y < MY(b); y++)
    for (x = 0; x < MX(b); x++)
      if (MG(b, x, y))
        Block(b2, (x >> 1)*(xbias1 + xbias2) + FOdd(x)*xbias1,
          (y >> 1)*(ybias1 + ybias2) + FOdd(y)*ybias1,
          ((x + 1) >> 1)*(xbias1 + xbias2) + FOdd(x + 1)*xbias1 - 1,
          ((y + 1) >> 1)*(ybias1 + ybias2) + FOdd(y + 1)*ybias1 - 1, fOn);
  DeallocateBitmap(b);
  *pb = b2;
  return fTrue;
}


// Resize a bitmap smaller, by trimming all blank rows and columns from its
// four edges.

flag FBitmapCollapse(bitmap *b)
{
  int xnew, ynew, i;

  // Trim off pixels from the right and bottom edges.
  for (ynew = MY(*b)-1; ynew >= 0; ynew--)
    for (i = 0; i < MX(*b); i++)
      if (MG(*b, i, ynew))
        goto LNext1;
LNext1:
  for (xnew = MX(*b)-1; xnew >= 0; xnew--)
    for (i = 0; i <= ynew; i++)
      if (MG(*b, xnew, i))
        goto LNext2;
LNext2:
  if (!FBitmapResizeTo(b, xnew + 1, ynew + 1))
    return fFalse;

  // Trim off pixels from the left and top edges.
  for (ynew = 0; ynew < MY(*b); ynew++)
    for (i = 0; i < MX(*b); i++)
      if (MG(*b, i, ynew))
        goto LNext3;
LNext3:
  for (xnew = 0; xnew < MX(*b); xnew++)
    for (i = ynew; i < MY(*b); i++)
      if (MG(*b, xnew, i))
        goto LNext4;
LNext4:
  return FBitmapShiftBy(b, -xnew, -ynew);
}


// Zoom a bitmap and its contents 200%. Then smooth the result so it doesn't
// look blocky with just 2x2 blocks everywhere.

long BitmapSmooth(bitmap *pb, flag fAll)
{
  bitmap b = *pb, b2;
  int x, y, x2, y2, grf;
  long count = 0;

  // Zoom the bitmap 200%.
  if (!FAllocateBitmap(&b2, MX(b) << 1, MY(b) << 1, *pb))
    return -1;
  for (y = 0; y < MY(b2); y++)
    for (x = 0; x < MX(b2); x++)
      MP(b2, x, y, MG(b, x >> 1, y >> 1));

  // Fill in certain corners of the zoomed bitmap with on pixels.
  for (y = 0; y < MY(b)-1; y++)
    for (x = 0; x < MX(b)-1; x++) {

      // See what pixels are set in each 2x2 block.
      grf = (MG(b, x, y) << 3) + (MG(b, x+1, y) << 2) +
        (MG(b, x, y+1) << 1) + MG(b, x+1, y+1);
      x2 = (x << 1) + 1; y2 = (y << 1) + 1;

      // If two opposite pixels are set in a 2x2 block, set pixels inside the
      // "stairs" between the two blocks to form a thick diagonal line.
      if (grf == 6) {
        M1(b2, x2, y2); M1(b2, x2+1, y2+1);
        count++;
      } else if (grf == 9) {
        M1(b2, x2+1, y2); M1(b2, x2, y2+1);
        count++;
      } else if (fAll) {

        // If three pixels are set in a 2x2 block, set the pixel in the corner
        // if it looks like the end of a diagonal line, but leave sharp right
        // angle corners alone.
        if (grf == 7 && ((Mg(b, x+2, y-1) && !Mg(b, x+1, y-1)) ||
          (Mg(b, x-1, y+2) && !Mg(b, x-1, y+1)))) {
          M1(b2, x2, y2);
          count++;
        } else if (grf == 11 && ((Mg(b, x-1, y-1) && !Mg(b, x, y-1)) ||
          (Mg(b, x+2, y+2) && !Mg(b, x+2, y+1)))) {
          M1(b2, x2+1, y2);
          count++;
        } else if (grf == 13 && ((Mg(b, x-1, y-1) && !Mg(b, x-1, y)) ||
          (Mg(b, x+2, y+2) && !Mg(b, x+1, y+2)))) {
          M1(b2, x2, y2+1);
          count++;
        } else if (grf == 14 && ((Mg(b, x+2, y-1) && !Mg(b, x+2, y)) ||
          (Mg(b, x-1, y+2) && !Mg(b, x, y+2)))) {
          M1(b2, x2+1, y2+1);
          count++;
        }
      }
    }
  DeallocateBitmap(b);
  *pb = b2;
  return count;
}


// Chisel off the corners of set content in a bitmap, by one pixel width.

long BitmapSmoothCorner(bitmap b, bit o)
{
  bitmap b2 = NULL;
  int x, y, i, j;
  long count = 0;

  if (!FBitmapCopy(b, &b2))
    return -1;
  j = 3 - (o << 1);
  for (y = -1; y <= MY(b); y++)
    for (x = -1; x <= MX(b); x++) {

      // For each 2x2 block in the bitmap, if only one pixel is set
      // appropriately, that means a corner or endpoint, so remove the pixel.
      i = Mg(b2, x, y) + Mg(b2, x+1, y) + Mg(b2, x, y+1) + Mg(b2, x+1, y+1);
      if (i == j) {
        Block(b, x, y, x+1, y+1, !o);
        count++;
      }
    }
  DeallocateBitmap(b2);
  return count;
}


// Thicken all content in a bitmap.

long BitmapThicken(bitmap b)
{
  int x, y;
  long count = 0;

  // For each on pixel, expand it into a 2x2 block of on pixels.
  for (y = MY(b)-1; y >= 0; y--)
    for (x = MX(b)-1; x >= 0; x--)
      if (MG(b, x, y)) {
        Mp(b, x+1, y, fOn); Mp(b, x, y+1, fOn); Mp(b, x+1, y+1, fOn);
        count++;
      }
  return count;
}


// Thicken all content in a bitmap by a specified diameter.

long BitmapThicken2(bitmap b, int z, flag fCorner)
{
  bitmap b2 = NULL, ba;
  char sz[cchSzDef];
  int x, y, x2, y2, h, f, i, iMax, grf;
  long count = 0;

  if (z <= 2)
    return BitmapThicken(b);
  if (!FBitmapCopy(b, &b2))
    return -1;

  // Compose a circle of on pixels to expand each bitmap pixel into.
  if (!FAllocateBitmap(&ba, z, z, b)) {
    DeallocateBitmap(b2);
    return -1;
  }
  BitmapOff(ba);
  Turtle(ba, "S1A0C1O1");
  h = (z - 1) >> 1; f = !FOdd(z);
  sprintf(S(sz), "BM0,%dOA%dBR%dOC%dBD%dOE%dBL%dOG%dBM%d,%d`",
    h, h, f, h, f, h, f, h, h, h);
  Turtle(ba, sz);

  // For each on pixel, expand it into a circle of on pixels.
  iMax = fCorner ? DIRS2 : DIRS;
  for (y = 0; y < MY(b); y++)
    for (x = 0; x < MX(b); x++)
      if (MG(b2, x, y)) {
        grf = 0;
        for (i = 0; i < iMax; i++)
          if (Mg(b2, x + xoff[i], y + yoff[i]))
            grf |= (1 << i);

        if ((grf & NBinary(0,1,0,1,0,1,0,1)) == NBinary(0,1,0,1,0,1,0,1))
          continue;
        for (y2 = 0; y2 < MY(ba); y2++)
          for (x2 = 0; x2 < MX(ba); x2++)
            if (MG(ba, x2, y2))
              Mp(b, x-h+x2, y-h+y2, fOn);

        if (fCorner) {
          if ((grf & NBinary(1,1,1,0,0,0,1,1)) == 0)
            Block(b, x,   y,   x-h,   y-h,   fOn);
          if ((grf & NBinary(1,1,1,1,1,0,0,0)) == 0)
            Block(b, x,   y+f, x-h,   y+h+f, fOn);
          if ((grf & NBinary(0,0,1,1,1,1,1,0)) == 0)
            Block(b, x+f, y+f, x+h+f, y+h+f, fOn);
          if ((grf & NBinary(1,0,0,0,1,1,1,1)) == 0)
            Block(b, x+f, y,   x+h+f, y-h,   fOn);
        }
        count++;
      }

  DeallocateBitmap(ba);
  DeallocateBitmap(b2);
  return count;
}


// Return whether an on pixel can be turned off without significantly
// affecting the topology of the on pixels around it, i.e. don't split shapes,
// create holes, or shorten lines. Used by BitmapThinner().

flag FBitmapThinCore(bitmap b, int x, int y, flag fAll)
{
  flag rgf[DIRS2];
  int c[DIRS], c1 = 0, d;

  // Look at the pixel's eight neighbors. Count the number of on pixels and
  // the number of on in the three pixels closest to each of the four corners.
  for (d = 0; d < DIRS2; d++)
    rgf[d] = Mg(b, x + xoff[d], y + yoff[d]);
  for (d = 0; d < DIRS; d++) {
    c[d] = rgf[d] + rgf[d+1 & DIRS1] + rgf[DIRS + d];
    c1 += rgf[d];
  }

  // If one corner is all on and the opposite corner all off, that means the
  // pixel in question is a corner and can be turned off. The pixel can also
  // be turned off if two adjacent corners are all on, and the edge opposite
  // them is off, meaning a small indentation can be made.
  for (d = 0; d < DIRS; d++)
    if (c[d] == 3 && ((c[d+2 & DIRS1] == 0) ||
      (fAll && c[d+1 & DIRS1] == 3 && c1 == 3))) {
      M0(b, x, y);
      return fTrue;
    }
  return fFalse;
}


// Make all on pixel content in a bitmap as thin as possible, without
// splitting shapes or making anything shorter. After this most if not all
// content will be one pixel wide, with few or no 2x2 blocks anywhere.

long BitmapThinner(bitmap b, flag fCorner)
{
  int x, y, f, d, cd;
  long count = 0, cOld;
  flag rgf[DIRS2];

  // First file down all corners as much as possible, then allow making
  // indentations or cavities within edges.

  for (f = 0; f <= 1; f++) {

    // Check all pixels from upper left to lower right.
    cOld = count;
    for (y = 0; y < MY(b); y++)
      for (x = 0; x < MX(b); x++)
        if (MG(b, x, y))
          count += FBitmapThinCore(b, x, y, f);
    if (count == cOld)
      continue;

    // Check all pixels from lower right to upper left.
    cOld = count;
    for (y = MY(b)-1; y >= 0; y--)
      for (x = MX(b)-1; x >= 0; x--)
        if (MG(b, x, y))
          count += FBitmapThinCore(b, x, y, f);
    if (count == cOld)
      continue;

    // Check all pixels from upper right to lower left.
    cOld = count;
    for (x = MX(b)-1; x >= 0; x--)
      for (y = 0; y < MY(b); y++)
        if (MG(b, x, y))
          count += FBitmapThinCore(b, x, y, f);
    if (count == cOld)
      continue;

    // Check all pixels from lower left to upper right.
    for (x = 0; x < MX(b); x++)
      for (y = MY(b)-1; y >= 0; y--)
        if (MG(b, x, y))
          count += FBitmapThinCore(b, x, y, f);
  }

  // After all possible orthogonal adjustments made, allow removals that only
  // leave pixels connected diagonally. This will break all 2x2 blocks left,
  // except those like an "X" whose four corners are connected diagonally.

  if (fCorner) {
    for (y = 0; y < MY(b); y++)
      for (x = 0; x < MX(b); x++)
        if (MG(b, x, y)) {
          cd = 0;
          for (d = 0; d < DIRS2; d++) {
            f = Mg(b, x + xoff[d], y + yoff[d]);
            rgf[d] = f;
            if (f)
              cd++;
          }
          if (cd < 3)
            continue;

          // Only delete this pixel if it won't cause a separation or a hole.
          // For each surrounding pixel, check if it's still connected to a
          // pixel counterclockwise. To delete there must be exactly one such
          // break. Less would make a hole, and more would make a separation.

          cd = 0;
          for (d = 0; d < DIRS2; d++)
            if (rgf[d] && (d < DIRS ? !rgf[d + 1 & DIRS1] && !rgf[DIRS + d] :
              !rgf[d + 1 & DIRS1]))
              cd++;
          if (cd == 1) {
            count++;
            M0(b, x, y);
          }
        }
   }
  return count;
}


// Edit pixels in a bitmap so they're a single convex shape, with no concave
// inlets, as if a rubber band were wrapped around the on pixels.

flag FBitmapConvex(bitmap b, int x1, int y1, int x2, int y2)
{
  bitmap b2 = NULL;
  int x, y, m, n, m2, n2;
  flag fRet = fFalse;

  // Determine the minimum bounding box containing on pixels.
  for (; x1 <= x2; x1++)
    for (y = y1; y <= y2; y++)
      if (MG(b, x1, y))
        goto LNext1;
  goto LDone;
LNext1:
  for (; x2 >= x1; x2--)
    for (y = y1; y <= y2; y++)
      if (MG(b, x2, y))
        goto LNext2;
  goto LDone;
LNext2:
  for (; y1 <= y2; y1++)
    for (x = x1; x <= x2; x++)
      if (MG(b, x, y1))
        goto LNext3;
  goto LDone;
LNext3:
  for (; y2 >= y1; y2--)
    for (x = x1; x <= x2; x++)
      if (MG(b, x, y2))
        goto LNext4;
  goto LDone;
LNext4:

  // Start with a solid block.
  if (!FBitmapCopy(b, &b2))
    goto LDone;
  Block(b, x1, y1, x2, y2, fOn);

  // Loop over the four quadrants in the bitmap.
  for (y = -1; y <= 1; y += 2)
    for (x = -1; x <= 1; x += 2) {
      m2 = x < 0 ? x1 : x2;
      n2 = y < 0 ? y1 : y2;
      for (m = m2, n = n2;; n -= y)
        if (MG(b2, m2, n))
          break;
      n += y;

      // In each quadrant, shave off lines of on pixels from the block, if
      // that line doesn't hit any on pixels in the original bitmap.
      while (n != n2 + y) {
        if (FLineUntil(b2, m2, n, m, n2, NULL, NULL, fOn, fFalse, fFalse))
          n += y;
        else {
          Line(b, m2, n, m, n2, fOff);
          m -= x;
        }
      }
    }
  fRet = fTrue;

LDone:
  if (b2 != NULL)
    DeallocateBitmap(b2);
  return fRet;
}


// Zoom a bitmap by 200% plus one extra pixel. Boundaries or edges between
// on and off pixels in the old bitmap, become lines of on pixels in the new.

flag FBitmapAccentBoundary(bitmap *pb)
{
  bitmap b = *pb, b2;
  int x, y, xnew, ynew;
  bit o0, o1, o2, o3;

  if (!FAllocateBitmap(&b2, (MX(b) << 1) + 1, (MY(b) << 1) + 1, b))
    return fFalse;
  BitmapOff(b2);
  for (y = -1; y < MY(b); y++) {
    ynew = (y + 1) << 1;
    for (x = -1; x < MX(b); x++) {
      xnew = (x + 1) << 1;
      o0 = Mg(b, x, y);   o1 = Mg(b, x+1, y);
      o2 = Mg(b, x, y+1); o3 = Mg(b, x+1, y+1);
      if (o1 != o3 || o2 != o3 || o0 != o3)
        M1(b2, xnew, ynew);
      if (o1 != o3)
        M1(b2, xnew+1, ynew);
      if (o2 != o3)
        M1(b2, xnew, ynew+1);
    }
  }
  DeallocateBitmap(b);
  *pb = b2;
  return fTrue;
}


// Change a bitmap to highlight areas of contrast, i.e. so that each pixel is
// on if there's a neighboring pixel that's a different value from it.

flag FBitmapAccentContrast(bitmap b, flag fCorner)
{
  bitmap b2 = NULL;
  int x, y, d, dMax = DIRS + fCorner*DIRS;
  bit o;

  if (!FBitmapCopy(b, &b2))
    return fFalse;
  BitmapOff(b);
  for (y = 0; y < MY(b); y++)
    for (x = 0; x < MX(b); x++) {
      o = MG(b2, x, y);
      for (d = 0; d < dMax; d++)
        if (Mg(b2, x + xoff[d], y + yoff[d]) != o) {
          M1(b, x, y);
          break;
        }
    }
  DeallocateBitmap(b2);
  return fTrue;
}


// Offset the content in a bitmap by the given number of pixels. This doesn't
// change the bitmap's size or lose any data, where pixels that move off an
// edge of the bitmap reappear on the opposite.

flag FBitmapSlide(bitmap b, int dx, int dy)
{
  bitmap b2 = NULL;

  if (!FBitmapCopy(b, &b2))
    return fFalse;
  while (dx < 0)
    dx += MX(b);
  while (dx >= MX(b))
    dx -= MX(b);
  while (dy < 0)
    dy += MY(b);
  while (dy >= MY(b))
    dy -= MY(b);

  // Copy content in four quadrants from the old bitmap to the new.
  BlockMove(b2, b, dx, dy, MX(b)-1, MY(b)-1, 0, 0);
  if (dx > 0)
    BlockMove(b2, b, 0, dy, dx-1, MY(b)-1, MX(b)-dx, 0);
  if (dy > 0)
    BlockMove(b2, b, dx, 0, MX(b)-1, dy-1, 0, MY(b)-dy);
  if (dx > 0 && dy > 0)
    BlockMove(b2, b, 0, 0, dx-1, dy-1, MX(b)-dx, MY(b)-dy);
  DeallocateBitmap(b2);
  return fTrue;
}


// Duplicate or delete a range of rows or columns in a bitmap, shifting the
// rest of the content appropriately.

flag FBitmapEditRowColumn(bitmap *pb, int z1, int z2, flag fX, flag fAdd)
{
  bitmap b = *pb, b2;
  int x, y, d;

  // Do nothing if either index is outside the bounds of the bitmap.
  if (z1 < 0 || (fX && z1 >= MX(b)) || (!fX && z1 >= MY(b)) ||
    z2 < 0 || (fX && z2 >= MX(b)) || (!fX && z2 >= MY(b)))
    return fFalse;
  SortN(&z1, &z2);

  // Figure out the new size of the bitmap.
  d = z2 - z1 + 1;
  x = MX(b); y = MY(b);
  if (fX)
    x += fAdd ? d : -d;
  else
    y += fAdd ? d : -d;
  if (!FAllocateBitmap(&b2, x, y, b))
    return fFalse;

  // Copy content in two halves from the old bitmap to the new.
  if (fX) {
    BlockMove(b, b2, 0, 0, fAdd ? z2 : z1-1, MY(b)-1, 0, 0);
    BlockMove(b, b2, fAdd ? z1 : z2+1, 0, MX(b)-1, MY(b)-1,
      fAdd ? z2+1 : z1, 0);
  } else {
    BlockMove(b, b2, 0, 0, MX(b)-1, fAdd ? z2 : z1-1, 0, 0);
    BlockMove(b, b2, 0, fAdd ? z1 : z2+1, MX(b)-1, MY(b)-1,
      0, fAdd ? z2+1 : z1);
  }
  DeallocateBitmap(b);
  *pb = b2;
  return fTrue;
}


// Highlight pixels within a given radius, that are definitely blocked by on
// pixels, when looking in all directions from a central point.

void BitmapSeen(bitmap b, int x0, int y0, int dr)
{
  int rSq, x, y, z, d, i, j, dx, dy, xAbs, yAbs, xInc, yInc;
  flag fCircle = dr < 0;

  if (fCircle) {
    neg(dr);
    rSq = Sq(dr);
  }

  // Loop in widening squares from the central point
  for (z = 1; z <= dr; z++) {
    x = x0 - z; y = y0 - z;
    d = DIRS1;
    for (i = 0; i < DIRS; i++) {

      // Loop along each edge of the current square
      for (j = 0; j < (z << 1); j++, x += xoff[d], y += yoff[d]) {
        if (!FLegal(b, x, y) ||
          (fCircle && Sq(x - x0) + Sq(y - y0) > rSq) ||
          MG(b, x, y))
          continue;
        dx = x - x0; dy = y - y0;
        xAbs = NAbs(dx); yAbs = NAbs(dy);
        xInc = NSgn(dx); yInc = NSgn(dy);

        // Check whether an off pixel is behind on pixel(s)
        if (Mg(b, x - xInc, y - yInc) &&
          (xInc == 0 || yInc == 0 || xAbs == yAbs ||
          Mg(b, x - xInc*(xAbs > yAbs), y - yInc*(yAbs > xAbs))))
          M1(b, x, y);
      }
      DirDec(d);
    }
  }
}


// Create a simple random dot stereogram in a bitmap, replacing a two level
// image in that bitmap. A second bitmap may be used for the random dots.

flag FStereogram(bitmap b, constbitmap bTexture, int width, int height)
{
  bitmap b2 = NULL;
  int x, y, h, w, wDrop = NAbs(height);
  flag fDeep = height < 0, fTexture, f;

  Assert(height < width);
  if (!FBitmapCopy(b, &b2))
    return fFalse;
  BitmapOff(b);
  fTexture = bTexture != NULL &&
    MX(bTexture) == MX(b) && MY(bTexture) == MY(b);

  for (y = 0; y < MY(b); y++) {
    for (x = 0; x < MX(b); x++) {
      f = MGCore(b2, x, y);
      h = f ? height : 0;
      w = width - h;
      if (x - w < 0 || (f == fDeep && MG(b2, x - wDrop, y) == !fDeep)) {
        if (fTexture ? MG(bTexture, x, y) : Rnd(0, 1))
          M1(b, x, y);
      } else {
        if (MG(b, x - w, y))
          M1(b, x, y);
      }
    }
  }
  DeallocateBitmap(b2);
  return fTrue;
}


// Treat the bitmap as a board of the "Life" cellular automaton, and create
// the next generation.

long LifeGenerate(bitmap b, flag fTorus)
{
  bitmap b2 = NULL;
  int x, y, i, n, xT, yT;
  long count = 0;

  if (!FBitmapCopy(b, &b2))
    return -1;
  for (y = 0; y < MY(b); y++)
    for (x = 0; x < MX(b); x++) {

      // Count the number of neighboring live cells.
      n = 0;
      if (!fTorus) {
        for (i = 0; i < DIRS2; i++)
          n += MgFast(b2, x + xoff[i], y + yoff[i]);
      } else {
        for (i = 0; i < DIRS2; i++) {
          xT = x + xoff[i]; yT = y + yoff[i];
          if (xT < 0)
            xT = MX(b)-1;
          else if (xT >= MX(b))
            xT = 0;
          if (yT < 0)
            yT = MY(b)-1;
          else if (yT >= MY(b))
            yT = 0;
          n += MG(b2, xT, yT);
        }
      }

      if (MG(b2, x, y)) {
        if (gs.grfLifeDie & (1 << n)) {
          // A cell dies.
          M0(b, x, y);
          count++;
        }
      } else {
        if (gs.grfLifeBorn & (1 << n)) {
          // A new cell is born.
          M1(b, x, y);
          count++;
        }
      }
    }
  DeallocateBitmap(b2);
  return count;
}


// Parse a "Life" rulestring such as "23/3", where the first sequence is the
// numbers of neighbors allowed to survive, and the second is neighbors needed
// for a new cell to be born. Implements the SetLife operation.

flag FSetLife(CONST char *sz, int cch)
{
  int grfDie = 0x1FF, grfBorn = 0, grf;
  flag fBorn = fFalse;
  char ch;

  while (cch-- > 0) {
    ch = *sz++;
    if (ch == '/')
      fBorn = fTrue;
    else if (!FBetween(ch, '0', '8'))
      return fFalse;
    else {
      grf = 1 << (ch - '0');
      if (fBorn)
        grfBorn |= grf;
      else
        grfDie &= ~grf;
    }
  }
  gs.grfLifeDie  = grfDie;
  gs.grfLifeBorn = grfBorn;
  return fTrue;
}


/*
******************************************************************************
** Bitmap File Routines
******************************************************************************
*/

// Read header information about a Windows bitmap from a file. Return its
// size, bits per pixel, and colors in the palette if any.

flag FReadBitmapHeader(FILE *file, flag fNoHeader,
  int *x, int *y, int *cf, int *ck)
{
  char ch, ch2;
  long l;
  int cbExtra, i;

  // BitmapFileHeader
  if (!fNoHeader) {
    ch = getbyte(); ch2 = getbyte();
    if (ch != 'B' || ch2 != 'M') {
      PrintSz_W("This file does not look like a Windows bitmap.\n");
      return fFalse;
    }
    skiplong();
    skipword(); skipword();
    skiplong();
  }
  // BitmapInfo / BitmapInfoHeader
  cbExtra = (int)getlong();
  *x = (int)getlong(); *y = (int)getlong();
  skipword(); *cf = getword();
  l = getlong();
  if (l != 0/*BI_RGB*/ && l != 3/*BI_BITFIELDS*/) {
    PrintSz_W("This Windows bitmap file can't be uncompressed.\n");
    return fFalse;
  }
  for (i = 0; i < 3; i++) {
    skiplong();
  }
  *ck = (int)getlong(); skiplong();
  for (cbExtra -= 40; cbExtra > 0; cbExtra--)
    skipbyte();
  if (l == 3/*BI_BITFIELDS*/)
    for (i = 0; i < 3; i++) {
      skiplong();
    }
  return fTrue;
}


// Load a monochrome bitmap from a Windows bitmap format file. It is assumed
// FReadBitmapHeader() has already been called.

flag FReadBitmapCore(bitmap *pb, FILE *file, int x, int y)
{
  bitmap b;
  dword il;
  char ch;

  // RgbQuad
  skiplong(); skiplong();
  // Data
  if (!FBitmapSizeSet(pb, x, y))
    return fFalse;
  b = *pb;
  for (y = MY(b)-1; y >= 0; y--) {
    for (x = 0; x < MX(b); x += 32) {
      il = MI(b, x, y);
      *MJ(b, il) = getlong();
    }
  }
  return fTrue;
}


// Save a monochrome bitmap to a Windows bitmap format file.

void WriteBitmap(constbitmap b, FILE *file, KV kv0, KV kv1)
{
  int x, y;
  dword dw;
  byte *pb = (byte *)&dw;

  // BitmapFileHeader
  putbyte('B'); putbyte('M');
  dw = 14+40 + 8 + 4*MY(b)*(((MX(b)-1) >> 5) + 1);
  putlong(dw);
  putword(0); putword(0);
  putlong(14+40 + 8);
  // BitmapInfo / BitmapInfoHeader
  putlong(40);
  putlong(MX(b)); putlong(MY(b));
  putword(1); putword(1);
  putlong(0 /*BI_RGB*/); putlong(0);
  putlong(0); putlong(0);
  putlong(0); putlong(0);
  // RgbQuad
  putbyte(RgbB(kv0)); putbyte(RgbG(kv0)); putbyte(RgbR(kv0)); putbyte(0);
  putbyte(RgbB(kv1)); putbyte(RgbG(kv1)); putbyte(RgbR(kv1)); putbyte(0);
  // Data
  for (y = MY(b)-1; y >= 0; y--)
    for (x = 0; x < MX(b); x += 32) {
      dw = MA(b, x, y);
      if (x + 32 > MX(b)) {
        SwapN(pb[0], pb[3]); SwapN(pb[1], pb[2]);
        dw &= ~((1 << (x + 32 - MX(b))) - 1);
        SwapN(pb[0], pb[3]); SwapN(pb[1], pb[2]);
      }
      putlong(dw);
    }
}


// Load a plain text file, into a new monochrome bitmap. Each character maps
// to one pixel.

flag FReadText(bitmap *pb, FILE *file)
{
  bitmap b;
  int x, xmax = 0, y = 0;
  char ch;

  // Figure out the number of rows and the longest row in the file.
  loop {
    ch = getbyte();
    if (ch == EOF || ch == chNull)
      break;
    x = 0;
    loop {
      if (ch < ' ') {
        skiplf();
        break;
      }
      x++;
      ch = getbyte();
    }
    if (x > xmax)
      xmax = x;
    y++;
  }
  if (!FBitmapSizeSet(pb, xmax, y))
    return fFalse;
  b = *pb;
  BitmapOff(b);
  fseek(file, 0, SEEK_SET);

  // Actually read in the content of the text file.
  y = 0;
  loop {
    ch = getbyte();
    if (ch == EOF || ch == chNull)
      break;
    x = 0;
    loop {
      if (ch < ' ') {
        skiplf();
        break;
      }
      if (ch >= chOn)
        M1(b, x, y);
      x++;
      ch = getbyte();
    }
    y++;
  }
  return fTrue;
}


// Save a monochrome bitmap to a plain text file. Use one character per pixel.

void WriteText(constbitmap b, FILE *file, flag fClip, flag fThin, flag fTab)
{
  int x, y, x2 = MX(b)-1, grf;

  for (y = 0; y < MY(b); y++) {
    if (fClip)
      for (x2 = MX(b)-1; x2 >= 0 && !MG(b, x2, y); x2--)
        ;
    for (x = 0; x <= x2; x++) {
      if (MG(b, x, y)) {
        if (fThin) {
          grf = Mg(b, x, y-1) | Mg(b, x-1, y) << 1 |
            Mg(b, x, y+1) << 2 | Mg(b, x+1, y) << 3;
          if (grf == 10)
            putbyte(chX);
          else if (grf == 5)
            putbyte(chY);
          else
            putbyte(chXY);
        } else
          putbyte(chOn);
      } else {
        if (!fClip || !fTab)
          putbyte(chOff);
      }
      if (fTab && x < x2)
        putbyte(chTab);
    }
    putbyte('\n');
  }
}


// Load an X11 bitmap from a file, into a new monochrome bitmap.

flag FReadXbm(bitmap *pb, FILE *file)
{
  bitmap b;
  char line[cchSzDef], ch;
  flag fSuper;
  int x, y, i, j;
  dword value;

  do {
    fscanf(file, "%s", S(line));
  } while (line[0] != '#');
  fscanf(file,   "%s%d",          S(line), &x);
  fscanf(file, "%s%s%d", S(line), S(line), &y);
  if (!FBitmapSizeSet(pb, x, y))
    return fFalse;
  b = *pb;
  fscanf(file, "%s", S(line));
  fscanf(file, "%s", S(line));
  fSuper = FEqSzI(line, "short");
  do {
    fscanf(file, "%s", S(line));
  } while (line[0] != '{');
  for (y = 0; y < MY(b); y++) {
    x = 0;
    while (x < MX(b)) {
      while ((ch = getbyte()) != 'x')
        ;
      ch = getbyte(); value = NHex(ch) << 4;
      ch = getbyte(); value += NHex(ch);
      if (fSuper) {
        ch = getbyte(); value = ((value << 4) + NHex(ch)) << 4;
        ch = getbyte(); value += NHex(ch);
      }
      j = fSuper ? 16 : 8;
      if (x + j > MX(b))
        j = MX(b) - x;
      for (i = 0; i < j; i++) {
        MP(b, x + i, y, (bit)((value & 1) ^ 1));
        value = value >> 1;
      }
      x += (fSuper ? 16 : 8);
    }
  }
  while ((ch = getbyte()) >= ' ')
    ;
  return fTrue;
}


// Save a monochrome bitmap to an X11 bitmap text file.

void WriteXbm(constbitmap b, FILE *file, char *szName, char mode)
{
  int x, y, i, temp = 0;
  dword value;
  char *pchStart, *pchEnd, chSav;

  for (pchEnd = szName; *pchEnd != chNull; pchEnd++)
    ;
  for (pchStart = pchEnd; pchStart > szName && *pchStart != '\\'; pchStart--)
    ;
  if (*pchStart == '\\')
    pchStart++;
  for (pchEnd = pchStart; *pchEnd != chNull && *pchEnd != '.'; pchEnd++)
    ;
  chSav = *pchEnd; *pchEnd = chNull;
  fprintf(file, "#define %s_width %d\n" , pchStart, MX(b));
  fprintf(file, "#define %s_height %d\n", pchStart, MY(b));
  fprintf(file, "static %s %s_bits[] = {",
    mode != 'S' ? "char" : "short", pchStart);
  *pchEnd = chSav;

  for (y = 0; y < MY(b); y++) {
    x = 0;
    do {
      if (y + x > 0)
        fprintf(file, ",");
      if (temp == 0)
        fprintf(file, "\n%s",
          mode == 'N' ? "  " : (mode == 'C' ? " " : ""));
      value = 0;
      for (i = (mode != 'S' ? 7 : 15); i >= 0; i--)
        value = (value << 1) + (!MG(b, x + i, y) && x + i < MX(b));
      if (mode == 'N')
        putbyte(' ');
      fprintf(file, "0x");
      if (mode == 'S')
        fprintf(file, "%c%c",
          ChHex(value >> 12), ChHex((value >> 8) & 15));
      fprintf(file, "%c%c",
        ChHex((value >> 4) & 15), ChHex(value & 15));
      temp++;
      if ((mode == 'N' && temp >= 12) || (mode == 'C' && temp >= 15) ||
        (mode == 'S' && temp >= 11))
        temp = 0;
      x += (mode != 'S' ? 8 : 16);
    } while (x < MX(b));
  }
  fprintf(file, "};\n");
}


// Load a Daedalus monochrome bitmap from a file, into a new bitmap.

flag FReadDaedalusBitmapCore(bitmap *pb, FILE *file, int x, int y)
{
  bitmap b;
  int xT, nCur;
  char ch;

  if (!FBitmapSizeSet(pb, x, y))
    return fFalse;
  b = *pb;
  BitmapOff(b);
  skipcrlf();
  for (y = 0; y < MY(b); y++) {
    x = 0;
    loop {
      ch = getbyte();
      if (ch < ' ') {
        skiplf();
        break;
      }
      if (ch < 'a') {
        // This character encodes 6 pixels.
        nCur = (int)(ch - '!');
        for (xT = 5; xT >= 0; xT--) {
          if (FOdd(nCur))
            M1(b, x + xT, y);
          nCur >>= 1;
        }
        x += 6;
      } else {
        // Check for the duplicate row marker '|'.
        if (ch == '|') {
          Assert(x == 0 && y > 0);
          for (nCur = 0, ch = getbyte(); FDigitCh(ch); ch = getbyte())
            nCur = nCur * 10 + (ch - '0');
          loop {
            BlockMove(b, b, 0, y-1, MX(b)-1, y-1, 0, y);
            if (--nCur <= 0)
              break;
            y++;
          }
          skipcrlf();
          break;
        }
        // This character contains some multiple of 6 on or off pixels.
        if (ch >= 'n') {
          LineX(b, x, x + (ch - 'n' + 2) * 6 - 1, y, fOn);
          x += (ch - 'n' + 2) * 6;
        } else
          x += (ch - 'a' + 2) * 6;
      }
    }
  }
  return fTrue;
}


// Save a bitmap to a Daedalus monochrome bitmap custom format file.

void WriteDaedalusBitmap(constbitmap b, FILE *file, flag fClip)
{
  int x, y, xT, xmax, nCur, nSav, cn, cRow = 0;
  flag fBlank;
  byte bT;

  fprintf(file, "DB\n%d %d 1\n\n", MX(b), MY(b));
  for (y = 0; y <= MY(b); y++) {

    // Check if current row same as previous. If so write '|' dup marker.
    if (y > 0) {
      if (y < MY(b)) {
        for (x = 0; x < MX(b); x++)
          if (MG(b, x, y) != MG(b, x, y-1))
            goto LNormal;
        cRow++;
        continue;
      }
LNormal:
      // Output number of times row was duplicated
      if (cRow > 0) {
        if (!fBlank || cRow > 1) {
          putbyte('|');
          if (cRow > 1)
            fprintf(file, "%d", cRow);
        }
        putbyte('\n');
        cRow = 0;
      }
    }
    if (y >= MY(b))
      continue;

    // Output a unique row.
    xmax = MX(b) - 1;
    if (fClip)
      while (xmax >= 0 && !MG(b, xmax, y))
        xmax--;
    fBlank = (xmax < 0);
    cn = 0;
    for (x = 0; x <= xmax + 6; x += 6) {
      // Compute a number encoding the next 6 pixels in this row.
      nCur = 0;
      for (xT = 0; xT < 6; xT++)
        nCur = (nCur << 1) + (x + xT < MX(b) && MG(b, x + xT, y));
      // Check for whether ready to write next character/run to file.
      if (x > 0 && (nCur != nSav ||
        (nCur != 0 && nCur != 63) || cn >= 14 || x > xmax)) {
        if (cn <= 1)
          putbyte('!' + nSav);
        else {
          Assert(nSav == 0 || nSav == 63);
          bT = nSav == 0 ? 'a' - 2 + cn : 'n' - 2 + cn;
          putbyte(bT);
        }
        cn = 0;
      }
      nSav = nCur;
      cn++;
    }
    putbyte('\n');
  }
}

/* graphics.cpp */
