/*
** Daedalus (Version 2.5) File: inside.cpp
** By Walter D. Pullen, Astara@msn.com, http://www.astrolog.org/labyrnth.htm
**
** IMPORTANT NOTICE: Daedalus and all Maze generation and general
** graphics routines used in this program are Copyright (C) 1998-2013 by
** Walter D. Pullen. Permission is granted to freely use, modify, and
** distribute these routines provided these credits and notices remain
** unmodified with any altered or distributed versions of the program.
** The user does have all rights to Mazes and other graphic output
** they make in Daedalus, like a novel created in a word processor.
**
** More formally: This program is free software; you can redistribute it
** and/or modify it under the terms of the GNU General Public License as
** published by the Free Software Foundation; either version 2 of the
** License, or (at your option) any later version. This program is
** distributed in the hope that it will be useful and inspiring, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** General Public License for more details, a copy of which is in the
** LICENSE.HTM included with Daedalus, and at http://www.gnu.org
**
** This file contains routines to draw 3D first person perspective views
** from inside a Maze's passages.
**
** Created: 6/2/2001.
** Last code change: 10/10/2013.
*/

#include <stdio.h>
#include <math.h>

#include "resource.h"
#include "util.h"
#include "graphics.h"
#include "color.h"
#include "threed.h"
#include "maze.h"
#include "draw.h"
#include "daedalus.h"


// Max number of semitransparent walls in line with each other.
#define itransMax 1000

#define iScale 4
#define iStep 16
#define iRound ((iStep) - 1)
#define IO(y) ((y) >> (iScale))
#define IP(y) ((y) << (iScale))

#define iVarScale2 3
#define iVar2 (1 << (iVarScale + iVarScale2))
#define iVarHalf ((iVar) << (iVarScale2 - 1))
#define iScale2 (iVarScale - 1 + iVarScale2)


/*
******************************************************************************
** Inside Display Routines
******************************************************************************
*/

// Convert a high-res inside coordinate to an low-res array coordinate.

int AFromInside(long z)
{
  long i;

  if (dr.fNarrow) {
    if (z >= 0)
      i = z / (dr.zCell + dr.zCellNarrow);
    else
      i = (z + 1) / (dr.zCell + dr.zCellNarrow) - 1;
    return (int)(i * 2 +
      (z - i * (dr.zCell + dr.zCellNarrow) >= dr.zCellNarrow));
  }
  return (int)(z / dr.zCell - (z < 0));
}


// Convert a low-res array coordinate to a high-res inside coordinate, such
// that one's positioned in the middle of a cell.

long InsideFromA(int z)
{
  if (dr.fNarrow)
    return (z - (z < 0)) / 2 * (dr.zCell + dr.zCellNarrow) +
      (FOdd(z) ? dr.zCellNarrow + (dr.zCell >> 1) : (dr.zCellNarrow >> 1));
  return z * dr.zCell + (dr.zCell >> 1);
}


// Convert a low-res array coordinate to a high-res inside coordinate, such
// that one's positioned at the top left corner of a cell.

long InsideFromA0(int z)
{
  if (dr.fNarrow)
    return (z - (z < 0)) / 2 * (dr.zCell + dr.zCellNarrow) +
      (FOdd(z) * dr.zCellNarrow);
  return z * dr.zCell;
}


// Draw a vertical line. Walls, floor markings, and mountains are all
// generated by drawing a sequence of vertical lines of appropriate heights.

#define LineI(x, y1, y2, kv) LineYTrans(c, x, y1, y2, kv, 0)


// Draw a horizontal line. Gradient backgrounds and clouds are all generated
// by drawing a sequence of horizontal lines at appropriate locations.

#define LineH(x1, x2, y, kv) LineXC(c, x1, x2, y, kv)


// Draw a gradient color vertical line on a color bitmap. Used for drawing
// background mountains in inside display modes.

void LineYGradient(colmap c, int x, int y1, int y2, KV kv1, KV kv2)
{
  byte *pb;
  int yClip1 = y1, yClip2 = y2, db, y,
    nR1, nG1, nB1, nR2, nG2, nB2, nRT, nGT, nBT, dR, dG, dB;

  // Don't worry about pixels that are off the bitmap.
  if (yClip1 < 0)
    yClip1 = 0;
  if (yClip2 > CY(c))
    yClip2 = CY(c);
  if (yClip2 <= yClip1)
    return;

  // Quickly draw the line, where things like the memory address only need to
  // be calculated once.
  nR1 = RgbR(kv1); nG1 = RgbG(kv1); nB1 = RgbB(kv1);
  pb = CJ(c, CI(c, x, yClip1));
  db = c->clRow << 2;
  if (kv2 < 0) {

    // If no gradient, just draw a solid color line.
    for (y = yClip1; y < yClip2; y++) {
      CL(pb, nR1, nG1, nB1);
      pb += db;
    }
  } else {

    // If gradient, gradually blend color from color #1 to color #2.
    nR2 = RgbR(kv2); nG2 = RgbG(kv2); nB2 = RgbB(kv2);
    nRT = nR1 << 16; nGT = nG1 << 16; nBT = nB1 << 16;
    dR = ((nR2 - nR1) << 16) / (y2 - y1);
    dG = ((nG2 - nG1) << 16) / (y2 - y1);
    dB = ((nB2 - nB1) << 16) / (y2 - y1);
    for (y = yClip1; y < yClip2; y++) {
      CL(pb, nRT >> 16, nGT >> 16, nBT >> 16);
      nRT += dR; nGT += dG; nBT += dB;
      pb += db;
    }
  }
}


// Draw a solid color vertical line on a color bitmap. Used for drawing walls
// in inside display modes.

void LineYTrans(colmap c, int x, int y1, int y2, KV kv, int nTrans)
{
  byte *pb;
  int db, y, nR, nG, nB, nRT, nGT, nBT;

  // Don't worry about pixels that are off the bitmap.
  if (y1 < 0)
    y1 = 0;
  if (y2 > CY(c))
    y2 = CY(c);

  // Quickly draw the line, where things like the memory address only need to
  // be calculated once.
  nR = RgbR(kv); nG = RgbG(kv); nB = RgbB(kv);
  pb = CJ(c, CI(c, x, y1));
  db = c->clRow << 2;
  if (nTrans == 0) {

    // If no semitransparency, just draw a solid color line.
    for (y = y1; y < y2; y++) {
      CL(pb, nR, nG, nB);
      pb += db;
    }
  } else {

    // If semitransparent, blend color with what's already in place.
    nTrans = nTrans * 128 / 100;
    for (y = y1; y < y2; y++) {
      CM(pb, nRT, nGT, nBT);
      nRT = nR + ((nRT - nR) * nTrans >> 7);
      nGT = nG + ((nGT - nG) * nTrans >> 7);
      nBT = nB + ((nBT - nB) * nTrans >> 7);
      CL(pb, nRT, nGT, nBT);
      pb += db;
    }
  }
}


// Draw a texture mapped vertical line on a color bitmap. Used for drawing
// walls in inside display modes.

void LineYTexture(
  colmap cDst, int xDst, int yDst1, int yDst2, int yClip1, int yClip2,
  colmap cSrc, int xSrc, int ySrc1, int ySrc2,
  int nTrans, KV kvFog, int nFog)
{
  int dyDst = yDst2 - yDst1, dySrc = ySrc2 - ySrc1, dbDst, dbSrc, y1 = yClip1,
    y2 = yClip2, y, yT, dy, nRF, nGF, nBF, nR, nG, nB, nRT, nGT, nBT;
  byte *pbDst, *pbSrc, *pbBase;

  // Don't worry about pixels that are off the bitmap.
  if (y1 < 0)
    y1 = 0;
  if (y2 > CY(cDst))
    y2 = CY(cDst);

  pbDst = CJ(cDst, CI(cDst, xDst, y1));
  pbBase = CJ(cSrc, CI(cSrc, xSrc, 0));
  dbDst = cDst->clRow << 2;
  dbSrc = cSrc->clRow << 2;
  if (dyDst == 0)
    dyDst = 1;
  yT = (y1 - yDst1) * dySrc;
  if ((yT << 10) >> 10 != yT)
    return;
  yT = ((ySrc1 << 16) + (yT << 10) / dyDst) << 6;
  if (yT < 0)
    return;
  dy = (dySrc << 16) / dyDst;
  if (nTrans == 0) {
    if (nFog <= 0) {

      // Draw a solid texture mapped line, without fog.
      for (y = y1; y < y2; y++) {
        pbSrc = pbBase + (yT >> 16) * dbSrc;
        CM(pbSrc, nR, nG, nB);
        CL(pbDst, nR, nG, nB);
        yT += dy;
        pbDst += dbDst;
      }
    } else {

      // Draw a solid texture mapped line, with fog.
      nFog = nFog * 128 / 100;
      nRF = RgbR(kvFog); nGF = RgbG(kvFog); nBF = RgbB(kvFog);
      for (y = y1; y < y2; y++) {
        pbSrc = pbBase + (yT >> 16) * dbSrc;
        CM(pbSrc, nR, nG, nB);
        nR += ((nRF - nR) * nFog >> 7);
        nG += ((nGF - nG) * nFog >> 7);
        nB += ((nBF - nB) * nFog >> 7);
        CL(pbDst, nR, nG, nB);
        yT += dy;
        pbDst += dbDst;
      }
    }
  } else {
    nTrans = nTrans * 128 / 100;
    if (nFog <= 0) {

      // Draw a semitransparent texture mapped line, without fog.
      for (y = y1; y < y2; y++) {
        pbSrc = pbBase + (yT >> 16) * dbSrc;
        CM(pbSrc, nR, nG, nB);
        CM(pbDst, nRT, nGT, nBT);
        nR += ((nRT - nR) * nTrans >> 7);
        nG += ((nGT - nG) * nTrans >> 7);
        nB += ((nBT - nB) * nTrans >> 7);
        CL(pbDst, nR, nG, nB);
        yT += dy;
        pbDst += dbDst;
      }
    } else {

      // Draw a semitransparent texture mapped line, with fog.
      nFog = nFog * 128 / 100;
      nRF = RgbR(kvFog); nGF = RgbG(kvFog); nBF = RgbB(kvFog);
      for (y = y1; y < y2; y++) {
        pbSrc = pbBase + (yT >> 16) * dbSrc;
        CM(pbSrc, nR, nG, nB);
        nR += ((nRF - nR) * nFog >> 7);
        nG += ((nGF - nG) * nFog >> 7);
        nB += ((nBF - nB) * nFog >> 7);
        CM(pbDst, nRT, nGT, nBT);
        nR += ((nRT - nR) * nTrans >> 7);
        nG += ((nGT - nG) * nTrans >> 7);
        nB += ((nBT - nB) * nTrans >> 7);
        CL(pbDst, nR, nG, nB);
        yT += dy;
        pbDst += dbDst;
      }
    }
  }
}


// Draw a texture map overlaying a vertical line on a color bitmap. The
// overlay shape comes from a monochrome bitmap, where optional coloring comes
// from another color bitmap.

void LineYMask(
  colmap cDst, int xDst, int yDst1, int yDst2, int yClip1, int yClip2,
  bitmap b, colmap cSrc, int xSrc, int ySrc1, int ySrc2,
  KV kvFog, int nFog, flag fSet)
{
  int dyDst = yDst2 - yDst1, dySrc = ySrc2 - ySrc1, db,
    y1 = yClip1, y2 = yClip2, y, yT, dy, k, nRF, nGF, nBF, nR, nG, nB;
  byte *pbDst, *pbSrc;

  // Don't worry about pixels that are off the bitmap.
  if (y1 < 0)
    y1 = 0;
  if (y2 > CY(cDst))
    y2 = CY(cDst);

  pbDst = CJ(cDst, CI(cDst, xDst, y1));
  db = cDst->clRow << 2;
  if (dyDst == 0)
    dyDst = 1;
  yT = (y1 - yDst1) * dySrc;
  if ((yT << 10) >> 10 != yT)
    return;
  yT = ((ySrc1 << 16) + (yT << 10) / dyDst) << 6;
  if (yT < 0)
    return;
  dy = (dySrc << 16) / dyDst;
  if (cSrc == NULL) {

    // Draw either black or white pixels on top of what's already present.
    k = fSet * 255;
    if (k > 0 && nFog > 0 && dr.fFogFloor) {
      nFog = nFog * 128 / 100;
      k -= (k * nFog >> 7);
    }
    for (y = y1; y < y2; y++) {
      if (MGCore(b, xSrc, yT >> 16) == fSet) {
        CL(pbDst, k, k, k);
      }
      yT += dy;
      pbDst += db;
    }
  } else {
    if (nFog <= 0 || !dr.fFogFloor) {

      // Draw a color mask from a texture on top of what's already present.
      for (y = y1; y < y2; y++) {
        if (MGCore(b, xSrc, yT >> 16) == fSet) {
          pbSrc = CJ(cSrc, CI(cSrc, xSrc, yT >> 16));
          CM(pbSrc, nR, nG, nB);
          CL(pbDst, nR, nG, nB);
        }
        yT += dy;
        pbDst += db;
      }
    } else {

      // Draw a color mask from a texture bitmap, with fog.
      nFog = nFog * 128 / 100;
      nRF = RgbR(kvFog); nGF = RgbG(kvFog); nBF = RgbB(kvFog);
      for (y = y1; y < y2; y++) {
        if (MGCore(b, xSrc, yT >> 16) == fSet) {
          pbSrc = CJ(cSrc, CI(cSrc, xSrc, yT >> 16));
          CM(pbSrc, nR, nG, nB);
          nR += ((nRF - nR) * nFog >> 7);
          nG += ((nGF - nG) * nFog >> 7);
          nB += ((nBF - nB) * nFog >> 7);
          CL(pbDst, nR, nG, nB);
        }
        yT += dy;
        pbDst += db;
      }
    }
  }
}


// Draw a variable color vertical line on a color bitmap. Used for drawing
// floor markings in inside display modes.

void LineYFloor(colmap c, int x, int yb, int dy1, int dy2,
  KV kv, int nTrans, real rScale)
{
  byte *pb;
  int y1 = yb + dy1, y2 = yb + dy2, db, y, n,
    nRF, nGF, nBF, nR, nG, nB, nRT, nGT, nBT, nRO, nGO, nBO;
  real rFogMax = dy1 >= 0 ? 128.0 : -128.0;

  // Don't worry about pixels that are off the bitmap.
  if (y1 < 0)
    y1 = 0;
  if (y2 > CY(c))
    y2 = CY(c);

  // Quickly draw the line, where things like the memory address only need to
  // be calculated once.
  nR = RgbR(kv); nG = RgbG(kv); nB = RgbB(kv);
  pb = CJ(c, CI(c, x, y1));
  db = c->clRow << 2;
  if (nTrans == 0) {
    if (dr.nFog <= 0 || !dr.fFogFloor) {

      // If no fog, just draw a solid color line.
      for (y = y1; y < y2; y++) {
        CL(pb, nR, nG, nB);
        pb += db;
      }
    } else {

      // If fog, blend each pixel with the fog color based on its distance.
      nRF = RgbR(dr.kvInFog); nGF = RgbG(dr.kvInFog); nBF = RgbB(dr.kvInFog);
      rScale = rScale / (real)(dr.nFog * 10) * rFogMax;
      for (y = y1; y < y2; y++) {
        n = IO((int)(rScale / (real)(y - yb)));
        if (n > 128)
          n = 128;
        nRT = nR + ((nRF - nR) * n >> 7);
        nGT = nG + ((nGF - nG) * n >> 7);
        nBT = nB + ((nBF - nB) * n >> 7);
        CL(pb, nRT, nGT, nBT);
        pb += db;
      }
    }
  } else {
    nTrans = nTrans * 128 / 100;
    if (dr.nFog <= 0 || !dr.fFogFloor) {

      // If no fog semitransparent, blend color with what's already in place.
      for (y = y1; y < y2; y++) {
        CM(pb, nRT, nGT, nBT);
        nRT = nR + ((nRT - nR) * nTrans >> 7);
        nGT = nG + ((nGT - nG) * nTrans >> 7);
        nBT = nB + ((nBT - nB) * nTrans >> 7);
        CL(pb, nRT, nGT, nBT);
        pb += db;
      }
    } else {

      // If fog semitransparent, blend with the fog color based on distance.
      nRF = RgbR(dr.kvInFog); nGF = RgbG(dr.kvInFog); nBF = RgbB(dr.kvInFog);
      rScale = rScale / (real)(dr.nFog * 10) * rFogMax;
      for (y = y1; y < y2; y++) {
        n = IO((int)(rScale / (real)(y - yb)));
        if (n > 128)
          n = 128;
        nRT = nR + ((nRF - nR) * n >> 7);
        nGT = nG + ((nGF - nG) * n >> 7);
        nBT = nB + ((nBF - nB) * n >> 7);
        CM(pb, nRO, nGO, nBO);
        nRT += ((nRO - nRT) * nTrans >> 7);
        nGT += ((nGO - nGT) * nTrans >> 7);
        nBT += ((nBO - nBT) * nTrans >> 7);
        CL(pb, nRT, nGT, nBT);
        pb += db;
      }
    }
  }
}


// Draw a texture mapped vertical line on a color bitmap. Used for drawing
// floor markings in inside display modes.

void LineYFloorTexture(colmap cDst, int x, int yb, int dy1, int dy2,
  bitmap b, colmap cSrc, real m1, real n1, real m2, real n2,
  flag fSet, real rScale)
{
  byte *pb, *pbT;
  int y1 = yb + IO(dy1), y2 = yb + IO(dy2), y, db, n, xT, yT, xMax, yMax,
    nRF, nGF, nBF, nR, nG, nB, nRT, nGT, nBT;
  real rFogMax, rScaleFog, dm, dn,
    rd1, rd2, rdd, rdy1, rddy, rdy, rdyCur, rdyInc, rT;
  flag fNoFog = cSrc == NULL || dr.nFog <= 0 || !dr.fFogFloor;

  // Don't worry about pixels that are off the bitmap.
  rdy = (real)(y2 - y1);
  if (y1 < 0) {
    y1 = 0;
    rdy1 = -IP(yb);
  } else
    rdy1 = (real)dy1;
  if (y2 > CY(cDst))
    y2 = CY(cDst);
  rddy = (real)(dy2 - dy1);

  rd1 = RDiv(rScale, (real)dy1);
  rd2 = RDiv(rScale, (real)dy2);
  rdd = rd2 - rd1;
  if (cSrc == NULL) {
    xMax = MX(b); yMax = MY(b);
  } else if (b == NULL) {
    xMax = CX(cSrc); yMax = CY(cSrc);
  } else {
    xMax = NMin(MX(b), CX(cSrc)); yMax = NMin(MY(b), CY(cSrc));
  }
  dm = RDiv((m2 - m1), rdd) * (real)xMax;
  dn = RDiv((n2 - n1), rdd) * (real)yMax;
  m1 *= (real)xMax; n1 *= (real)yMax;
  rdyCur = rdy1 > rMinute || rdy1 < -rMinute ? rdy1 :
    (rdy >= 0 ? rMinute : -rMinute);
  rdyInc = rddy / rdy;

  pb = CJ(cDst, CI(cDst, x, y1));
  db = cDst->clRow << 2;
  if (!fNoFog) {
    nRF = RgbR(dr.kvInFog); nGF = RgbG(dr.kvInFog); nBF = RgbB(dr.kvInFog);
    rFogMax = dy1 >= 0 ? 128.0 : -128.0;
    rScaleFog = rScale / (real)(dr.nFog * 10) * rFogMax;
  }
  if (cSrc == NULL)
    nR = nG = nB = fSet * 255;

  // Texture mapping floors is more complicated than walls. For floors the
  // vertical line crosses the texture at an oblique angle, requiring
  // calculating the texture pixel in both x and y. Also floors are viewed at
  // an oblique angle, so the rate moving across the texture changes.

  for (y = y1; y < y2; y++) {
    rT = (rScale / rdyCur) - rd1;
    xT = (int)(m1 + rT * dm);
    yT = (int)(n1 + rT * dn);
    if (xT >= xMax)
      xT = xMax - 1;
    if (yT >= yMax)
      yT = yMax - 1;
    if (b != NULL) {
      if (MGCore(b, xT, yT) != fSet)
        goto LNext;
      if (cSrc != NULL) {
        pbT = CJ(cSrc, CI(cSrc, xT, yT));
        CM(pbT, nR, nG, nB);
      }
    } else {
      pbT = CJ(cSrc, CI(cSrc, xT, yT));
      CM(pbT, nR, nG, nB);
    }
    if (fNoFog) {
      CL(pb, nR, nG, nB);
    } else {
      n = IO((int)(rScaleFog / (real)(y - yb)));
      if (n > 128)
        n = 128;
      nRT = nR + ((nRF - nR) * n >> 7);
      nGT = nG + ((nGF - nG) * n >> 7);
      nBT = nB + ((nBF - nB) * n >> 7);
      CL(pb, nRT, nGT, nBT);
    }
LNext:
    pb += db;
    rdyCur += rdyInc;
  }
}


// Return a randomly chosen color for a star or meteor.

KV KvStarRandom()
{
  int nR, nG, nB, nR0, nR1, nG0, nG1, nB0, nB1;

  // Start with a monochrome brightness
  nR = nG = nB = Rnd(0, 255);

  // Then add in a little coloring if appropriate
  if (dr.lStarColor != 0 && Rnd(0, 98) < (int)(dr.lStarColor / 1000000L)) {
    switch (Rnd(0, 2)) {
    case 0:
      nR0 = (int)(dr.lStarColor / 100000L % 10);
      nR1 = (int)(dr.lStarColor / 10000L  % 10);
      nR += Rnd(0, (nR0 + nR1) * 511 / 18) - (nR0 * 511 / 18);
      EnsureBetween(nR, 0, 255);
      break;
    case 1:
      nG0 = (int)(dr.lStarColor / 1000L % 10);
      nG1 = (int)(dr.lStarColor / 100L  % 10);
      nG += Rnd(0, (nG0 + nG1) * 511 / 18) - (nG0 * 511 / 18);
      EnsureBetween(nG, 0, 255);
      break;
    case 2:
      nB0 = (int)(dr.lStarColor / 10L % 10);
      nB1 = (int)(dr.lStarColor       % 10);
      nB += Rnd(0, (nB0 + nB1) * 511 / 18) - (nB0 * 511 / 18);
      EnsureBetween(nB, 0, 255);
      break;
    }
  }
  return Rgb(nR, nG, nB);
}


// Create and draw the stars for the inside display modes.

flag FCreateInsideStars(colmap c, real rd, int yAdd, flag fPerspective)
{
  int i, x, y;
  real rx, dInside2 = dr.dInside*2.0, rT;

  // If the list of stars hasn't been generated yet, do so first.
  if (dr.rgstar == NULL) {
    dr.rgstar = RgAllocate(istarMax, STAR);
    if (dr.rgstar == NULL)
      return fFalse;
    for (i = 0; i < istarMax; i++) {
      dr.rgstar[i].x = Rnd(0, 16383);
      dr.rgstar[i].y = Rnd(0, 16383);
      dr.rgstar[i].kv = KvStarRandom();
    }
    if (c == NULL)
      return fTrue;
  }

  // Draw the list of stars in their appropriate locations and colors.
  if (fPerspective)
    rT = RTanD(dr.dInside) * 2.0;
  for (i = 0; i < dr.cStar; i++) {
    rx = (real)dr.rgstar[i].x * rDegMax / 16384.0;
    if ((rd + dInside2 < rDegMax) ?
      (rx > rd && rx < rd + dInside2) :
      (rx > rd || rx < rd + dInside2 - rDegMax)) {
      rx -= rd;
      if (rx < 0.0)
        rx += rDegMax;
      if (fPerspective)
        rx = RTanD(rx - dr.dInside) / rT + 0.5;
      else
        rx /= dInside2;
      x = ws.xpClient-1 - (int)((real)ws.xpClient * rx);
      y = yAdd + (int)((real)(ws.ypClient >> (int)!dr.fSkyAll) *
        (real)dr.rgstar[i].y / 16384.0);
      if (fPerspective || CG(c, x, y) == dr.kvInSky)
        Cp(c, x, y, dr.rgstar[i].kv);
    }
  }
  return fTrue;
}


// Draw a solid circle, with an optional circle shaped bite out of it in a
// different color. Called from DrawSunMoon() to draw the Sun or Moon.

void DrawCrescent(colmap c, int x, int y, int r, int x2, int y2,
  KV kv, KV kv2)
{
  int h1 = x-r, v1 = y-r, h2 = x+r, v2 = y+r, h, v;
  long n = Sq(r) + r - 1, l;

  Legalize(c, &h1, &v1);
  Legalize(c, &h2, &v2);
  for (v = v1; v <= v2; v++)
    for (h = h1; h <= h2; h++) {
      l = Sq(v - y) + Sq(h - x);
      if (l <= n) {
        if (kv2 < 0)
          CP(c, h, v, kv);
        else {
          l = Sq(v - y2) + Sq(h - x2);
          CP(c, h, v, l <= n ? kv2 : kv);
        }
      }
    }
}


// Draw the Sun or Moon for the inside display mode.

void DrawSunMoon(colmap c, real rd, int yAdd)
{
  int x, y, xi, yi, zSun, d;
  real rx = dr.dInside, dInside2 = dr.dInside*2.0;

  zSun = ws.ypClient / 20;
  if (dr.rLightFactor >= 0.0)
    rx += rDegHalf;
  if (!((rd + dInside2 < rDegMax) ?
    (rx >= rd - 0.9 && rx <= rd + dInside2 + 0.9) :
    (rx >= rd - 0.9 || rx <= rd + dInside2 + 0.9 - rDegMax)))
    return;

  rx -= rd;
  if (rx < 0.0)
    rx += rDegMax;
  rx = RTanD(rx - dr.dInside) / RTanD(dr.dInside) / 2.0 + 0.5;
  x = ws.xpClient-1 - (int)((real)ws.xpClient * rx);
  y = yAdd + (ws.ypClient >> 1) * dr.ySunMoon / 1000;
  if (dr.fStar) {

    // For the Moon, draw a bite out of the circle.
    d = NMultDiv(zSun*2, dr.nSunMoon, 100);
    DrawCrescent(c, x, y, zSun, x-d, y, dr.kvMoon,
      dr.kvInSky2 < 0 ? dr.kvInSky :
      KvBlendN(dr.kvInSky, dr.kvInSky2, dr.ySunMoon, 1000));
  } else {

    // For the Sun, draw rays from the circle.
    DiskC(c, x-zSun, y-zSun, x+zSun, y+zSun, dr.kvSun);
    zSun += NMultDiv(zSun*2, NAbs(dr.nSunMoon), 100);
    for (d = 0; d < 6; d++) {
      xi = (int)((real)(zSun)*RCosD((real)(d*30)));
      yi = (int)((real)(zSun)*RSinD((real)(d*30)));
      LineC(c, x-xi, y-yi, x+xi, y+yi, dr.kvSun);
    }
  }
}


// Draw a rainbow opposite the Sun for the inside display mode.

flag FDrawRainbow(colmap c, real rd, int yAdd)
{
  int x, y, y2, m, n, xT, yT, nSav1, nSav2, nR, nG, nB,
    nHi, nLo, nWid, nTrans;
  long l, q1, q2;
  flag fSav;
  real rx = dr.dInside, dInside1 = 89.0, dInside2 = dInside1*2.0;
  KV kv1;
  byte *pb1, *pb2, *pbT;

  rx = dInside1;
  if (dr.rLightFactor < 0.0)
    rx += rDegHalf;
  if (!((rd + dInside2 < rDegMax) ?
    (rx >= rd - 0.9 && rx <= rd + dInside2 + 0.9) :
    (rx >= rd - 0.9 || rx <= rd + dInside2 + 0.9 - rDegMax)))
    return fFalse;

  nHi = dr.nRainbow / 1000000;
  nLo = (dr.nRainbow / 10000) % 100;
  nWid = 99 - ((dr.nRainbow / 100) % 100);
  nTrans = 99 - (dr.nRainbow % 100);

  rx -= rd;
  if (rx < 0.0)
    rx += rDegMax;
  rx = RTanD(rx - dInside1) / RTanD(dr.dInside) / 2.0 + 0.5;
  x = ws.xpClient-1 - (int)((real)ws.xpClient * rx);
  y = (ws.ypClient >> 1) * nHi / 99;
  if (nTrans <= 0 || x+y < 0 || x-y >= ws.xpClient)
    return fTrue;

  // If the rainbow pixels haven't been generated yet, do so first.
  if (bm.cR == NULL || MY(bm.cR) != y+1) {
    if (!FColmapSizeSet(&bm.cR, y*2+1, y+1))
      return fFalse;
    ColmapSet(bm.cR, kvBlack);
    y2 = y * nWid / 99;
    q1 = Sq(y) + y - 1; q2 = Sq(y2) + y2 - 1;
    if (y2 == y)
      y2--;
    fSav = cs.fRainbowBlend; cs.fRainbowBlend = fTrue;
    nSav1 = cs.nRainbowStart; cs.nRainbowStart = 0;
    nSav2 = cs.nRainbowDistance; cs.nRainbowDistance = nDegMax;
    for (n = 0; n <= y; n++)
      for (m = -y; m <= y; m++) {
        l = Sq(n) + Sq(m);
        if (l <= q1 && l >= q2)
          CP(bm.cR, y + m, y - n,
            Hue((y*10 - NSqr(l*100)) * (nHueMax * 31 / 360) / (y-y2)));
      }
    cs.fRainbowBlend = fSav;
    cs.nRainbowStart = nSav1; cs.nRainbowDistance = nSav2;
  }

  // Blend the rainbow bitmap with what's already present.
  yT = yAdd + (ws.ypClient >> 1) + (y * nLo / 99);
  for (n = 0; n <= y; n++) {
    if (yT < 0 || yT >= CY(c))
      continue;
    pb1 = CJ(bm.cR, CI(bm.cR, 0, y - n));
    pb2 = CJ(c, CI(c, 0, yT));
    for (m = -y; m <= y; m++) {
      kv1 = CK(pb1);
      if (kv1 != kvBlack) {
        xT = x + m;
        if (xT >= 0 && xT < CX(c)) {
          pbT = pb2 + xT*cbPixelC;
          CM(pbT, nR, nG, nB);
          nR += (RgbR(kv1) - nR) * nTrans / 99;
          nG += (RgbG(kv1) - nG) * nTrans / 99;
          nB += (RgbB(kv1) - nB) * nTrans / 99;
          CL(pbT, nR, nG, nB);
        }
      }
      pb1 += cbPixelC;
    }
    yT--;
  }
  return fTrue;
}


// Create and draw background mountains for the inside display mode.

flag FCreateInsideMountains(colmap c, real rd, int yAdd)
{
  int i, i1, i2, x, y, y0, y1, y2, yb;
  real rx;

  // If the list of elevations hasn't been generated yet, do so first.
  if (dr.rgmtn == NULL) {
    dr.rgmtn = RgAllocate(imtnMax, short);
    if (dr.rgmtn == NULL)
      return fFalse;

    // Start with large blocky shapes to give general definition.
    y0 = y1 = Rnd(0, 10000);
    i1 = 0;
    while (i1 < imtnMax) {
      if (i1 + 90 < imtnMax) {
        i2 = i1 + Rnd(30, 60);
        y2 = Rnd(Max(y1 - 5000, 0), Min(y1 + 5000, 10000));
      } else {
        // Make sure the last elevation matches up with the first.
        i2 = imtnMax;
        y2 = y0;
      }
      for (i = i1; i < i2; i++)
        dr.rgmtn[i] = y1 + NMultDiv(i - i1, y2 - y1, i2 - i1);
      i1 = i2;
      y1 = y2;
    }

    // Then add in smaller adjustments to make the mountains seem more random
    // and natural. Two stages makes these simple fractal mountains! :)
    y0 = y1 = Rnd(-3000, -1000);
    i1 = 0;
    while (i1 < imtnMax) {
      if (i1 + 30 < imtnMax) {
        i2 = i1 + Rnd(10, 20);
        y2 = Rnd(-3000, -1000);
      } else {
        // Make sure the last elevation matches up with the first.
        i2 = imtnMax;
        y2 = y0;
      }
      for (i = i1; i < i2; i++)
        dr.rgmtn[i] += y1 + NMultDiv(i - i1, y2 - y1, i2 - i1);
      i1 = i2;
      y1 = y2;
    }
    for (i = 0; i < imtnMax; i++)
      if (dr.rgmtn[i] < 0)
        dr.rgmtn[i] = 0;
    if (c == NULL)
      return fTrue;
  }

  // Draw each part of the horizon at its appropriate elevation.
  yb = (ws.ypClient >> (int)!dr.fSkyAll) + yAdd;
  for (x = 0; x < ws.xpClient; x++) {
    rx = rd - dr.rgcalc[x].rAngleColumn;
    if (rx < 0)
      rx += rDegMax;
    else if (rx >= rDegMax)
      rx -= rDegMax;
    i = (int)(rx * (real)imtnMax / rDegMax);
    y = NMultDiv(ws.ypClient, (long)dr.rgmtn[i] * dr.nPeakHeight, 1600000);
    LineYGradient(c, x, yb-y, yb,
      dr.kvInMtn2 < 0 ? dr.kvInMtn : dr.kvInMtn2, dr.kvInMtn);
  }
  return fTrue;
}


#define dyCld 250
#define csCld (4 + (dyCld << 1))

// Create and draw background clouds for the inside display mode.

flag FCreateInsideClouds(colmap c, real rd, int yAdd)
{
  int i, is, j, k, x1, y1, x2, y2, i1, i2,
    ySky = ws.ypClient >> (int)!dr.fSkyAll;
  real rx1, rx2, rx3, rx4,
    dInside2 = dr.dInside*2.0, rT = RTanD(dr.dInside) * 2.0;
  KV kv = dr.kvInCld;
  flag f;

  // If the list of cloud positions hasn't been generated yet, do so first.
  if (dr.rgcld == NULL) {
    dr.rgcld = RgAllocate(icldMax*csCld, short);
    if (dr.rgcld == NULL)
      return fFalse;
    for (i = 0; i < icldMax; i++) {
      is = i * csCld;

      // Determine bounding box for this cloud. The first 16 clouds will never
      // be due south or north, i.e. they will never block the Sun/Moon.
      do {
        j = Rnd(0, 16383);
      } while (i < 16 && (j <= 3000 || FBetween(j, 8192, 8192+3000)));
      k = Rnd(3000, 6000);
      dr.rgcld[is] = j;
      dr.rgcld[is + 1] = Rnd(-(k >> 1), 16383 - (k >> 1));
      dr.rgcld[is + 2] = Rnd(1000, 3000);
      dr.rgcld[is + 3] = k;

      // Randomize the left and right sides of the cloud.
      for (f = 0; f <= 1; f++) {

        // Start with large blocky shapes to give general definition.
        y1 = Rnd(0, 4000);
        i1 = 0;
        while (i1 < dyCld) {
          if (i1 + 80 < dyCld)
            i2 = i1 + Rnd(40, 60);
          else
            i2 = dyCld;
          y2 = Rnd(0, 4000);
          for (j = i1; j < i2; j++) {
            k = y1 + NMultDiv(j - i1, y2 - y1, i2 - i1);
            dr.rgcld[is + 4 + f + (j << 1)] = 8192 - k;
          }
          i1 = i2;
          y1 = y2;
        }

        // Then add in smaller adjustments to make the cloud seem more random
        // and natural. Note how similar this process is to doing mountains.
        y1 = Rnd(0, 1000);
        i1 = 0;
        while (i1 < dyCld) {
          if (i1 + 20 < dyCld)
            i2 = i1 + Rnd(10, 15);
          else
            i2 = dyCld;
          y2 = Rnd(0, 1000);
          for (j = i1; j < i2; j++) {
            k = y1 + NMultDiv(j - i1, y2 - y1, i2 - i1);
            dr.rgcld[is + 4 + f + (j << 1)] -= k;
          }
          i1 = i2;
          y1 = y2;
        }
      }

      // Finally round the corners of the cloud so it has a bounding ellipse.
      for (j = 0; j < dyCld; j++) {
        k = NMultDiv(NAbs(j - 125), 8192, 125);
        k = NSqr(Sq(8192) - Sq(k));
        dr.rgcld[is + 4 + (j << 1)] = 8192 -
          NMultDiv(dr.rgcld[is + 4 + (j << 1)], k, 8192);
        dr.rgcld[is + 5 + (j << 1)] = 8192 +
          NMultDiv(dr.rgcld[is + 5 + (j << 1)], k, 8192);
      }
    }
    if (c == NULL)
      return fTrue;
  }

  // Draw each cloud at its appropriate position.
  f = (rd + dInside2 < rDegMax);
  for (i = 0; i < dr.cCloud; i++) {
    is = i * csCld;
    rx1 = (real)dr.rgcld[is] * rDegMax / 16384.0;
    rx2 = (real)(dr.rgcld[is] + dr.rgcld[is + 2]) * rDegMax / 16384.0;
    rx4 = rx2;
    if (rx4 >= rDegMax)
      rx4 -= rDegMax;
    if (!(f ? (rx1 > rd && rx1 < rd + dInside2) ||
      (rx4 > rd && rx4 < rd + dInside2) :
      (rx1 > rd || rx1 < rd + dInside2 - rDegMax) ||
      (rx4 > rd || rx4 < rd + dInside2 - rDegMax)))
      continue;
    y1 = yAdd + (int)((real)ySky * (real)dr.rgcld[is + 1] / 16384.0);
    y2 = y1   + (int)((real)ySky * (real)dr.rgcld[is + 3] / 16384.0);

    // For each pixel row, draw a horizontal line for that part of the cloud.
    for (j = Max(y1, 0); j < Min(y2, yAdd + ySky); j++) {
      if (dr.kvInCld2 >= 0)
        kv = KvBlendN(dr.kvInCld, dr.kvInCld2, j - y1, y2 - y1);
      k = is + 4 + (NMultDiv(j - y1, dyCld, y2 - y1) << 1);
      rx3 = rx1 + (real)dr.rgcld[k] * (rx2 - rx1) / 16384.0;
      rx4 = rx1 + (real)dr.rgcld[k + 1] * (rx2 - rx1) / 16384.0;
      if (rx3 >= rDegMax)
        rx3 -= rDegMax;
      if (rx4 >= rDegMax)
        rx4 -= rDegMax;
      if (f ? (rx3 > rd && rx3 < rd + dInside2) ||
        (rx4 > rd && rx4 < rd + dInside2) :
        (rx3 > rd || rx3 < rd + dInside2 - rDegMax) ||
        (rx4 > rd || rx4 < rd + dInside2 - rDegMax)) {
        if (f) {
          if (rx3 < rd || rx3 > rd + dInside2)
            rx3 = rd - 0.5;
          if (rx4 < rd || rx4 > rd + dInside2)
            rx4 = rd + dInside2 + 0.5;
        } else {
          if (rx3 < rd && rx3 > rd + dInside2 - rDegMax)
            rx3 = rd - 0.5;
          if (rx4 < rd && rx4 > rd + dInside2 - rDegMax)
            rx4 = rd + dInside2 - rDegMax + 0.5;
        }
        rx3 -= rd;
        if (rx3 < 0.0)
          rx3 += rDegMax;
        rx3 = RTanD(rx3 - dr.dInside) / rT + 0.5;
        rx4 -= rd;
        if (rx4 < 0.0)
          rx4 += rDegMax;
        rx4 = RTanD(rx4 - dr.dInside) / rT + 0.5;
        x1 = ws.xpClient-1 - (int)((real)ws.xpClient * rx3);
        x2 = ws.xpClient-1 - (int)((real)ws.xpClient * rx4);
        LineH(x1, x2, j, kv);
      }
    }
  }
  return fTrue;
}


// Adjust a cloud's horizontal coordinate by a specified offset.

flag FMoveCloud(int iCloud, int x)
{
  int icldLo, icldHi, icld, is, x0, dx;
  flag fVariable;

  // Return right away if there are no clouds, or not moving any distance.
  if (dr.rgcld == NULL)
    return fFalse;
  if (x == 0)
    return fTrue;

  // An invalid cloud index means move all clouds.
  if (FBetween(iCloud, 0, icldMax-1))
    icldLo = icldHi = iCloud;
  else {
    icldLo = 0; icldHi = icldMax-1;
  }

  // An invalid (too extreme) delta means move each cloud a different amount.
  fVariable = !FBetween(x, -16383, 16383);
  if (!fVariable)
    dx = x;
  else {
    x0 = NSgn(x);
    x -= x0 * 16384;
  }

  // Loop over the set of clouds to move.
  for (icld = icldLo; icld <= icldHi; icld++) {
    if (fVariable)
      dx = x0 + NMultDiv(icld - icldLo, x, 16384);
    is = icld * csCld;
    dr.rgcld[is] = dr.rgcld[is] - dx & 16383;
  }
  return fTrue;
}


// Regenerate the background components of the inside display mode.

void ResetInside()
{
  if (dr.rgstar != NULL) {
    DeallocateP(dr.rgstar);
    dr.rgstar = NULL;
  }
  if (dr.rgmtn != NULL) {
    DeallocateP(dr.rgmtn);
    dr.rgmtn = NULL;
  }
  if (dr.rgcld != NULL) {
    DeallocateP(dr.rgcld);
    dr.rgcld = NULL;
  }
  FCreateInsideStars(NULL, 0.0, 0, fFalse);
  FCreateInsideMountains(NULL, 0.0, 0);
  FCreateInsideClouds(NULL, 0.0, 0);
}


// Show a compass, i.e. print the direction the dot is facing.

void FormatCompass(char sz[cchSzDef], flag fOffset, int *pcch)
{
  CONST char *pch1, *pch2;
  int cch, a, d;

  if (fOffset || dr.fCompassAngle) {

    // Complex version: Figure out which of the eight directions to print.
    pch1 = pch2 = NULL;
    a = dr.dir * 90 + dr.nOffsetD;
    if (a < 0)
      a += nDegMax;
    d = a + 22;
    if (d > nDegMax)
      d -= nDegMax;
    d /= 45;
    if (d >= 7 || d <= 1)
      pch1 = rgszDir[0];
    else if (FBetween(d, 3, 5))
      pch1 = rgszDir[2];
    if (FBetween(d, 1, 3))
      pch2 = rgszDir[1];
    else if (FBetween(d, 5, 7))
      pch2 = rgszDir[3];
    if (dr.nOffsetD == 0) {
      if (pch1 != NULL)
        pch2 = pch1;
      pch1 = dr.fCompassAngle ? "Due " : NULL;
    }
    if (pch1 == NULL)
      pch1 = "";
    else if (pch2 == NULL)
      pch2 = "";
    if (dr.fCompassAngle)
      sprintf(SS(sz, cchSzDef), "%s%s a=%d", pch1, pch2, a);
    else
      sprintf(SS(sz, cchSzDef), "%s%s", pch1, pch2);
    cch = CchSz(sz);
  } else {

    // Simple version: Just print North, West, South, or East.
    sprintf(SS(sz, cchSzDef), rgszDir[dr.dir]);
    cch = 5 - (dr.dir & 1);
  }
  *pcch = cch;
}


// Show the dot's location, i.e. print its 2D or 3D coordinates.

void FormatLocation(char sz[cchSzDef], flag f3D)
{
  if (!f3D)
    sprintf(SS(sz, cchSzDef), "x=%d, y=%d",
      dr.fNarrow ? dr.x >> 1 : dr.x, dr.fNarrow ? dr.y >> 1 : dr.y);
  else
    sprintf(SS(sz, cchSzDef), "x=%d, y=%d, z=%d",
      dr.fNarrow ? (dr.x+1) >> 1 : dr.x, dr.fNarrow ? (dr.y+1) >> 1 : dr.y,
      (dr.fNarrow ? dr.z >> 1 : dr.z) + 1);
}


// Show the number of milliseconds it took to render the current scene.

void FormatTimer(char sz[cchSzDef], ulong lTimer, ulong lTimer2)
{
  dr.nDelay = lTimer2 - lTimer;
  sprintf(SS(sz, cchSzDef), "Frame Delay: %d", dr.nDelay);
}


// Do all background drawing, i.e. all things the actual walls are later drawn
// on top of, for the inside display mode.

void DrawBackground(colmap c, flag f3D, real rd, int yb)
{
  colmap cT;
  real rx, rT;
  int x, y, xSrc1, xSrc2, xDst, iTexture, xT, yT;
  KV kvSky = f3D && dr.z > 0 ? dr.kvInCeil : dr.kvInSky,
    kvSky2 = f3D && dr.z > 0 ? dr.kvInCeil2 : dr.kvInSky2;

  // Draw a rectangle for the sky or ceiling half of the screen.
  BlockC(c, 0, 0, ws.xpClient-1, yb-1, kvSky);

  // Draw a rectangle for the ground or floor half of the screen.
  if (dr.fSkyAll) {
    BlockC(c, 0, yb, ws.xpClient-1, ws.ypClient-1, dr.kvInDirt);
    if (dr.kvInDirt2 >= 0) {
      yT = ws.ypClient >> 1;
      for (y = Min(yT, ws.ypClient - yb)-1; y >= 0; y--)
        LineH(0, ws.xpClient, yb+y,
          KvBlendN(dr.kvInDirt2, dr.kvInDirt, y, yT));
    }
  }

  // Draw gradient between two colors over the sky.
  if (kvSky2 >= 0) {
    yT = ws.ypClient >> 1;
    for (y = Min(yT, yb)-1; y >= 0; y--)
      LineH(0, ws.xpClient, yb-1-y,
        KvBlendN(kvSky2, kvSky, y, yT));
  }

  // Do texture mapping over both the sky and ground part of the screen.
  if (dr.fTexture && (dr.nTexture > 0 || dr.nTexture2 > 0)) {
    iTexture = !f3D || dr.z <= 0 || dr.fSky3D ? dr.nTexture : dr.nTexture2;
    cT = *ColmapGetTexture(iTexture);
    if (iTexture > 0 && cT != NULL && !FNull(cT)) {
      yb -= ws.ypClient >> 1;
      rT = RTanD(dr.dInside);
      yT = ws.ypClient >> (int)!dr.fSkyAll;
      xDst = xSrc1 = 0;
      for (x = 0; x < ws.xpClient; x++) {
        rx = (real)(x << 1) / (real)ws.xpClient - 1.0;
        rx = rd - RAtnD(rx * rT);
        if (rx < 0)
          rx += rDegMax;
        else if (rx >= rDegMax)
          rx -= rDegMax;
        xSrc2 = CX(cT) - 1 - (int)(rx * (real)CX(cT) / rDegMax);
        if (xSrc2 != xSrc1 || x == ws.xpClient-1) {
          if (x > 0) {
            for (xT = xDst; xT <= x; xT++)
              LineYTexture(c, xT, yb, yb + yT, yb, yb + yT,
                cT, xSrc1, 0, CY(cT), 0, kvBlack, 0);
            xDst = x+1;
          }
          xSrc1 = xSrc2;
        }
      }
      yb += ws.ypClient >> 1;
    }
  }

  // Normally only do the outside things for 2D Mazes or the top level of 3D
  // Mazes. Draw things in order from farthest away to nearest.
  if (!f3D || dr.z <= 0 || dr.fSky3D) {
    if (dr.fStar)
      FCreateInsideStars(c, rd, yb-(ws.ypClient >> 1), fTrue);
    if (dr.fSunMoon)
      DrawSunMoon(c, rd, yb-(ws.ypClient >> 1));

    // Draw meteors if appropriate.
    if (dr.fStar && dr.nMeteor > 0) {
      yb -= ws.ypClient >> 1;
      y = (ws.ypClient >> (int)!dr.fSkyAll) - 1;
      while (Rnd(0, 9999) < dr.nMeteor)
        LineC(c, Rnd(0, ws.xpClient-1), yb + Rnd(0, y),
          Rnd(0, ws.xpClient-1), yb + Rnd(0, y), KvStarRandom());
      yb += ws.ypClient >> 1;
    }

    if (dr.cCloud > 0)
      FCreateInsideClouds(c, rd, yb-(ws.ypClient >> 1));
    if (dr.fRainbow)
      FDrawRainbow(c, rd, yb-(ws.ypClient >> 1));
    if (dr.fMountain)
      FCreateInsideMountains(c, rd, yb-(ws.ypClient >> 1));
  }

  // Draw a rectangle for the ground or floor half of the screen.
  if (!dr.fSkyAll) {
    BlockC(c, 0, yb, ws.xpClient-1, ws.ypClient-1, dr.kvInDirt);
    if (dr.kvInDirt2 >= 0) {
      yT = ws.ypClient >> 1;
      for (y = Min(yT, ws.ypClient - yb)-1; y >= 0; y--)
        LineH(0, ws.xpClient, yb+y,
          KvBlendN(dr.kvInDirt2, dr.kvInDirt, y, yT));
    }
  }
}


// A bunch of macros used by RedrawInsidePerspective().

#define ypWallMax (1 << 28)
#define LengthI(z, t) (Sq(z) + (Sq((z)*(t))))
#define DistanceM() RAbs((real)(me - ml) * md0)
#define DistanceN() RAbs((real)(ne - nl) * nd0)

#define ComputeDir(k) \
  k = (yo - yg + 1) + ((xo > xg) << 1)

#define ShadeKv(kv) \
  ComputeDir(k); \
  if (k & 1) \
    ; \
  else if (k == 0) \
    kv = KvShade(kv, -dr.rLightFactor); \
  else \
    kv = KvShade(kv, dr.rLightFactor); \
  if (fFog) \
    kv = KvApplyFog(kv);

#define GetKvCore2(kv, c, x, y) \
  if (FLegalC(c, x, y)) \
    kv = CG(c, x, y);

#define GetKvCore(kv, c, x, y, z) \
  if (f3D) { \
    GetKvCore2(kv, c, Tx(bm.a, x, z), Ty(bm.a, y, z)); \
  } else { \
    GetKvCore2(kv, c, x, y); \
  }

#define GetKv(kv, c, x, y, z, kvElse) \
  kv = kvElse; \
  GetKvCore(kv, c, x, y, z)

#define GetKvIf(kv, x, y, z, kvElse) \
  kv = kvElse; \
  if (fColor) { \
    GetKvCore(kv, bm.c, x, y, z) \
  }

#define GetKvMark(kv, c, x, y) \
  GetKv(kv, c, x, y, dr.z, kvMarkDef)

#define Ensure3DColor(x, y, dz) \
  GetKvIf(kv3D, x, y, dr.z + dz, dr.kvIn3D) \
  ShadeKv(kv3D);

#define KvApplyFog(kv) \
  KvBlendR(kv, dr.kvInFog, Min(zd / rFogDist, 1.0))

#define DrawTexture3D(dz) \
  if (fTextureWall) { \
    ComputeTexture(&iTexture, &iMask, &rTexture, dr.nTextureWall, \
      xg, yg, dz, k, xq, yq, me, ml, ne, nl, md, nd, mt, nt); \
    if (iTexture > 0 || iMask > 0) \
      DrawTexture(c, x, y, dz > 0 ? yExtra : -yExtra2, IP(yb), iTexture, \
        iMask, rTexture, 0, fFog ? (int)(Min(zd / rFogDist, 1.0) * 100.0) : \
        0, ws.rgbMask[0] != NULL && Mg(ws.rgbMask[0], xg, yg)); \
  }

#define FWallVar(x, y) \
  ((dr.zElev != 0) || (fWallVar2 && Mg3(bWallVar, x, y, dr.z, f3D)))

#define ComputeTextureOffset(x, y) \
  if (md < nd) \
    rTexture = RAbs((real)(me - ml)*mt) - RAbs((real)(ne - nl)); \
  else \
    rTexture = RAbs((real)(ne - nl)*nt) - RAbs((real)(me - ml)); \
  rTexture = RAbs(rTexture); \
  if (!dr.fNarrow || FOdd(x & y) || FOdd(x & (k ^ 1)) || FOdd(y & k)) \
    rTexture /= (real)dr.zCell; \
  else \
    rTexture /= (real)dr.zCellNarrow; \
  if ((xq < 1) ^ (yq < 1) ^ FOdd(k)) \
    rTexture = 1.0 - rTexture;

#define ComputeTextureOffsetFloor1(rx1, ry1) \
  switch (k) { \
  case 0 : rx1 = 1.0 - rTexture; ry1 = 0.0; break; \
  case 1 : rx1 = 0.0; ry1 = rTexture; break; \
  case 2 : rx1 = rTexture; ry1 = 1.0; break; \
  default: rx1 = 1.0; ry1 = 1.0 - rTexture; break; \
  }

#define ComputeTextureOffsetFloor2(rx2, ry2) \
  switch (k) { \
  case 0 : rx2 = 1.0 - rTexture; ry2 = 1.0; break; \
  case 1 : rx2 = 1.0; ry2 = rTexture; break; \
  case 2 : rx2 = rTexture; ry2 = 0.0; break; \
  default: rx2 = 0.0; ry2 = 1.0 - rTexture; break; \
  }

#define ComputeTextureOffsetFloor3(rx3, ry3) \
  switch (k) { \
  case 0 : rx3 = rTexture; ry3 = 0.0; break; \
  case 1 : rx3 = 1.0; ry3 = rTexture; break; \
  case 2 : rx3 = 1.0 - rTexture; ry3 = 1.0; break; \
  default: rx3 = 0.0; ry3 = 1.0 - rTexture; break; \
  }

#define ComputeTextureOffsetFloor4(rx4, ry4) \
  switch (k) { \
  case 0 : rx4 = rTexture; ry4 = 1.0; break; \
  case 1 : rx4 = 0.0; ry4 = rTexture; break; \
  case 2 : rx4 = 1.0 - rTexture; ry4 = 0.0; break; \
  default: rx4 = 1.0; ry4 = 1.0 - rTexture; break; \
  }


// Determine what texture mapping to apply to the current wall at the current
// column, if any.

void ComputeTexture(int *piTexture, int *piMask, real *prTexture, int ic,
  int xg, int yg, int dz, int k, int xq, int yq,
  long me, long ml, long ne, long nl, real md, real nd, real mt, real nt)
{
  int iTexture = 0, iMask = 0;
  real rTexture = 0.0;
  colmap c = ws.rgcTexture[ic];
  byte *pb;
  KV kv;

  if (dr.f3D) {
    xg = Tx(bm.a, xg, dr.z + dz); yg = Ty(bm.a, yg, dr.z + dz);
  }
  if (!FLegalC(c, xg, yg))
    goto LDone;
  pb = CJ(c, CI(c, xg, yg));
  kv = CK(pb);
  if (kv == kvBlack)
    goto LDone;

  // Determine the color and monochrome texture to apply to this wall.
  iTexture = ITextureWall(kv, xg, yg, k);
  if (iTexture > 0) {
    if (iTexture < ws.cbMask && ws.rgbMask[iTexture] != NULL)
      iMask = iTexture;
    if (iTexture >= ws.ccTexture || ws.rgcTexture[iTexture] == NULL)
      iTexture = 0;

    // If any texture, determine horizontal fraction within texture to draw.
    if (iTexture > 0 || iMask > 0) {
      ComputeTextureOffset(xg, yg);
    }
  }
LDone:
  *piTexture = iTexture; *piMask = iMask; *prTexture = rTexture;
}


// Determine what texture mapping to apply to the current floor cell, if any.

void ComputeTextureFloor(int *piTexture, int *piMask, int xg, int yg,
  flag fBlock, flag fCeiling)
{
  int nTexture = fBlock ? dr.nTextureVar : dr.nTextureDirt,
    iTexture = 0, iMask = 0;
  colmap c = ws.rgcTexture[nTexture];
  byte *pb;
  KV kv;

  if (dr.f3D) {
    xg = Tx(bm.a, xg, dr.z); yg = Ty(bm.a, yg, dr.z);
  }
  if (!FLegalC(c, xg, yg))
    goto LDone;
  pb = CJ(c, CI(c, xg, yg));
  kv = CK(pb);
  if (kv == kvBlack)
    goto LDone;

  // Determine the color and monochrome texture to apply to the floor here.
  iTexture = fCeiling ? UdD(kv) : UdU(kv);
  if (iTexture > 0) {
    if (iTexture < ws.cbMask && ws.rgbMask[iTexture] != NULL)
      iMask = iTexture;
    if (iTexture >= ws.ccTexture || ws.rgcTexture[iTexture] == NULL)
      iTexture = 0;
  }
LDone:
  *piTexture = iTexture; *piMask = iMask;
}


// Draw a texture at the given column. For solid color textures, this
// completely covers whatever solid wall color would otherwise be present. For
// textures overlays, this is done on top of whatever solid wall color is
// already present.

void DrawTexture(colmap c, int x, int y, int y2, int yb,
  int iTexture, int iMask, real rTexture, int nTrans, int nFog, flag fMaskSet)
{
  bitmap bT;
  colmap cT;
  int xTexture, yDst1, yDst2, yClip1, yClip2;

  // Figure out the pixel offset within the texture bitmaps. Color bitmap
  // only: Solid texture. Monochrome bitmap only: Black or white overlay on
  // top of the wall. Both color and monochrome: Merge color texture with the
  // black or white overlay, resulting in a color overlay on top of the wall.

  if (iMask > 0)
    bT = ws.rgbMask[iMask];
  if (iTexture > 0) {
    cT = ws.rgcTexture[iTexture];
    xTexture = (int)((real)CX(cT) * rTexture);
    if (xTexture >= CX(cT))
      xTexture = CX(cT)-1;
  } else {
    cT = NULL;
    xTexture = (int)((real)MX(bT) * rTexture);
    if (xTexture >= MX(bT))
      xTexture = MX(bT)-1;
  }

  // Figure out the actual pixel coordinates to draw between.
  if (yb == ~0) {
    yDst1 = IO(y); yDst2 = IO(y2);
    yClip1 = yDst1; yClip2 = yDst2;
  } else if (y2 == 0) {
    yDst1 = IO(yb)-IO(y); yDst2 = IO(yb)+IO(y);
    yClip1 = yDst1; yClip2 = yDst2;
  } else if (y2 > 0) {
    yDst1 = IO(yb+y); yDst2 = IO(yb+y*3);
    yClip1 = yDst1; yClip2 = IO(yb+Min(y2, y*3));
  } else { // y2 < 0
    yDst1 = IO(yb-y*3+iRound); yDst2 = IO(yb-y+iRound);
    yClip1 = IO(yb-Min(-y2, y*3)+iRound); yClip2 = yDst2;
  }

  // Go draw the solid texture map or texture overlay.
  if (iTexture > 0 && iMask == 0)
    LineYTexture(c, x, yDst1, yDst2, yClip1, yClip2,
      cT, xTexture, 0, CY(cT), nTrans, dr.kvInFog, nFog);
  else
    LineYMask(c, x, yDst1, yDst2, yClip1, yClip2,
      bT, cT, xTexture, 0, cT != NULL ? CY(cT) : MY(bT), dr.kvInFog, nFog,
      fMaskSet);
}


// Compute and return the height in inside units of the bottom and top of a
// variable height wall.

void ComputeWallVar(flag fWallVar2, bitmap bWallVar, colmap cWallVar,
  int zWall, int x, int y, int *pyLo, int *pyHi)
{
  int yLo, yHi, yT;

  if (fWallVar2 && Mg3(bWallVar, x, y, dr.z, dr.f3D)) {

    // For variable height walls, translate from percentage to inside units.
    yT = Cg3F(cWallVar, x, y, dr.z, dr.f3D);
    yLo = (UdD(yT) << iVarScale2) * dr.zWall / zWall;
    yHi = (UdU(yT) << iVarScale2) * dr.zWall / zWall;
    if (yLo > yHi)
      SwapN(yLo, yHi);
  } else {

    // Non-variable height walls are always the default wall height.
    yLo = 0; yHi = iVar2 * dr.zWall / zWall;
  }

  // Adjust height by the viewing height. Levitating over the tops of walls is
  // implemented by making all walls shorter by the right proportion.

  if (dr.zElev < 0) {
    yT = iVar2 - (iVar2 * dr.zWall / zWall);
    yLo += yT; yHi += yT;
  }
  *pyLo = yLo; *pyHi = yHi;
}


// Draw the appropriate set of precomputed semitransparent and variable height
// walls at a column. These are drawn on top of whatever solid walls or
// background are already in place.

void DrawTransparent(colmap c, int x, int yb, int cTrans, TRANS *trans,
  real rScale)
{
  int i, yLo, yHi, yLo1, yHi1, yLo2, yHi2,
    ypLo, ypHi, ypHiU = -1, ypLoU = -1, ypMax = CY(c);
  KV kv, kvT;
  flag fWall;

  for (i = 0; i < cTrans; i++) {
    yLo = trans[i].yLo; yHi = trans[i].yHi;

    // Compute pixel coordinates.
    yLo1 = IP(yb) + trans[i].y;
    yLo2 = IP(yb) + trans[i].y2;
    if (yHi == iVar2) {
      yHi1 = IP(yb) - trans[i].y;
      yHi2 = IP(yb) - trans[i].y2;
    } else {
      yHi1 = yLo1 - LMultShift(trans[i].y,  yHi, iScale2) + iRound;
      yHi2 = yLo2 - LMultShift(trans[i].y2, yHi, iScale2) + iRound;
    }
    if (yLo > 0) {
      yLo1 -= LMultShift(trans[i].y,  yLo, iScale2) - iRound;
      yLo2 -= LMultShift(trans[i].y2, yLo, iScale2) - iRound;
    }

    // Don't bother with this wall if it's entirely behind other solid walls.
    trans[i].fSkip = fFalse;
    ypLo = IO(Max(yLo1, yLo2)); ypLo = Min(ypLo, ypMax);
    ypHi = IO(Min(yHi1, yHi2)); ypHi = Max(ypHi, 0);
    if (ypLo <= ypLoU && ypHi >= ypHiU) {
      trans[i].fSkip = fTrue;
      continue;
    }

    // If wall is solid and overlaps current solid range, expand the range.
    if (!trans[i].fTrans) {
      if ((ypHi < ypHiU && ypLo >= ypHiU) ||
        (ypLo > ypLoU && ypHi <= ypLoU)) {
        if (ypLo > ypLoU)
          ypLoU = ypLo;
        if (ypHi < ypHiU)
          ypHiU = ypHi;
      } else if (ypLo - ypHi > ypLoU - ypHiU) {
        ypLoU = ypLo; ypHiU = ypHi;
      }
    }
    trans[i].yLo1 = yLo1; trans[i].yHi1 = yHi1;
    trans[i].yLo2 = yLo2; trans[i].yHi2 = yHi2;

    // If solid wall range covers entire screen, skip anything else behind.
    if (ypHiU <= 0 && ypLoU >= ypMax) {
      cTrans = i+1;
      break;
    }
  }

  // Draw semitransparent and variable height walls from back to front,
  // blending each semitransparent pixel with whatever is already in place.

  rScale /= (real)iVar2;
  for (i = cTrans-1; i >= 0; i--) {
    if (trans[i].fSkip)
      continue;
    kv = trans[i].kv;
    yLo  = trans[i].yLo;  yHi  = trans[i].yHi;
    yLo1 = trans[i].yLo1; yHi1 = trans[i].yHi1;
    yLo2 = trans[i].yLo2; yHi2 = trans[i].yHi2;
    fWall = !(dr.nTrans == nTransFast && trans[i].fTrans && i > 0 &&
      trans[i-1].fTrans && trans[i].y == trans[i-1].y2 &&
      yHi == trans[i-1].yHi && yLo == trans[i-1].yLo);

    if (yHi >= 0) {

      // Draw a variable height wall.
      if (!trans[i].fTrans || dr.nTrans != nTransVery) {

        // Draw a solid or simple semitransparent variable height wall.
        fWall &= (trans[i].fTrans || i == 0 || yHi > trans[i-1].yHi ||
          yLo < trans[i-1].yLo || ((yHi < iVarHalf || yLo > iVarHalf) &&
          trans[i].y != trans[i-1].y2) || trans[i-1].fTrans);
        if (fWall)
          LineYTrans(c, x, IO(yHi1), IO(yLo1), kv,
            trans[i].fTrans ? dr.nTransPct : 0);
        if (yHi < iVarHalf) {

          // Looking down on top of wall.
          kvT = KvShade(trans[i].kvOrig, RAbs(dr.rLightFactor) / 2.0);
          if (!dr.fFogFloor)
            LineYTrans(c, x, IO(yHi2), IO(yHi1), kvT,
              trans[i].fTrans ? dr.nTransPct2 : 0);
          else
            LineYFloor(c, x, yb, IO(yHi2)-yb, IO(yHi1)-yb, kvT,
              trans[i].fTrans ? dr.nTransPct2 : 0,
              rScale * (real)((iVarHalf - yHi) << 1));
          if (trans[i].iMaskVar > 0 || trans[i].iTextureVar > 0)
            LineYFloorTexture(c, x, yb, yHi2-IP(yb), yHi1-IP(yb),
              trans[i].iMaskVar > 0 ? ws.rgbMask[trans[i].iMaskVar] : NULL,
              trans[i].iTextureVar > 0 ? ws.rgcTexture[trans[i].iTextureVar] :
              NULL, trans[i].rx2, trans[i].ry2, trans[i].rx1, trans[i].ry1,
              trans[i].fMaskSetVar, rScale * (real)((iVarHalf - yHi) << 1));
        } else if (yLo > iVarHalf) {

          // Looking up at bottom of wall.
          kvT = KvShade(trans[i].kvOrig, RAbs(dr.rLightFactor) / -2.0);
          if (!dr.fFogFloor)
            LineYTrans(c, x, IO(yLo1), IO(yLo2), kvT,
              trans[i].fTrans ? dr.nTransPct2 : 0);
          else
            LineYFloor(c, x, yb, IO(yLo1)-yb, IO(yLo2)-yb, kvT,
              trans[i].fTrans ? dr.nTransPct2 : 0,
              rScale * (real)((yLo - iVarHalf) << 1));
          if (trans[i].iMaskVar2 > 0 || trans[i].iTextureVar2 > 0)
            LineYFloorTexture(c, x, yb, yLo1-IP(yb), yLo2-IP(yb),
              trans[i].iMaskVar2 > 0 ? ws.rgbMask[trans[i].iMaskVar2] : NULL,
              trans[i].iTextureVar2 > 0 ? ws.rgcTexture[trans[i].iTextureVar2]
              : NULL, trans[i].rx3, trans[i].ry3, trans[i].rx4, trans[i].ry4,
              trans[i].fMaskSetVar, rScale * (real)((iVarHalf - yHi) << 1));
        }
      } else {

        // Draw a very thorough semitransparent variable height wall.
        if (yHi < iVarHalf) {

          // Looking down on top of wall.
          kvT = KvShade(trans[i].kvOrig, RAbs(dr.rLightFactor) / 2.0);
          if (yHi1 <= yLo2) {
            if (fWall) {
              LineYTrans(c, x, IO(yLo2), IO(yLo1), kv, dr.nTransPct2);
              LineYTrans(c, x, IO(yHi1), IO(yLo2), kv, dr.nTransPct);
            }
            if (!dr.fFogFloor)
              LineYTrans(c, x, IO(yHi2), IO(yHi1), kvT, dr.nTransPct2);
            else
              LineYFloor(c, x, yb, IO(yHi2)-yb, IO(yHi1)-yb, kvT,
                dr.nTransPct2, rScale * (real)((iVarHalf - yHi) << 1));
          } else {
            if (fWall)
              LineYTrans(c, x, IO(yHi1), IO(yLo1), kv, dr.nTransPct2);
            if (!dr.fFogFloor) {
              LineYTrans(c, x, IO(yLo2), IO(yHi1), trans[i].kvOrig,
                dr.nTransPct);
              LineYTrans(c, x, IO(yHi2), IO(yLo2), kvT, dr.nTransPct2);
            } else {
              LineYFloor(c, x, yb, IO(yLo2)-yb, IO(yHi1)-yb,
                trans[i].kvOrig, dr.nTransPct,
                rScale * (real)((iVarHalf - yHi) << 1));
              LineYFloor(c, x, yb, IO(yHi2)-yb, IO(yLo2)-yb,
                kvT, dr.nTransPct2, rScale * (real)((iVarHalf - yHi) << 1));
            }
          }
          if (trans[i].iMaskVar > 0 || trans[i].iTextureVar > 0)
            LineYFloorTexture(c, x, yb, yHi2-IP(yb), yHi1-IP(yb),
              trans[i].iMaskVar > 0 ? ws.rgbMask[trans[i].iMaskVar] : NULL,
              trans[i].iTextureVar > 0 ? ws.rgcTexture[trans[i].iTextureVar] :
              NULL, trans[i].rx2, trans[i].ry2, trans[i].rx1, trans[i].ry1,
              trans[i].fMaskSetVar, rScale * (real)((iVarHalf - yHi) << 1));
        } else if (yLo > iVarHalf) {

          // Looking up at bottom of wall.
          kvT = KvShade(trans[i].kvOrig, RAbs(dr.rLightFactor) / -2.0);
          if (yLo1 >= yHi2) {
            if (fWall) {
              LineYTrans(c, x, IO(yHi1), IO(yHi2), kv, dr.nTransPct2);
              LineYTrans(c, x, IO(yHi2), IO(yLo1), kv, dr.nTransPct);
            }
            if (!dr.fFogFloor)
              LineYTrans(c, x, IO(yLo1), IO(yLo2), kvT, dr.nTransPct2);
            else
              LineYFloor(c, x, yb, IO(yLo1)-yb, IO(yLo2)-yb, kvT,
                dr.nTransPct2, rScale * (real)((yLo - iVarHalf) << 1));
          } else {
            if (fWall)
              LineYTrans(c, x, IO(yHi1), IO(yLo1), kv, dr.nTransPct2);
            if (!dr.fFogFloor) {
              LineYTrans(c, x, IO(yLo1), IO(yHi2), trans[i].kvOrig,
                dr.nTransPct);
              LineYTrans(c, x, IO(yHi2), IO(yLo2), kvT, dr.nTransPct2);
            } else {
              LineYFloor(c, x, yb, IO(yLo1)-yb, IO(yHi2)-yb,
                trans[i].kvOrig, dr.nTransPct,
                rScale * (real)((yLo - iVarHalf) << 1));
              LineYFloor(c, x, yb, IO(yHi2)-yb, IO(yLo2)-yb,
                kvT, dr.nTransPct2, rScale * (real)((yLo - iVarHalf) << 1));
            }
          }
          if (trans[i].iMaskVar2 > 0 || trans[i].iTextureVar2 > 0)
            LineYFloorTexture(c, x, yb, yLo1-IP(yb), yLo2-IP(yb),
              trans[i].iMaskVar2 > 0 ? ws.rgbMask[trans[i].iMaskVar2] : NULL,
              trans[i].iTextureVar2 > 0 ? ws.rgcTexture[trans[i].iTextureVar2]
              : NULL, trans[i].rx3, trans[i].ry3, trans[i].rx4, trans[i].ry4,
              trans[i].fMaskSetVar, rScale * (real)((iVarHalf - yHi) << 1));
        } else {

          // Looking straight through wall.
          if (fWall) {
            LineYTrans(c, x, IO(yHi1), IO(yHi2), kv, dr.nTransPct2);
            LineYTrans(c, x, IO(yHi2), IO(yLo2), kv, dr.nTransPct);
            LineYTrans(c, x, IO(yLo2), IO(yLo1), kv, dr.nTransPct2);
          }
        }
      }
    } else if (fWall) {

      // Draw a default height wall.
      if (dr.nTrans != nTransVery)
        LineYTrans(c, x, IO(yHi1), IO(yLo1), kv, dr.nTransPct);
      else {
        LineYTrans(c, x, IO(yHi1), IO(yHi2), kv, dr.nTransPct2);
        LineYTrans(c, x, IO(yHi2), IO(yLo2), kv, dr.nTransPct);
        LineYTrans(c, x, IO(yLo2), IO(yLo1), kv, dr.nTransPct2);
      }
    }

    // Texture map the wall.
    if (fWall && (trans[i].iTexture > 0 || trans[i].iMask > 0))
      DrawTexture(c, x, yHi1, yLo1, ~0,
        trans[i].iTexture, trans[i].iMask, trans[i].rTexture,
        trans[i].fTrans ? dr.nTransPct : 0, trans[i].nFog, trans[i].fMaskSet);
  }
}


// Draw the 3D first person perspective inside view in the window. This is one
// of the largest and most interesting functions in Daedalus. It has many
// cases, but its core is a pretty simple ray casting algorithm. It's very
// fast because it only shoots one ray per pixel column. Each ray is shot from
// the viewing position, looking at each bitmap array cell it passes over
// until it hits a wall i.e. set cell. Take the distance to the hit, and that
// determines the height of the wall at that column. This method has the same
// frame rate for a 10x10 bitmap array as it does in a 10000x10000 array,
// where it doesn't matter how many polygons the array contains. The classic
// PC games Wolfenstein 3D and DOOM do similar "one dimensional ray tracing".

flag RedrawInsidePerspective(colmap c)
{
  // Main variables for standard walls.
  KV kvWall, kv1, kv1Old, kv2, kv2Old;
  long xl, yl, ml, nl, me, ne;
  int xc = ws.xpClient >> 1, yc = ws.ypClient >> 1, yb, x, y, xq, yq, xo, yo,
    xg, yg, xmax, ymax, mi, ni, ms, ns, mg, ng, i, k = 1, kOld = 1, zWall;
  real zl, z, zi, zj, mt, nt, md, nd, md0, nd0, zd, rScale, rd, rFogDist, rT;
  flag fOffset = FBetween(dr.nInside, nInsideSmooth, nInsideVeryFree),
    fBeenIn, fColor = (bm.c != NULL), fAdd;

  // Variables for special features, such as colored floor and ceiling
  // squares, 3D passages in the floor and ceiling, semitransparent walls, and
  // variable height walls.
  bitmap bMark2, bWallVar = NULL;
  colmap cMark = dr.fMarkColor ? bm.c2 : bm.c, cMark2, cWallVar = NULL;
  KV kv3D, kvMark, kvMarkOld, kvMark2, kvMarkOld2, kvMarkDef, kvT;
  flag fExtra, fFog = (dr.nFog > 0), fWallVar, fWallVar2, fInWallVar = fFalse,
    f3D = dr.f3D, fIn3D = fFalse, fIn3D2 = fFalse, fMark = (bm.a2 != NULL &&
    MX(bm.a2) <= MX(bm.a) && MY(bm.a2) <= MY(bm.a) &&
    dr.zElev >= -dr.zWall / 2), fInMark = fFalse, fMark2, fInMark2 = fFalse,
    fTrans = (dr.nTrans != nTransNone && bm.a3 != NULL), fInTrans = fFalse;
  int yExt, yExtra, yExtra2, itransMac = 10, cTrans;
  flag fColorMark = (cMark != NULL), fColorMark2,
    fDoCompute = fFalse, fDoEnd = fFalse, fDoWall = fFalse,
    fDoSeal = fFalse, fDoTrans = fFalse, fDoWallVar = fFalse,
    fDoEnterMark = fFalse, fDoEnterMark2 = fFalse, fDoEnter3D = fFalse,
    fDoEnter3D2 = fFalse, fDoExitMark = fFalse, fDoExitMark2 = fFalse,
    fDoExit3D = fFalse, fDoExit3D2 = fFalse, fDoExitTrans = fFalse,
    fDoExitWallVar = fFalse, fTransStart = fFalse, fWallVarStart = fFalse,
    fDoTextureVar = fFalse;
  TRANS *trans, *ptransCur, *ptransOld;

  // Variables for texture mapping.
  int iTexture, iMask = 0,
    iTextureMark = 0, iMaskMark = 0, iTextureMark2 = 0, iMaskMark2 = 0;
  real rTexture, rx1, ry1, rx2 = 0.0, ry2 = 0.0, rx3, ry3, rx4, ry4;
  flag fTextureWall = dr.fTexture && dr.nTextureWall >= 0 &&
    ws.ccTexture > dr.nTextureWall && ws.rgcTexture[dr.nTextureWall] != NULL;
  flag fTextureDirt = dr.fTexture && dr.nTextureDirt >= 0 &&
    ws.ccTexture > dr.nTextureDirt && ws.rgcTexture[dr.nTextureDirt] != NULL;
  bitmap bTextureDirt = fTextureDirt && ws.cbMask > dr.nTextureDirt ?
    ws.rgbMask[dr.nTextureDirt] : NULL;
  flag fTextureVar = dr.fTexture && dr.nTextureVar >= 0 &&
    ws.ccTexture > dr.nTextureVar && ws.rgcTexture[dr.nTextureVar] != NULL;
  bitmap bTextureVar = fTextureVar && ws.cbMask > dr.nTextureVar ?
    ws.rgbMask[dr.nTextureVar] : NULL;

  // Prepare the window for drawing.
  rT = RTanD(dr.dInside); rd = (real)xc / rT;
  if (ws.xpClient != bm.xaI || rd != dr.rdCalc) {
    if (ws.xpClient != bm.xaI) {
      if (dr.rgcalc != NULL)
        DeallocateP(dr.rgcalc);
      dr.rgcalc = RgAllocate(ws.xpClient, CALC);
      if (dr.rgcalc == NULL)
        return fFalse;
    }
    for (x = 0; x < ws.xpClient; x++) {
      dr.rgcalc[x].rAngleColumn = RAtnD((real)(x - xc) / rd);
      dr.rgcalc[x].rDistCorrect = RCosD(dr.rgcalc[x].rAngleColumn);
    }
    dr.rdCalc = rd;
  }

  // Initialize a bunch of variables.
  if (f3D) {
    xmax = TX(bm.a); ymax = TY(bm.a);
  } else {
    xmax = MX(bm.a); ymax = MY(bm.a);
  }
  xl = InsideFromA(dr.x) + dr.nOffsetX; yl = InsideFromA(dr.y) + dr.nOffsetY;
  zl = (real)(dr.dir * 90 + dr.nOffsetD);
  if (zl < 0.0)
    zl += rDegMax;
  yb = yc * (dr.yHorizon + 1000) / 1000;
  if (fOffset && f3D)
    yb -= NMultDiv(dr.nOffsetZ, yc, dr.zCell >> 1);
  xo = dr.x; yo = dr.y;
  zWall = dr.zWall + (NAbs(dr.zElev) << 1);
  rScale = (real)(zWall * yc << iScale) * 0.801 / rT;
  rFogDist = (real)(dr.nFog * 10);
  kvWall = (f3D ? dr.kvIn3D : dr.kvInWall);
  GetKvIf(kv1, xo, yo, dr.z, kvWall);
  kv2 = kv1; kv1Old = kv1; kv2Old = kv2;
  kvMark = kvMarkOld = kvMark2 = kvMarkOld2 = kvMarkDef =
    (f3D ? dr.kvInWall : dr.kvIn3D);
  fMark2 = (dr.nMarkSky >= 0 && ws.cbMask > dr.nMarkSky &&
    (bMark2 = ws.rgbMask[dr.nMarkSky]) != NULL) && dr.zElev <= 0;
  cMark2 = fMark2 && ws.ccTexture > dr.nMarkSky &&
    ws.rgcTexture[dr.nMarkSky] != NULL ? ws.rgcTexture[dr.nMarkSky] : cMark;
  fColorMark2 = cMark2 != NULL;
  fWallVar2 = (dr.nWallVar >= 0 && ws.cbMask > dr.nWallVar && ws.ccTexture >
    dr.nWallVar && (bWallVar = ws.rgbMask[dr.nWallVar]) != NULL &&
    (cWallVar = ws.rgcTexture[dr.nWallVar]) != NULL);
  fWallVar = fWallVar2 || dr.zElev != 0;
  fExtra = fMark || fMark2 || f3D;
  if (fTrans || fWallVar) {
    if (fWallVar)
      itransMac = Min(dr.nClip + 1, itransMax);
    if (dr.rgtrans == NULL)
      dr.rgtrans = RgAllocate(itransMax, TRANS);
    trans = dr.rgtrans;
    if (trans == NULL)
      return fFalse;
    dr.nTransPct = ds.nTrans > 0 ? ds.nTrans / 100 : 50;
    dr.nTransPct2 = dr.nTransPct * 4 / 5;
  }

  // Clear the window and draw all background stuff behind the walls.
  DrawBackground(c, f3D, zl, yb);

  // If the dot is located within a solid or semitransparent wall, draw or
  // prepare to draw later over the whole window.
  if (fTrans && Mg3(bm.a3, xo, yo, dr.z, f3D))
    fTransStart = fTrue;
  if (fTransStart || Mg3(bm.a, xo, yo, dr.z, f3D)) {
    if (fWallVar && FWallVar(xo, yo)) {
      fWallVarStart = fTrue;
      ComputeWallVar(fWallVar2, bWallVar, cWallVar, zWall, xo, yo,
        &trans[0].yLo, &trans[0].yHi);
    }
    if (!fTransStart && (!fWallVarStart ||
      (trans[0].yHi > iVarHalf && trans[0].yLo < iVarHalf))) {
      ColmapSet(c, kv2);
      return fTrue;
    }
  }
  if (fTransStart || fWallVarStart) {
    GetKvIf(kvT, xo, yo, dr.z, kvWall)
    trans[0].kv = trans[0].kvOrig = kvT;
    trans[0].y = trans[0].y2 = ypWallMax >> iVarScale2;
    trans[0].fTrans = fTransStart;
    if (!fWallVarStart) {
      trans[0].yLo = 0; trans[0].yHi = iVar2;
    }
    trans[0].nFog = trans[0].iTexture = trans[0].iMask = 0;
    if (fTextureVar) {
      ComputeTextureFloor(&trans[0].iTextureVar,
        &trans[0].iMaskVar, xo, yo, fTrue, fFalse);
      ComputeTextureFloor(&trans[0].iTextureVar2,
        &trans[0].iMaskVar2, xo, yo, fTrue, fTrue);
      trans[0].fMaskSetVar = bTextureVar != NULL && Mg(bTextureVar, xo, yo);
    } else
      trans[0].iTextureVar = trans[0].iMaskVar =
        trans[0].iTextureVar2 = trans[0].iMaskVar2 = 0;
  }

  // Loop over each column in the window, to draw each column separately.
  for (x = 0; x < ws.xpClient; x++) {

    // Figure out the angle of this ray, the corresponding horizontal and
    // vertical increments to get from cell to cell, and the characteristics
    // of the starting cell.
    zi = dr.rgcalc[x].rAngleColumn;
    z = zl - zi;
    if (z < 0.0)
      z += rDegMax;
    else if (z >= rDegMax)
      z -= rDegMax;
    xq = (z >= rDegHalf) ? 1 : -1;
    yq = (z >= rDegQuad && z < rDeg34) ? 1 : -1;
    zj = z - ((real)(dr.dir*90));
    if (zj > rDegHalf)
      zj -= rDegMax;
    if FOdd(dr.dir) {
      mi = xq; ni = yq;
      ml = xl; nl = yl;
    } else {
      mi = yq; ni = xq;
      ml = yl; nl = xl;
    }
    mg = AFromInside(ml); ng = AFromInside(nl);
    mt = RAbs(RTanD(zj));
    nt = RDiv(1.0, mt);
    me = InsideFromA0(mg + (mi > 0));
    ne = InsideFromA0(ng + (ni > 0));
    ms = ns = dr.zCell;
    if (dr.fNarrow) {
      if (FOdd(mg))
        ms = dr.zCellNarrow;
      if (FOdd(ng))
        ns = dr.zCellNarrow;
    }
    rT = (real)(dr.zCell << 1);
    md0 = RSqr(LengthI(rT, mt)) / rT;
    nd0 = RSqr(LengthI(rT, nt)) / rT;
    md = DistanceM();
    nd = DistanceN();
    fBeenIn = f3D ? FLegalCubeLevel(bm.a, xo, yo) : FLegal(bm.a, xo, yo);

    fInMark = fInMark2 = fIn3D = fIn3D2 = fFalse;
    if (fExtra || fTextureVar) {
      if (f3D && !Qg3(bm.a, xo, yo, dr.z + 1)) {
        fIn3D = fTrue;
        yExtra = ypWallMax;
      } else if (fMark && Mg3(bm.a2, xo, yo, dr.z, f3D)) {
        fInMark = fTrue;
        yExtra = ypWallMax;
      }
      if (f3D && !Qg3(bm.a, xo, yo, dr.z - 1)) {
        fIn3D2 = fTrue;
        yExtra2 = ypWallMax;
      } else if (fMark2 && Mg3(bMark2, xo, yo, dr.z, f3D)) {
        fInMark2 = fTrue;
        yExtra2 = ypWallMax;
      }
      if (fInMark || fInMark2 || fTextureVar) {
        if (fColorMark) {
          GetKvMark(kvMark, cMark, xo, yo);
        }
        if (fColorMark2) {
          GetKvMark(kvMark2, cMark2, xo, yo);
        }
        if (fTextureDirt || fTextureVar) {
          if (fTextureDirt) {
            ComputeTextureFloor(&iTextureMark,  &iMaskMark,  xo, yo,
              fFalse, fFalse);
            ComputeTextureFloor(&iTextureMark2, &iMaskMark2, xo, yo,
              fFalse, fTrue);
          }
          i = !dr.fNarrow || FOdd(xo) ? dr.zCell : dr.zCellNarrow;
          rx1 = (real)(dr.nOffsetX + (i >> 1)) / (real)i;
          i = !dr.fNarrow || FOdd(yo) ? dr.zCell : dr.zCellNarrow;
          ry1 = (real)(dr.nOffsetY + (i >> 1)) / (real)i;
          rx3 = 1.0 - rx1; ry3 = ry1;
          Assert(FBetween(rx1, 0.0, 1.0) && FBetween(ry1, 0.0, 1.0));
        }
        if (fTextureVar && fWallVarStart) {
          trans[0].rx1 = rx1; trans[0].ry1 = ry1;
          trans[0].rx2 = rx2; trans[0].ry2 = ry2;
          if (trans[0].iTextureVar > 0 || trans[0].iMaskVar > 0 ||
            trans[0].iTextureVar2 > 0 || trans[0].iMaskVar2 > 0)
            fDoTextureVar = fTrue;
        }
      }
    }
    fInTrans = fTransStart;
    fInWallVar = fWallVarStart;
    cTrans = fInTrans || fInWallVar;

    // Shoot the ray. Jump from cell to neighboring cell until a wall is hit,
    // not giving up until the clipping plane is reached.
    for (i = 0; i < dr.nClip; i++) {
      if (md < nd)
        mg += mi;
      else
        ng += ni;
      if (FOdd(dr.dir)) {
        xg = mg; yg = ng;
      } else {
        xg = ng; yg = mg;
      }

      // Is the ray not on the bitmap?
      if (f3D ? !FLegalCubeLevel(bm.a, xg, yg) : !FLegal(bm.a, xg, yg)) {

        // If the ray from on to off the bitmap, do some cleanup, such as
        // finish the current floor marking in progress.
        if (fBeenIn) {
          fDoEnd = fTrue;
          if (dr.fSealEntrance)
            fDoSeal = fDoCompute = fTrue;
          if (fTrans && dr.nTrans == nTransVery && fInTrans)
            fDoExitTrans = fDoCompute = fTrue;
          if (fWallVar && fInWallVar)
            fDoExitWallVar = fDoCompute = fTrue;
        }

        // Performance: If off the bitmap, and moving farther away from the
        // bitmap, a wall will never be hit, so finish this column now.
        if ((yo < 0 && yg < yo) || (xo < 0 && xg < xo) ||
          (yo >= ymax && yg > yo) || (xo >= xmax && xg > xo))
          break;
      } else
        fBeenIn = fTrue;

      // Check for existence of a semitransparent or variable height wall. If
      // so, remember it for later, then keep searching for a solid wall.
      if (fTrans) {
        if (Mg3(bm.a3, xg, yg, dr.z, f3D) && cTrans < itransMac)
          fDoTrans = fDoCompute = fTrue;
        else if (fInTrans) {
          fDoExitTrans = fTrue;
          if (dr.nTrans == nTransVery)
            fDoCompute = fTrue;
        }
      }
      if (fWallVar) {
        if ((fDoTrans || Mg3(bm.a, xg, yg, dr.z, f3D)) && FWallVar(xg, yg) &&
          cTrans < itransMac)
          fDoWallVar = fDoCompute = fTrue;
        else if (fInWallVar)
          fDoExitWallVar = fDoCompute = fTrue;
      }
      if (fDoTrans || fDoWallVar)
        goto LNotSet;

      // Is this a solid wall? If so go draw it.
      if (Mg3Fast(bm.a, xg, yg, dr.z, f3D)) {
        fDoWall = (cTrans < itransMac);
        fDoEnd = fDoCompute = fTrue;
      } else {

        // This is not a solid wall. There are things that might need drawing
        // in the floor or ceiling area at the current cell as the search
        // continues for the wall in the distance. Colored floor and ceiling
        // cell markings are implemented by getting the vertical coordinate as
        // if there were a wall there, then drawing the column when the floor
        // changes color. Pits in the floor and ceiling are the same as floor
        // markings, just that columns are colored appropriately based on what
        // direction they exit the cell, i.e. what side of the pit seems to
        // face you.
LNotSet:
        if (fExtra) {
          if (f3D) {
            if (!fIn3D && !Qg3(bm.a, xg, yg, dr.z + 1)) {
              fDoEnter3D = fDoCompute = fTrue;
              if (fInMark)
                fDoExitMark = fTrue;
            } else if (fIn3D && Qg3(bm.a, xg, yg, dr.z + 1))
              fDoExit3D = fDoCompute = fTrue;
            if (!fIn3D2 && !Qg3(bm.a, xg, yg, dr.z - 1)) {
              fDoEnter3D2 = fDoCompute = fTrue;
              if (fInMark2)
                fDoExitMark2 = fTrue;
            } else if (fIn3D2 && Qg3(bm.a, xg, yg, dr.z - 1))
              fDoExit3D2 = fDoCompute = fTrue;
          }
          if (fMark) {
            if (!fInMark && (!fIn3D || fDoExit3D) && !fDoEnter3D &&
              Mg3(bm.a2, xg, yg, dr.z, f3D))
              fDoEnterMark = fDoCompute = fTrue;
            else if (fInMark) {
              if (!Mg3(bm.a2, xg, yg, dr.z, f3D))
                fDoExitMark = fDoCompute = fTrue;
              else {
                if (fColorMark) {
                  kvMarkOld = kvMark;
                  GetKvMark(kvMark, cMark, xg, yg);
                  if (kvMark != kvMarkOld)
                    fDoExitMark = fDoEnterMark = fDoCompute = fTrue;
                }
                if (fTextureDirt)
                  fDoExitMark = fDoEnterMark = fDoCompute = fTrue;
              }
            }
          }
          if (fMark2) {
            if (!fInMark2 && (!fIn3D2 || fDoExit3D2) && !fDoEnter3D2 &&
              Mg3(bMark2, xg, yg, dr.z, f3D))
              fDoEnterMark2 = fDoCompute = fTrue;
            else if (fInMark2) {
              if (!Mg3(bMark2, xg, yg, dr.z, f3D))
                fDoExitMark2 = fDoCompute = fTrue;
              else {
                if (fColorMark2) {
                  kvMarkOld2 = kvMark2;
                  GetKvMark(kvMark2, cMark2, xg, yg);
                  if (kvMark2 != kvMarkOld2)
                    fDoExitMark2 = fDoEnterMark2 = fDoCompute = fTrue;
                }
                if (fTextureDirt)
                  fDoExitMark2 = fDoEnterMark2 = fDoCompute = fTrue;
              }
            }
          }
        }

        // Ran off the end of the clipping plane, so leave current column.
        if (i >= dr.nClip - 1)
          fDoEnd = fTrue;
      }

      // If ending column, indicate anything in progress should be finished.
      if (fDoEnd) {
        if (fExtra) {
          if (fIn3D)
            fDoExit3D = fDoCompute = fTrue;
          else if (fInMark)
            fDoExitMark = fDoCompute = fTrue;
          if (fIn3D2)
            fDoExit3D2 = fDoCompute = fTrue;
          else if (fInMark2)
            fDoExitMark2 = fDoCompute = fTrue;
        }
        if (fInTrans)
          fDoExitTrans = fDoCompute = fTrue;
        if (fInWallVar)
          fDoExitWallVar = fDoCompute = fTrue;
      }
      if (fDoCompute) {
        zd = Min(md, nd) * dr.rgcalc[x].rDistCorrect;
        if (zd < 1.0)
          zd = 1.0;
        rT = rScale / zd;
        y = rT > (real)ypWallMax ? ypWallMax : (int)rT;
        yExt = dr.zElev >= 0 ? y :
          y - LMultShift(y, iVar2 - (iVar2 * dr.zWall / zWall), iScale2);
        fDoCompute = fFalse;
      }

      // There's a semitransparent or variable height wall here. Compute its
      // coordinates for later drawing, then keep searching for a solid wall.
      if (fDoTextureVar) {
        ptransOld = cTrans > 0 ? &trans[cTrans-1] : NULL;
        ComputeDir(k);
        ComputeTextureOffset(xg, yg);
        if (ptransOld != NULL) {
          if (ptransOld->iTextureVar > 0 || ptransOld->iMaskVar > 0) {
            ComputeTextureOffsetFloor2(ptransOld->rx2, ptransOld->ry2);
          }
          if (ptransOld->iTextureVar2 > 0 || ptransOld->iMaskVar2 > 0) {
            ComputeTextureOffsetFloor4(ptransOld->rx4, ptransOld->ry4);
          }
        }
        fDoTextureVar = fFalse;
      }
      if (fDoTrans || fDoWallVar) {
        ptransOld = cTrans > 0 ? &trans[cTrans-1] : NULL;
        GetKvIf(kvT, xg, yg, dr.z, kvWall);
        fAdd = cTrans <= 0 || kvT != ptransOld->kvOrig ||
          (fDoTrans && (!fInTrans || fDoExitWallVar)) ||
          (fDoWallVar && (!fInWallVar || (!fDoTrans && ptransOld->fTrans)));
        if (fAdd || fDoWallVar || fTextureWall || fTextureVar) {

          // Add new entry to the semitransparent / variable height wall list.
          ptransCur = &trans[cTrans];
          if (fDoWallVar)
            ComputeWallVar(fWallVar2, bWallVar, cWallVar, zWall, xg, yg,
              &ptransCur->yLo, &ptransCur->yHi);
          else {
            ptransCur->yLo = 0; ptransCur->yHi = iVar2;
          }
          if (fTextureWall) {
            ComputeDir(k);
            ComputeTexture(&ptransCur->iTexture, &ptransCur->iMask,
              &ptransCur->rTexture, dr.nTextureWall,
              xg, yg, 0, k, xq, yq, me, ml, ne, nl, md, nd, mt, nt);
            ptransCur->fMaskSet =
              ws.rgbMask[0] != NULL && Mg(ws.rgbMask[0], xg, yg);
          } else
            ptransCur->iTexture = ptransCur->iMask = 0;
          if (fTextureVar) {
            ComputeTextureFloor(&ptransCur->iTextureVar,
              &ptransCur->iMaskVar, xg, yg, fTrue, fFalse);
            ComputeTextureFloor(&ptransCur->iTextureVar2,
              &ptransCur->iMaskVar2, xg, yg, fTrue, fTrue);
            if (ptransCur->iTextureVar > 0 || ptransCur->iMaskVar > 0 ||
              ptransCur->iTextureVar2 > 0 || ptransCur->iMaskVar2 > 0) {
              ComputeDir(k);
              ComputeTextureOffset(xg, yg);
              if (ptransCur->iTextureVar > 0 || ptransCur->iMaskVar > 0) {
                ComputeTextureOffsetFloor1(ptransCur->rx1, ptransCur->ry1);
              }
              if (ptransCur->iTextureVar2 > 0 || ptransCur->iMaskVar2 > 0) {
                ComputeTextureOffsetFloor3(ptransCur->rx3, ptransCur->ry3);
              }
              ptransCur->fMaskSetVar =
                bTextureVar != NULL && Mg(bTextureVar, xg, yg);
              fDoTextureVar = fTrue;
            }
          } else
            ptransCur->iTextureVar = ptransCur->iMaskVar =
              ptransCur->iTextureVar2 = ptransCur->iMaskVar2 = 0;
          if (fAdd || ptransCur->iTexture > 0 || ptransCur->iMask > 0 ||
            ptransCur->yLo != ptransOld->yLo ||
            ptransCur->yHi != ptransOld->yHi || fTextureVar) {
            ptransCur->y = ptransCur->y2 = y;
            ptransCur->fTrans = fDoTrans;
            ptransCur->nFog =
              fFog ? (int)(Min(zd / rFogDist, 1.0) * 100.0) : 0;
            ptransCur->kvOrig = kvT;
            ShadeKv(kvT);
            ptransCur->kv = kvT;
            if ((fInTrans || fInWallVar) && ptransOld != NULL)
              ptransOld->y2 = ptransCur->y;
            cTrans++;
          }
        }
        fInTrans = fDoTrans;
        fInWallVar = fDoWallVar;
        fDoTrans = fDoWallVar = fFalse;
      }
      if (fDoExitTrans || fDoExitWallVar) {
        if (cTrans > 0) {
          ptransOld = &trans[cTrans-1];
          ptransOld->y2 = y;
        }
        if (fDoExitTrans)
          fInTrans = fDoExitTrans = fFalse;
        if (fDoExitWallVar)
          fInWallVar = fDoExitWallVar = fFalse;
      }

      if (fExtra) {
        if (fDoExit3D) {
          Ensure3DColor(xg, yg, 1);
          LineI(x, yb+IO(yExt), yb+IO(yExtra), kv3D);
          DrawTexture3D(1);
          fIn3D = fDoExit3D = fFalse;
        } else if (fDoExitMark) {
          if (iTextureMark == 0 || iMaskMark > 0)
            LineYFloor(c, x, yb, IO(yExt), IO(yExtra),
              fDoEnterMark ? kvMarkOld : kvMark, 0, rScale);
          if (fTextureDirt && (iTextureMark > 0 || iMaskMark > 0)) {
            ComputeDir(k);
            ComputeTextureOffset(xo, yo);
            ComputeTextureOffsetFloor2(rx2, ry2);
            LineYFloorTexture(c, x, yb, yExt, yExtra,
              iMaskMark > 0 ? ws.rgbMask[iMaskMark] : NULL,
              iTextureMark > 0 ? ws.rgcTexture[iTextureMark] : NULL,
              rx2, ry2, rx1, ry1,
              bTextureDirt != NULL && Mg(bTextureDirt, xo, yo), rScale);
          }
          fInMark = fDoExitMark = fFalse;
        }
        if (fDoExit3D2) {
          Ensure3DColor(xg, yg, -1);
          LineI(x, yb-IO(yExtra2), yb-IO(yExt), kv3D);
          DrawTexture3D(-1);
          fIn3D2 = fDoExit3D2 = fFalse;
        } else if (fDoExitMark2) {
          if (iTextureMark2 == 0 || iMaskMark2 > 0)
            LineYFloor(c, x, yb, -IO(yExtra2), -IO(yExt),
              fDoEnterMark2 ? kvMarkOld2 : kvMark2, 0, rScale);
          if (fTextureDirt && (iTextureMark2 > 0 || iMaskMark2 > 0)) {
            ComputeDir(k);
            ComputeTextureOffset(xo, yo);
            ComputeTextureOffsetFloor4(rx4, ry4);
            LineYFloorTexture(c, x, yb, iRound - yExtra2, iRound - yExt,
              iMaskMark2 > 0 ? ws.rgbMask[iMaskMark2] : NULL,
              iTextureMark2 > 0 ? ws.rgcTexture[iTextureMark2] : NULL,
              rx3, ry3, rx4, ry4,
              bTextureDirt != NULL && Mg(bTextureDirt, xo, yo), rScale);
          }
          fInMark2 = fDoExitMark2 = fFalse;
        }

        if (fDoEnter3D) {
          yExtra = yExt;
          fIn3D = fTrue;
          fDoEnter3D = fFalse;
        } else if (fDoEnterMark) {
          yExtra = yExt;
          if (fColorMark) {
            GetKvMark(kvMark, cMark, xg, yg);
          }
          if (fTextureDirt) {
            ComputeTextureFloor(&iTextureMark, &iMaskMark, xg, yg,
              fFalse, fFalse);
            if (iTextureMark > 0 || iMaskMark > 0) {
              ComputeDir(k);
              ComputeTextureOffset(xo, yo);
              ComputeTextureOffsetFloor1(rx1, ry1);
            }
          }
          fInMark = fTrue;
          fDoEnterMark = fFalse;
        }
        if (fDoEnter3D2) {
          yExtra2 = yExt;
          fIn3D2 = fTrue;
          fDoEnter3D2 = fFalse;
        } else if (fDoEnterMark2) {
          yExtra2 = yExt;
          if (fColorMark2) {
            GetKvMark(kvMark2, cMark2, xg, yg);
          }
          if (fTextureDirt) {
            ComputeTextureFloor(&iTextureMark2, &iMaskMark2, xg, yg,
              fFalse, fTrue);
            if (iTextureMark2 > 0 || iMaskMark2 > 0) {
              ComputeDir(k);
              ComputeTextureOffset(xo, yo);
              ComputeTextureOffsetFloor3(rx3, ry3);
            }
          }
          fInMark2 = fTrue;
          fDoEnterMark2 = fFalse;
        }
      }

      if (fDoEnd) {
        if (fDoSeal) {
          LineI(x, yb-IO(y), yb+IO(y), dr.kvInEdge);
          fDoSeal = fFalse;
        }
        fDoEnd = fFalse;
        break;
      }

      // Follow the ray to the next neighboring cell.
      xo = xg; yo = yg;
      if (md < nd) {
        me += mi*ms;
        if (dr.fNarrow)
          ms = (ms == dr.zCell ? dr.zCellNarrow : dr.zCell);
        md = DistanceM();
      } else {
        ne += ni*ns;
        if (dr.fNarrow)
          ns = (ns == dr.zCell ? dr.zCellNarrow : dr.zCell);
        nd = DistanceN();
      }
    }

    if (fDoWall) {
      fDoWall = fFalse;

      // Do texture mapping for the wall at the current column.
      ComputeDir(k);
      if (fTextureWall) {
        ComputeTexture(&iTexture, &iMask, &rTexture, dr.nTextureWall,
          xg, yg, 0, k, xq, yq, me, ml, ne, nl, md, nd, mt, nt);
        if (iTexture > 0 && iMask == 0) {
          DrawTexture(c, x, y, 0, IP(yb), iTexture, 0, rTexture,
            0, fFog ? (int)(Min(zd / rFogDist, 1.0) * 100.0) : 0, fFalse);
          goto LNext;
        }
      }

      // Figure out what color to draw the wall at the current column in.
      if (fColor) {
        GetKv(kv1, bm.c, xg, yg, dr.z, kvWall);
      }
      if (k != kOld || kv1 != kv1Old || fFog) {
        kOld = k; kv1Old = kv1;
        if (k & 1)
          kv2 = kv1;
        else if (k == 0)
          kv2 = KvShade(kv1, -dr.rLightFactor);
        else
          kv2 = KvShade(kv1, dr.rLightFactor);
        if (fFog)
          kv2 = KvApplyFog(kv2);
      }

      // Actually draw the wall for the current column here.
      LineYGradient(c, x, yb-IO(y), yb+IO(y), kv2, dr.kvInWall2);

      // Do texture mapping on top of the drawn wall at the current column.
      if (fTextureWall && iMask > 0)
        DrawTexture(c, x, y, 0, IP(yb), iTexture, iMask, rTexture,
          0, fFog ? (int)(Min(zd / rFogDist, 1.0) * 100.0) : 0,
          ws.rgbMask[0] != NULL && Mg(ws.rgbMask[0], xg, yg));
    }

LNext:
    // Finally draw the list of semitransparent lines over top of this column.
    if ((fTrans || fWallVar) && cTrans > 0)
      DrawTransparent(c, x, yb, cTrans, trans, rScale);
    xo = AFromInside(xl); yo = AFromInside(yl);
  }
  return fTrue;
}


/*
******************************************************************************
** Simple Perspective Inside Display
******************************************************************************
*/

#define ZCell(z) ((z) * 3 / 5)
#define ZCellNarrow(z) \
  ((z) * (32 - 16 * dr.zCellNarrow / dr.zCell) / 32)

// Draw a line on the window twice, at the given coordinates and also at the
// coordinates flipped vertically. Used by RedrawInside(), where the top and
// bottom halves of the window when drawing a Maze in simple perspective mode
// are mirror images of each other.

void DrawInside(colmap c, int x1, int y1, int x2, int y2, int i)
{
  int x = ws.xpClient-1, y = ws.ypClient-1;

  i = i > 0;
  LineC(c, i ? x1 : x-x1, y1,   i ? x2 : x-x2, y2,   dr.kvInEdge);
  LineC(c, i ? x1 : x-x1, y-y1, i ? x2 : x-x2, y-y2, dr.kvInEdge);
}


// Draw the simple mode version of 3D first person perspective inside view.
// This only works for 2D Mazes, and can only display views looking straight
// down passages. The algorithm draws lines indicating corners between walls.
// There's a finite set of lines, where it's a matter of determining whether
// or not to draw each line.

void RedrawInside(colmap c)
{
  KV kv = dr.kvInWall, kvT;
  int xc = ws.xpClient >> 1, yc = ws.ypClient >> 1,
    xm, ym, xn, yn, xd, yd, xe, ye, xl, yl, xf, yf, xx, i, j, k;

  // Initialize Windows variables for drawing walls and such.
  if (bm.c != NULL && FLegal(bm.c, dr.x, dr.y)) {
    kvT = CG(bm.c, dr.x, dr.y);
    if (kvT != dr.kvInEdge)
      kv = kvT;
  }
  ColmapSet(c, kv);

  // Draw the left and right halves of the window separately.
  for (i = -1; i <= 1; i += 2) {
    xm = dr.x; ym = dr.y;
    j = (dr.dir + i) & DIRS1;
    xd = xm + xoff[j]; yd = ym + yoff[j];
    xl = (xc * 3 / 4) * 5 / 3; yl = (yc * 3 / 4) * 5 / 3;
    xx = 0;

    // Loop over each cell looking down the passage from the viewing point.
    for (j = 0; j < 20; j++) {
      xn = xm + xoff[dr.dir]; yn = ym + yoff[dr.dir];
      xe = xd + xoff[dr.dir]; ye = yd + yoff[dr.dir];
      if (!dr.fNarrow || !FOdd(j)) {
        xf = ZCell(xl); yf = ZCell(yl);
      } else {
        xf = ZCellNarrow(xl); yf = ZCellNarrow(yl);
      }

      // Done when a wall is reached, or when the walls will be drawn at a
      // size of one pixel.
      if (Mg(bm.a, xm, ym) ||
        ((xl == xf || yl == yf) && (dr.zCellNarrow != 0 || !FOdd(j))))
        break;

      // For passages that go outside the Maze, draw a horizon line.
      if (Mg(bm.a, xd, yd)) {
        DrawInside(c, xc-xl, yc-yl, xc-xf, yc-yf, i);
        xx = xc-xf;
      } else if (Mg(bm.a, xe, ye)) {
        DrawInside(c, xx, yc-yf, xc-xf, yc-yf, i);
        xx = xc-xf;
      }

      // Draw lines on the window where appropriate.
      if (Mg(bm.a, xn, yn)) {
        DrawInside(c, xc-xf, yc-yf, xc, yc-yf, i);
        if (Mg(bm.a, xd, yd) || !Mg(bm.a, xe, ye))
          DrawInside(c, xc-xf, yc-yf, xc-xf, yc, i);
        if (!Mg(bm.a, xe, ye)) {
          if (j == 0)
            DrawInside(c, xc-xl, yc-yf+(yl-yf), xc-xf, yc-yf, i);
          else
            DrawInside(c, xx, yc, xc-xf, yc, i);
        }
      } else if (Mg(bm.a, xd, yd) != Mg(bm.a, xe, ye)) {
        DrawInside(c, xc-xf, yc-yf, xc-xf, yc, i);
        xx = xc-xf;
      }

      // Draw "X" on the ground if marking there.
      if (bm.a2 != NULL && FLegal(bm.a2, xm, ym) && MG(bm.a2, xm, ym) &&
        !Mg(bm.a, xm, ym) && j > 0 && (!dr.fNarrow || !FOdd(j))) {
        k = yc + yl+(yf-yl)*63/100;
        LineC(c, xc + i*xl*50/100, yc + yl+(k-yc-yl)*50/100, xc, k,
          dr.kvInEdge);
        LineC(c, xc + i*xf*63/100, yc + yf+(k-yc-yf)*37/100, xc, k,
          dr.kvInEdge);
      }

      // Figure out the coordinates for the next cell down the passage.
      xm = xn; ym = yn;
      xn += xoff[dr.dir]; yn += yoff[dr.dir];
      xd = xe; yd = ye;
      xe += xoff[dr.dir]; ye += yoff[dr.dir];
      xl = xf; yl = yf;
    }
    if (yl == yf && (dr.zCellNarrow != 0 || !FLegal(bm.a, xm, ym)))
      DrawInside(c, xx, yc, xc, yc, i);
  }

  // Decorate the exposed parts of the sky and ground.
  if (!Mg(bm.a, dr.x, dr.y)) {
    FFillC(c, xc, ws.ypClient-1, dr.kvInDirt, -1, fFalse);
    FFillC(c, xc, 0,             dr.kvInSky,  -1, fFalse);
  }
  if (dr.fStar)
    FCreateInsideStars(c, (real)(dr.dir * 90), 0, fFalse);
}


// Draw a line on the window, flipping the horizontal and vertical coordinates
// when specified.

void DrawInside3(colmap c, int x1, int y1, int x2, int y2, int i, int j)
{
  int x = ws.xpClient-1, y = ws.ypClient-1;

  i = i > 0;
  LineC(c, i ? x1 : x-x1, j ? y-y1 : y1,
    i ? x2 : x-x2, j ? y-y2 : y2, dr.kvInEdge);
}


// A version of RedrawInside() that works for 3D Mazes.

void RedrawInside3(colmap c)
{
  KV kv = dr.kvIn3D, kvT;
  int xc = ws.xpClient >> 1, yc = ws.ypClient >> 1,
    xm, ym, z, xn, yn, xd, yd, xe, ye, xl, yl, xf, yf, i, j, k;

  // Initialize Windows variables for drawing walls and such.
  if (bm.c != NULL && FLegal(bm.c, dr.x, dr.y)) {
    kvT = CG(bm.c, dr.x, dr.y);
    if (kvT != dr.kvInEdge)
      kv = kvT;
  }
  ColmapSet(c, kv);

  // Draw the left and right halves of the window separately.
  for (i = -1; i <= 1; i += 2) {
    xm = dr.x; ym = dr.y; z = dr.z;
    j = (dr.dir + i) & DIRS1;
    xd = xm + xoff[j]; yd = ym + yoff[j];
    xl = (xc * 3 / 4) * 5 / 3; yl = (yc * 3 / 4) * 5 / 3;

    // Loop over each cell looking down the passage from the viewing point.
    for (j = 0; j < 20; j++) {
      xn = xm + xoff[dr.dir]; yn = ym + yoff[dr.dir];
      xe = xd + xoff[dr.dir]; ye = yd + yoff[dr.dir];
      if (!dr.fNarrow || !FOdd(j)) {
        xf = ZCell(xl); yf = ZCell(yl);
      } else {
        xf = ZCellNarrow(xl); yf = ZCellNarrow(yl);
      }

      // Done when a wall is reached, or when the walls will be drawn at a
      // size of one pixel.
      if (Qg(bm.a, xm, ym, z) ||
        ((xl == xf || yl == yf) && (dr.zCellNarrow != 0 || !FOdd(j))))
        break;

      // Draw lines on the window where appropriate.
      if (Qg(bm.a, xm, ym, z-1) ^ Qg(bm.a, xn, yn, z-1) ^ Qg(bm.a, xn, yn, z))
        DrawInside3(c, xc-xf, yc-yf, xc, yc-yf, i, fFalse);
      if (Qg(bm.a, xm, ym, z+1) ^ Qg(bm.a, xn, yn, z+1) ^ Qg(bm.a, xn, yn, z))
        DrawInside3(c, xc-xf, yc-yf, xc, yc-yf, i, fTrue);
      if (!Qg(bm.a, xd, yd, z) && Qg(bm.a, xe, ye, z))
        DrawInside(c, xc-xl, yc-yf, xc-xf, yc-yf, i);
      if (Qg(bm.a, xe, ye, z) ^ Qg(bm.a, xd, yd, z) ^ Qg(bm.a, xn, yn, z))
        DrawInside(c, xc-xf, yc-yf, xc-xf, yc, i);
      if (!Qg(bm.a, xm, ym, z-1) && Qg(bm.a, xn, yn, z-1))
        DrawInside3(c, xc-xf, yc-yl, xc-xf, yc-yf, i, fFalse);
      if (!Qg(bm.a, xm, ym, z+1) && Qg(bm.a, xn, yn, z+1))
        DrawInside3(c, xc-xf, yc-yl, xc-xf, yc-yf, i, fTrue);
      if (Qg(bm.a, xd, yd, z) ^ Qg(bm.a, xm, ym, z-1) ^ Qg(bm.a, xd, yd, z-1))
        DrawInside3(c, xc-xl, yc-yl, xc-xf, yc-yf, i, fFalse);
      if (Qg(bm.a, xd, yd, z) ^ Qg(bm.a, xm, ym, z+1) ^ Qg(bm.a, xd, yd, z+1))
        DrawInside3(c, xc-xl, yc-yl, xc-xf, yc-yf, i, fTrue);

      // Draw "X" on the ground if marking there.
      if (bm.a2 != NULL && FLegalMaze3(bm.a2, xm, ym, z) &&
        TG(bm.a2, xm, ym, z) && !Qg(bm.a, xm, ym, z) && j > 0 &&
        (!dr.fNarrow || !FOdd(j))) {
        k = yc + yl+(yf-yl)*63/100;
        LineC(c, xc + i*xl*50/100, yc + yl+(k-yc-yl)*50/100, xc, k,
          dr.kvInEdge);
        LineC(c, xc + i*xf*63/100, yc + yf+(k-yc-yf)*37/100, xc, k,
          dr.kvInEdge);
      }

      // Figure out the coordinates for the next cell down the passage.
      xm = xn; ym = yn;
      xn += xoff[dr.dir]; yn += yoff[dr.dir];
      xd = xe; yd = ye;
      xe += xoff[dr.dir]; ye += yoff[dr.dir];
      xl = xf; yl = yf;
    }
  }

  // For passages that go outside the Maze, fill them with the ground or
  // sky color to indicate an entrance or exit door.
  if ((ym < 0 && dr.dir == 0) || (ym >= TY(bm.a) && dr.dir == 2))
    FFillC(c, xc, yc, ym < 0 ? dr.kvInDirt : dr.kvInSky, -1, fFalse);
}


// Draw a Maze contained in one bitmap as line segments overlaying another
// bitmap. Implements the Map Overlay setting in perspective inside view.

void DrawOverlay(colmap c, bitmap b)
{
  int zd, zs, xc, yc, d, xx, xy, yx, yy, x0, y0, xo, yo,
    x1, y1, x2, y2, x, y, xp, yp, zT;
  flag fX, fY, fArrow;

  // Determine segment length and how much of the Maze is visible.
  xc = MX(c) >> 1; yc = MY(c) >> 1;
  if (bm.nWhat != 1) {
    zd = Min(bm.xCell, bm.yCell) << 1;
  } else {
    zd = Max(CX(c), CY(c)) / Min(bm.xCell, bm.yCell);
  }
  zd = Max(zd, 10);
  zs = Max(CX(c), CY(c)) / zd;
  zs = NSqr(Sq(zs) << 1);

  // Determine x and y pixel offsets of the rotated Maze axes.
  d = nDegMax - (dr.dir * 90 + dr.nOffsetD);
  if (d > nDegMax)
    d -= nDegMax;
  xx = (int)((real)zd * RCosD((real)d));
  xy = (int)((real)zd * RSinD((real)d));
  yx = (int)((real)zd * RCosD((real)(d + 90)));
  yy = (int)((real)zd * RSinD((real)(d + 90)));

  // Determine the offset of the dot within the Maze.
  if (!dr.f3D) {
    x0 = dr.x; y0 = dr.y;
  } else {
    x0 = Tx(b, dr.x, dr.z); y0 = Ty(b, dr.y, dr.z);
  }
  if (dr.fNarrow) {
    zT = dr.zCell + dr.zCellNarrow;
    x = dr.nOffsetX - FOdd(x0)*(zT >> 1);
    y = dr.nOffsetY - FOdd(y0)*(zT >> 1);
  } else {
    zT = dr.zCell;
    x = dr.nOffsetX;
    y = dr.nOffsetY;
  }
  xo = xc - (x * xx + y * xy) / zT;
  yo = yc - (x * yx + y * yy) / zT;

  x1 = x0 - zs; x1 = Max(x1, 0);
  y1 = y0 - zs; y1 = Max(y1, 0);
  x2 = x0 + zs; x2 = Min(x2, MX(b)-1);
  y2 = y0 + zs; y2 = Min(y2, MY(b)-1);

  // Draw the Maze overlay.
  if (dr.fNarrow) {

    // Narrow Walls on means each segment spans two Maze bitmap pixels.
    x1 &= ~1; y1 &= ~1; x2 &= ~1; y2 &= ~1;
    for (y = y1; y <= y2; y += 2)
      for (x = x1; x <= x2; x += 2)
        if (MGCore(b, x, y)) {
          xp = xo + ((x - x0) >> 1) * xx + ((y - y0) >> 1) * xy;
          yp = yo + ((x - x0) >> 1) * yx + ((y - y0) >> 1) * yy;
          if (MgFast(b, x+1, y))
            LineC(c, xp, yp, xp + xx, yp - xy, dr.kvMap);
          if (MgFast(b, x, y+1))
            LineC(c, xp, yp, xp - yx, yp + yy, dr.kvMap);
        }
  } else {

    // Narrow Walls off means each segment spans one Maze bitmap pixel.
    for (y = y1; y <= y2; y++)
      for (x = x1; x <= x2; x++)
        if (MGCore(b, x, y)) {
          xp = xo + (x - x0) * xx + (y - y0) * xy;
          yp = yo + (x - x0) * yx + (y - y0) * yy;
          fX = MgFast(b, x+1, y); fY = MgFast(b, x, y+1);
          if (fX)
            LineC(c, xp, yp, xp + xx, yp - xy, dr.kvMap);
          if (fY)
            LineC(c, xp, yp, xp - yx, yp + yy, dr.kvMap);

          // Draw diagonal lines if no x or y axis connections available.
          if (!fX && !fY && MgFast(b, x+1, y+1))
            LineC(c, xp, yp, xp + xx - yx, yp - xy + yy, dr.kvMap);
          if (!fY && !MgFast(b, x-1, y) && MgFast(b, x-1, y+1))
            LineC(c, xp, yp, xp - xx - yx, yp + xy + yy, dr.kvMap);
        }
  }

  // Draw the 2nd dot.
  d = (zd << 1) / 3;
  x = dr.x2; y = dr.y2;
  if (dr.fNarrow) {
    x |= 1; y |= 1; d >>= 1;
  }
  zT = (d << 1) + 1;

  if (dr.fDot2 && FBetween(x, x1, x2) && FBetween(y, y1, y2)) {
    if (x == dr.x && y == dr.y && dr.nOffsetX == 0 && dr.nOffsetY == 0) {
      xp = xc; yp = yc;
    } else {
      xp = xo + ((x - x0 - 1) >> dr.fNarrow) * xx +
        ((y - y0 + 1) >> dr.fNarrow) * xy + ((xx - xy) >> dr.fNarrow);
      yp = yo + ((x - x0 - 1) >> dr.fNarrow) * yx +
        ((y - y0 + 1) >> dr.fNarrow) * yy + ((yx - yy) >> dr.fNarrow);
    }
    if (dr.fCircle && zT > 2)
      DiskC(c, xp-d, yp-d, xp+d, yp+d, dr.kvDot2);
    else
      BlockC(c, xp - d, yp - d, xp + d, yp + d, dr.kvDot2);
  }

  // Draw the dot.
  if (dr.fDot) {
    fArrow = dr.fArrow && zT >= 9;
    if (!fArrow || dr.kvDot != dr.kvInEdge) {
      if (dr.fCircle && zT > 2)
        DiskC(c, xc-d, yc-d, xc+d, yc+d, dr.kvDot);
      else
        BlockC(c, xc - d, yc - d, xc + d, yc + d, dr.kvDot);
    }
    if (fArrow) {
      x = xc; y = yc - zT*3/8;
      LineC(c, x, y, x, y + zT*6/8, dr.kvInEdge);
      LineC(c, x, y, x - zT*11/32, y + zT*11/32, dr.kvInEdge);
      LineC(c, x, y, x + zT*11/32, y + zT*11/32, dr.kvInEdge);
    }
  }
}

/* inside.cpp */
