/*
** Daedalus (Version 2.5) File: maze.h
** By Walter D. Pullen, Astara@msn.com, http://www.astrolog.org/labyrnth.htm
**
** IMPORTANT NOTICE: Daedalus and all Maze generation and general
** graphics routines used in this program are Copyright (C) 1998-2013 by
** Walter D. Pullen. Permission is granted to freely use, modify, and
** distribute these routines provided these credits and notices remain
** unmodified with any altered or distributed versions of the program.
** The user does have all rights to Mazes and other graphic output
** they make in Daedalus, like a novel created in a word processor.
**
** More formally: This program is free software; you can redistribute it
** and/or modify it under the terms of the GNU General Public License as
** published by the Free Software Foundation; either version 2 of the
** License, or (at your option) any later version. This program is
** distributed in the hope that it will be useful and inspiring, but
** WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
** General Public License for more details, a copy of which is in the
** LICENSE.HTM included with Daedalus, and at http://www.gnu.org
**
** This file contains definitions for Maze creation algorithms, solving
** algorithms, and other Maze utilities.
**
** Created: 11/22/1996.
** Last code change: 10/10/2013.
*/


#define iVarScale 8
#define iVar (1 << (iVarScale))
#define iCircuitMax 30

enum {
  femsEvenSize  = 0x1,
  femsEvenSize2 = 0x2,
  femsEvenStart = 0x4,
  femsNoResize  = 0x8,
  femsNoSection = 0x10,
  femsOddSize   = 0x20,
  femsMinSize   = 0x40,
  fems64K       = 0x80,
};

enum {
  lcsCircle  = 0,
  lcsRounded = 1,
  lcsSquare  = 2,
  lcsFlat    = 3,
};

enum {
  lccRadius   = 0,
  lccEntrance = 1,
  lccExit     = 2,
  lccBalanced = 3,
  lccMerged   = 4,
};


/*
******************************************************************************
** Types
******************************************************************************
*/

typedef struct _mazesettings {
  // File settings

  long nInfGen;
  long nInfCutoff;
  long nInfFileCutoff;
  flag fInfEller;
  flag fInfFile;
  flag fInfFileDos;
  flag fInfAutoStart;
  int xCutoff;

  // Random settings

  int nRndSeed;
  int nRndBias;
  int nRndRun;

  // Dot settings

  int nRadar;

  // Maze settings

  flag fSection;
  flag fTeleportEntrance;
  flag fPoleNoDeadEnd;
  flag fSolveEveryPixel;
  flag fSolveDotExit;
  flag fRandomPath;
  flag fCountShortest;
  int nTweakPassage;
  long cMaze;
  int nWeaveOff;
  int nWeaveOn;
  int nWeaveRail;
  int nEntrancePos;

  // Labyrinth settings

  int nLabyrinth;
  int lcs;
  int nClassical;
  flag fClassicalCirc;
  char szCustom[cchSzDef];
  flag fCustomAuto;
  int nCustomFold;
  int nCustomAsym;
  int nCustomPartX;
  int nCustomPartY;
  int lcc;
  int zCustomMiddle;

  // Create settings

  flag fRiver;
  flag fRiverEdge;
  flag fRiverFlow;
  int omega;
  int omega2;
  int omegas;
  int omegaf;
  char szPlanair[cchSzMax];
  flag fTreeWall;
  flag fTreeRandom;
  int nTreeRiver;
  int cSpiral;
  int cSpiralWall;
  int cRandomAdd;
  int xFractal;
  int yFractal;
  int zFractal;
  flag fFractalI;
  flag fCrackOff;
  int nCrackLength;
  int nCrackPass;
  int nCrackSector;
  int nSparse;
  flag fKruskalPic;
  flag fWeaveCorner;
  flag fTiltDiamond;

  // Macro accessible only settings

  int nCellMax;
  int nFractalD;
  int nFractalL;

  // Internal settings

  int xEntrance;
  int yEntrance;
  int xExit;
  int yExit;
  int cRunRnd;
  int dirRnd;
  int cRunRndSeg;
  int dirRndSeg;
  long iSpiralIndex;
  FILE *fileInf;
  flag fInfMiddle;
  long nInfCount;
  short *rgnInfEller;
  int cnInfEller;
} MS;

typedef struct _krus {
  struct _krus *next;
  long count;
} KRUS;

extern MS ms;
extern int xl, yl, xh, yh;
extern CONST char *rgszChip[7];


/*
******************************************************************************
** Maze Utilities
******************************************************************************
*/

#define LPrime(l) ((l) % 109L != 0 ? 109L : ((l) % 191L != 0 ? 191L : 199L))
#define RndSkip(n1, n2) ((n1) + (Rnd(0, (((n2) - (n1)) >> 1)) << 1))
#define fCellMax (ms.nCellMax >= 0 && --ms.nCellMax < 0)

#define SetXyh(b) xl = yl = 0; xh = MX(b)-1; yh = MY(b)-1

#define FLegalMaze(x, y) \
  ((x) >= xl && (x) < xh && (y) >= yl && (y) < yh)
#define FLegalMaze2(x, y) \
  ((x) >= xl && (x) <= xh && (y) >= yl && (y) <= yh)

#define Count(b, x, y) \
  (MG(b, (x), (y)-1) + MG(b, (x)-1, (y)) + \
  MG(b, (x), (y)+1) + MG(b, (x)+1, (y)))
#define CountSlow(b, x, y) \
  (Mg(b, (x), (y)-1) + Mg(b, (x)-1, (y)) + \
  Mg(b, (x), (y)+1) + Mg(b, (x)+1, (y)))
#define Count2(b, x, y) \
  (MG(b, (x), (y)-2) + MG(b, (x)-2, (y)) + \
  MG(b, (x), (y)+2) + MG(b, (x)+2, (y)))
#define CountWeave(b, x, y) \
  (Mg(b, (x), (y)-4) + Mg(b, (x)-4, (y)) + \
  Mg(b, (x), (y)+4) + Mg(b, (x)+4, (y)))

#define FOnMaze(b, x, y) \
  (MG(b, (x)-1, (y)-1) && MG(b, (x)+1, (y)-1) && \
  MG(b, (x)-1, (y)+1) && MG(b, (x)+1, (y)+1))
#define FOnMazeSlow(b, x, y) \
  (Mg(b, (x)-1, (y)-1) && Mg(b, (x)+1, (y)-1) && \
  Mg(b, (x)-1, (y)+1) && Mg(b, (x)+1, (y)+1))
#define FOnMaze2(b, x, y) \
  (MG(b, (x)-2, (y)-2) && MG(b, (x)+2, (y)-2) && \
  MG(b, (x)-2, (y)+2) && MG(b, (x)+2, (y)+2))
#define FOnMazeZeta(b, x, y) \
  (MG(b, (x)-1, (y)-2) && MG(b, (x)+2, (y)-1) && \
  MG(b, (x)-2, (y)+1) && MG(b, (x)+1, (y)+2))

#define FOnWall(b, x, y) \
  (MG(b, (x) & ~1, (y) & ~1) & MG(b, (x) + FOdd(x), (y) + FOdd(y)))
#define FOnPassage(b, x, y) \
  (!MG(b, (x) | 1, (y) | 1) && !MG(b, (x) - FOdd(y), (y) - FOdd(x)))

extern flag FMazeSizeError(int, int);
extern flag FEnsureMazeSize(bitmap *, int, int);
extern void BlockOuter(bitmap, bit);
extern void MazeClear(bitmap, bit);
extern int AddEntranceExit(bitmap, flag);
extern void MakeEntranceExit(bitmap, int);
extern flag MakeIsolationDetachment(bitmap, flag);
extern flag FWouldMakeDeadEnd(constbitmap, int, int);
extern flag FWouldMakeIsolation(constbitmap, int, int);
extern int PeekWall(constbitmap, int, int, int, int, flag, int);
extern int FollowWall(constbitmap, int *, int *, int, int);
extern int FollowPassage(constbitmap, int *, int *, int *, int, flag);
extern int PeekRandom(constbitmap, int, int, int, int, flag);
extern flag FFindPassage(constbitmap, int *, int *, flag);
extern int DirFindUncreatedCore(constbitmap, int *, int *, flag);
#define DirFindUncreated(b, x, y) DirFindUncreatedCore(b, x, y, fFalse)
extern int RndDir(void);


/*
******************************************************************************
** Maze Operations
******************************************************************************
*/

#define FMazeZoomBias(b) FBitmapBias(b, 1, 3, 1, 3)
extern long MazeNormalize(bitmap, flag);
extern long MazeZoomAndExpandSetCells(bitmap *);
extern long MazeRoomifyThinnedCells(bitmap);
extern long MazeTweakEndpoints(bitmap);
extern long MazeTweakPassages(bitmap);
extern long DoSetAllCellsToPoles(bitmap);
extern long DoRemoveIsolationDetachment(bitmap, flag);
#define DoCrackIsolations(b) DoRemoveIsolationDetachment(b, fFalse)
#define DoConnectDetachments(b) DoRemoveIsolationDetachment(b, fTrue)
extern long DoConnectPoles(bitmap, flag);
extern long DoDeletePoles(bitmap, flag);
extern long DoCrackDeadEnds(bitmap);
extern long DoCrackEntrances(bitmap);
extern long DoCrackPassages(bitmap);
extern long DoCrackRooms(bitmap);
extern long DoCrackBoundaryWall(bitmap, flag);
extern long DoCrackTubes(bitmap);
extern long DoCrackIsland(bitmap, int, int, flag);
extern long DoCrackIslands(bitmap, flag);
extern long DoMakeSymmetric(bitmap, int, flag);
extern void BlockMoveMaze(bitmap, bitmap,
  int, int, int, int, int, int, int, int, int, int);
extern void MazeAnalyze(constbitmap);
extern void MazeAnalyze2(constbitmap);
extern ulong MazeCountPossible(int, int);
extern void WriteText2(constbitmap, FILE *, flag);
extern void WriteTextDOS(constbitmap, FILE *, int, int);
extern void WriteText3D(cubemap, FILE *, flag, flag);

extern flag PerfectGenerate(bitmap, flag, int, int);
extern long BraidConnectWalls(bitmap);
extern long UnicursalApply(bitmap, flag);
extern flag RecursiveGenerate(bitmap, int, int);
extern flag PrimGenerate(bitmap, flag, flag, int, int);
extern flag TreeGenerate(bitmap, flag, int, int);
extern void EllerMakeRow(bitmap, short *, short *, int, int, int, int, int,
  flag);
extern void BraidMakeRow(bitmap, short *, short *, int, int, int, int, int,
  int);
extern flag FractalPartial(bitmap *, long, long, int, int,
  long *, long *);
extern flag WeaveGenerate(bitmap, flag, int, int);
extern flag ConvertWeaveTo3D(bitmap *, bitmap *);
extern flag ConvertWeaveToInside(bitmap *, colmap *, colmap *, bitmap *,
  int, KV, KV, KV);
extern flag FClarifyWeave(bitmap *, int, int, int);
extern flag CavernGenerate(bitmap, flag, int, int);
extern flag ZetaGenerate(bitmap, flag, int, int);
extern void SegmentGetSz(char *, int);
extern void SegmentParseSz(CONST char *);


/*
******************************************************************************
** Maze Creation Routines
******************************************************************************
*/

#define SpiralMax     30
#define SpiralWallMax 60
#define SpiralDead    0xFFFF

KRUS *KruskalFind(KRUS *);
void KruskalUnion(KRUS *, KRUS *);

extern flag CreateMazePerfect(bitmap *);
extern flag CreateMazePerfect2(bitmap *);
extern flag CreateMazeBraid(bitmap *);
extern flag CreateMazeUnicursal(bitmap *);
extern flag CreateMazeRecursive(bitmap *);
extern flag CreateMazePrim(bitmap *);
extern flag CreateMazeKruskal(bitmap *, flag, colmap, colmap);
extern flag CreateMazeTree(bitmap *);
extern flag CreateMazeAldousBroder(bitmap *);
extern flag CreateMazeWilson(bitmap *);
extern flag CreateMazeEller(bitmap *);
extern flag CreateMazeBraidEller(bitmap *);
extern flag CreateMazeDivision(bitmap *);
extern flag CreateMazeBinary(bitmap *);
extern flag CreateMazeSidewinder(bitmap *);
extern flag CreateMazeSpiral(bitmap *);
extern flag CreateMazeDiagonal(bitmap *);
extern flag CreateMazeSegment(bitmap *);
extern flag CreateMazeFractal(bitmap *);
extern int CreateMazeFractal2(bitmap *, colmap *);
extern flag CreateMazeInfinite(bitmap *, int);
extern flag CreateMazeWeave(bitmap *);
extern flag CreateMazeCavern(bitmap);
extern long CreateMazeCrack(bitmap, flag);
extern flag CreateMazeZeta(bitmap *);
extern flag CreateMazeArrow(bitmap *, colmap *);
extern flag CreateMazeDungeon(bitmap, colmap *, bitmap *,
  int, int, int, int, flag, int, int *, int *);

extern void CreateMazeTilt(bitmap);
extern flag CreateLabyrinth(bitmap *);
extern flag CreateLabyrinthChartres(bitmap *);
extern flag FCopyRgchToSzCustom(CONST char *, int, char *);
extern flag FValidSzCustom(CONST char *, int *, int *);
extern flag FComputeNextCustom(char *, flag);
extern flag CreateLabyrinthCustom(bitmap *);


/*
******************************************************************************
** Generic Maze Routines
******************************************************************************
*/

extern void CreateMazeGamma(bitmap *);
extern void CreateMazeDelta(bitmap *);
extern void CreateMazeSigma(bitmap *);
extern void CreateMazeTheta(bitmap *);
extern void CreateMazeUpsilon(bitmap *);


/*
******************************************************************************
** 3D Maze Routines
******************************************************************************
*/

#define dimMax 16
#define CLAR 10

#define FLegalMaze3(t, x, y, z) \
  ((x) > 0 && (x) < TX(t)-1 && (y) > 0 && (y) < TY(t)-1 && \
  (z) >= 0 && (z) <= TZ(t)-1)
#define FLegalMaze4(t, w, x, y, z) \
  ((x) > 0 && (x) < TX(t)-1 && (y) > 0 && (y) < TY(t)-1 && \
  (z) >= 0 && (z) <= TZ(t)-1 && (w) >= 0 && (w) <= TW(t)-1)
#define Count3(t, x, y, z) \
  (TG(t, x, y-1, z) + TG(t, x-1, y, z) + \
  TG(t, x, y+1, z) + TG(t, x+1, y, z) + \
  ((z) <= 0 || TG(t, x, y, z-1)) + ((z) >= Even(TZ(t))-2 || TG(t, x, y, z+1)))
#define Count4(t, w, x, y, z) \
  (UG(t, w, x, y-1, z) + UG(t, w, x-1, y, z) + \
  UG(t, w, x, y+1, z) + UG(t, w, x+1, y, z) + \
  ((z) <= 0 || UG(t, w, x, y, z-1)) + \
  ((z) >= Even(TZ(t))-2 || UG(t, w, x, y, z+1)) + \
  ((w) <= 0 || UG(t, w-1, x, y, z)) + \
  ((w) >= Even(TW(t))-2 || UG(t, w+1, x, y, z)))
#define Mg3(b, x, y, z, f3D) ((f3D) ? Qg3(b, x, y, z) : Mg(b, x, y))
#define Mg3Fast(b, x, y, z, f3D) ((f3D) ? Qg3(b, x, y, z) : MgFast(b, x, y))
#define FLegalLevel(b, x, y, f3D) \
  ((f3D) ? FLegalCubeLevel(b, x, y) : FLegal(b, x, y))

extern bit Qg(constcubemap, int, int, int);
extern bit Qg3(constcubemap, int, int, int);
extern flag CreateMaze3D(bitmap *);
extern flag CreateMazeBinary3D(bitmap *);
extern flag FClarify3D(bitmap *);
extern flag FColmapClarify3D(constcubemap, colmap *, KV, KV);
extern bit Rg(constcubemap, int, int, int, int);
extern flag CreateMaze4D(bitmap *);
extern flag FClarify4D(bitmap *);
extern flag FColmapClarify4D(constcubemap, colmap *, KV, KV);
extern flag FConvert4DTo3D(cubemap *, bitmap *, colmap *,
  KV, KV, KV, KV, KV);
extern flag CreateMazeHyper(bitmap *);
extern void MapND(constcubemap, CONST int *, int *, int *);
extern flag CreateMazeND(bitmap *);
extern void RenderND(bitmap, colmap, constcubemap, int *, int *, int *, int *,
  CONST long *, CONST long *, KV, int);
extern flag MoveND(constcubemap, long *, CONST long *, int *, int *, flag);


/*
******************************************************************************
** Planair Maze Routines
******************************************************************************
*/

// Normal: "a0a1a2a3"
// Torus: "a2a3a0a1"
// Cube: "e2d3f0b1 e7a3f3c1 e4b3f6d1 e1c3f5a1 c4d0a0b4 a2d6c6b2"

extern flag FValidPlanair(constbitmap, CONST char *);
extern flag CreateMazePlanair(bitmap *);


/*
******************************************************************************
** Maze Solving Routines
******************************************************************************
*/

extern long SolveMazeFillDeadEnds(bitmap, int, int, int, int);
extern long DoFillDeadEnds(bitmap, int, int, int, int);
extern long DoMarkDeadEnds(bitmap, int, int, int, int);
extern long DoMarkCulDeSacs(bitmap, int, int, int, int);
extern long SolveMazeFillCulDeSacs(bitmap, int, int, int, int);
extern long DoMarkBlindAlleys(bitmap);
extern long SolveMazeFillBlindAlleys(bitmap, int, int, int, int);
extern long DoMarkCollisions(bitmap, int, int);
extern long SolveMazeFillCollisions(bitmap, int, int, int, int);
extern long SolveMazeRecursive(bitmap, int, int, int, int, flag);
extern long SolveMazeShortest(bitmap, int, int, int, int, flag);
extern long SolveMazeShortest2(bitmap, int, int, int, int, flag);
extern long SolveMazeFollowWall(bitmap, int, int, int, int, int, flag);
extern long SolveMazePledge(bitmap, int, int, int, int, int, flag);
extern long SolveMazeChain(bitmap, int, int, int, int);
extern long SolveMazeTremaux(constbitmap, colmap *, int, int, int, int,
  KV, KV, KV, KV);
extern flag DoShowBottlenecks(bitmap, int, int, int, int, int);
extern long SolveMazeWeave(bitmap);
extern long SolveMaze3D(cubemap);
extern long SolveMaze4D(cubemap);
extern long SolveMazeND(cubemap);
extern long SolveMazeHyper(cubemap);
extern int SolveMazeFractal2(bitmap);

/* maze.h */
